 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTApplication.c
 *
 * Description: Receives, forks, and executes a binary application
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/
#define ROVER_ACCESSMGR

#include "sys/types.h"
#include "unistd.h"
#include "sys/stat.h"
#ifndef linux
#include "sys/uio.h"
#endif /* linux */
#include "fcntl.h"
#include <sys/wait.h>

#include "RoverBase.h"
#include "HTUtils.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTCache.h"
#include "HTSession.h"
#include "HTLog.h"
#include "HTNet.h"
#include "HTObject.h"
#include "RoverZlib.h"
#include "RoverTCP.h"


/* Definitions for procedures not in the include files */
extern HTConverter HTMIMEConvert;
extern char **environ;
extern int select(int nfds, fd_set *readfds, fd_set *writefds,
		  fd_set *exceptfds, struct timeval *timeout);
#ifdef decstation
extern int waitpid(int pid, union wait *status, int options);
#else
extern pid_t waitpid(int , int *, int);
#endif /* decstation */

#ifndef linux
extern int putenv(char *string);
#endif /* !linux */

extern int getdtablesize();

PUBLIC RoverQ **ServerWriteQ;
PUBLIC HTEventCallBack **Qhandlers;
PUBLIC int *Pipe2App;
PUBLIC int HTAppCount = 0;

PUBLIC HTEventState RoverServerHandlerRcv();

PRIVATE int App_abort(HTStream *me, HTError e);

struct _AppShare {
  RoverCacheable cache;
  dependV *dv;			/* Dependency vector */
  char *tdv;			/* Marshalled dependency vector */
  int dvsize;			/* Size of dependency vector */
  char sourceFileName[256];	/* Name of application's file */
  time_t loadDelay;
  time_t lastRef;
  int count;
};
typedef struct _AppShare AppShare;

/* The number of applications is limited by the size of the cache's
   callback vector: one 32-bit unsigned integer */
#define MAXAPPS 32
struct _HTApp {
  char *url;
  int pid;
  int topipe;
  int frompipe;
  AppShare *share;
  HTEventCallBack *user;
  RoverRcv Rstate;
  int updates;
};
typedef struct _HTApp HTApp;

PRIVATE Tcl_HashTable AppTable;
PRIVATE int RoverAppID = 0;
PRIVATE int *PipeTable;
PRIVATE	HTList *reqlist = NULL;		  	/* List of active requests */
PRIVATE char *signals[] = {
  NULL,
  "SIGHUP",
  "SIGINT",
  "SIGQUIT",
  "SIGILL",
  "SIGTRAP",
  "SIGIOT",
  "SIGBUS",
  "SIGFPE",
  "SIGKILL",
  "SIGUSR1",
  "SIGSEGV",
  "SIGUSR2",
  "SIGPIPE",
  "SIGALRM",
  "SIGTERM",
  "SIGSTKFLT",
  "SIGCHLD",
  "SIGCONT",
  "SIGSTOP",
  "SIGTSTP",
  "SIGTTIN",
  "SIGTTOU",
  "SIGURG",
  "SIGXCPU",
  "SIGXFSZ",
  "SIGVTALRM",
  "SIGPROF",
  "SIGWINCH",
  "SIGIO",
  "SIGPWR",
  "SIGUNUSED"
};

PRIVATE int App_create(HTStream *me, RoverCBList callback, QRPCState *state);
PRIVATE int App_process(HTStream *me, int verify);
PRIVATE HTApp *AppSearch(Tcl_HashSearch **search, int *appID);
PRIVATE void AppSearchEnd(Tcl_HashSearch **search);

#define APP_MEM_CACHED		 0
#define APP_DISK_CACHED		 1
#define APP_NOT_CACHED		-1

/* Create an application entry */
PRIVATE HTApp *AppCreate(int *appID)
{
  Tcl_HashEntry *hentry;
  int new;
  HTApp *apentry;

  /* Set up the entry */
  if (RoverDebug) fprintf(stderr, "%s: Creating application %d\n", myAppID,
			  RoverAppID);
  *appID = RoverAppID++;
  hentry = Tcl_CreateHashEntry(&AppTable, (char *) (*appID), &new);
  if (!new) {
    fprintf(stderr, "Found duplicate app ID %d [AppCreate:HTApplication.c]\n",
	    *appID);
    *appID = ROVER_ERR;
    return(NULL);
  }
  if ((apentry = (HTApp *)malloc(sizeof(HTApp))) == NULL) {
    fprintf(stderr, "Out of memory [AppCreate:HTApplication.c]\n");
    *appID = ROVER_NO_MEM;
    return(NULL);
  }

  /* Save cache entry */
  Tcl_SetHashValue(hentry, (char *) apentry);
  
  return(apentry);
}

/* Look up an application entry by ID */
PRIVATE HTApp *AppGetInfo(int appID)
{
  Tcl_HashEntry *hentry;

  if ((hentry = Tcl_FindHashEntry(&AppTable, (char *) appID)) == NULL) {
    fprintf(stderr,"App ID %d not found [AppGetInfo:HTApplication.c]\n",appID);
    return(NULL);
  }
  return((HTApp *) Tcl_GetHashValue(hentry));
}


/* Look up an application name by ID */
PUBLIC char *RoverAppGetURL(int appID)
{
  Tcl_HashEntry *hentry;

  if ((hentry = Tcl_FindHashEntry(&AppTable, (char *) appID)) == NULL) {
    fprintf(stderr,
	    "App ID %d not found [RoverAppGetURL:HTApplication.c]\n", appID);
    return(NULL);
  }
  return(((HTApp *) Tcl_GetHashValue(hentry))->url);
}

/* Look up an application pid by ID */
PUBLIC int RoverAppGetPID(int appID)
{
  Tcl_HashEntry *hentry;

  if ((hentry = Tcl_FindHashEntry(&AppTable, (char *) appID)) == NULL) {
    fprintf(stderr,
	    "App ID %d not found [RoverAppGetName:HTApplication.c]\n", appID);
    return(-1);
  }
  return(((HTApp *) Tcl_GetHashValue(hentry))->pid);
}

/* Look up an application ID by toPipe */
PUBLIC int RoverAppPipe2AppID(int toPipe)
{
  Tcl_HashSearch *search = NULL;
  HTApp *apentry;
  int appID;

  if (RoverDebug) fprintf(stderr, "%s: Checking cache for toPipe: %d\n",
			  myAppID, toPipe);

  while ((apentry = AppSearch(&search, &appID)) != NULL) {
    if (appID == 0) continue;		/* Skip Access Manager */
    if (apentry->topipe == toPipe) {
      if (RoverDebug)
	fprintf(stderr, "%s: Found %s w/ ID %d\n", myAppID,apentry->url,appID);
      AppSearchEnd(&search);
      return(appID);
    }
  }
  if (RoverDebug) fprintf(stderr, "%s: Cache entry for toPipe: %d NOT found\n",
			  myAppID, toPipe);
  return(-1);
}

/* Set watch information for an application */
PUBLIC RoverErr RoverAppSetWatch(int appID, int watch)
{
  Tcl_HashEntry *hentry;

  if ((hentry = Tcl_FindHashEntry(&AppTable, (char *) appID)) == NULL) {
    fprintf(stderr,
	    "App ID %d not found [RoverAppGetName:HTApplication.c]\n", appID);
    return(ROVER_ERR);
  }
  ((HTApp *) Tcl_GetHashValue(hentry))->updates = watch;
  return(ROVER_OK);
}


/* Delete an application by ID */
PRIVATE RoverErr AppDelete(int appID)
{
  Tcl_HashEntry *hentry;

  if ((hentry = Tcl_FindHashEntry(&AppTable, (char *) appID)) == NULL) {
    fprintf(stderr,
	    "App ID %d not found [AppDelete:HTApplication.c]\n", appID);
    return(ROVER_ERR);
  }
  free(Tcl_GetHashValue(hentry));
  Tcl_DeleteHashEntry(hentry);
  return(ROVER_OK);
}

/* Search for an application */
PRIVATE HTApp *AppSearch(Tcl_HashSearch **search, int *appID)
{
  Tcl_HashEntry *hentry;
  HTApp *apentry;

  if (*search == NULL) {
    if ((*search = (Tcl_HashSearch *)malloc(sizeof(Tcl_HashSearch))) == NULL){
      fprintf(stderr,"Out of memory! [AppSearch:HTApplication.c]\n");
      return((HTApp *) NULL);
    }
    hentry = Tcl_FirstHashEntry(&AppTable, *search); 
  } else 
    hentry = Tcl_NextHashEntry(*search);
  if (hentry) {
    apentry = (HTApp *) Tcl_GetHashValue(hentry);
    *appID = (int) Tcl_GetHashKey(&AppTable, hentry);
    return(apentry);
  }
  free(*search);
  *search = NULL;
  *appID = -1;
  return((HTApp *)NULL);
}

/* End the search for an application */
PRIVATE void AppSearchEnd(Tcl_HashSearch **search)
{
  if (*search) free(*search);
}

/* Check the in-memory hash table and Rover (memory and disk) caches for an
 *  application. Returns the location of the application, if found.
 */
PRIVATE int App_CheckCache(char *name, int *idx, char **fname,
			   RoverCacheable *cache, char **dv,
			   unsigned long *dvsize)
{
  Tcl_HashSearch *search = NULL;
  HTApp *apentry;
  int appID;
  RoverObjType type;

  if (RoverDebug) fprintf(stderr, "%s: Checking cache for app: %s\n", myAppID,
			  name);
  *idx = -1;
  *cache = ROVER_NO_CACHE;
  *fname = NULL;      
  *dv = NULL;
  *dvsize = 0;
  while ((apentry = AppSearch(&search, &appID)) != NULL) {
    if (appID == 0) continue;		/* Skip Access Manager */

    if (RoverDebug) fprintf(stderr,"%s: Comparing %s\n",myAppID,apentry ->url);
    if (!strcmp(apentry->url, name)) {
      *cache = apentry->share->cache;
      *idx = appID;
      if (RoverDebug)
	fprintf(stderr, "%s: Found %s in memory at %d: %s\n", myAppID,
		apentry->url, *idx, apentry->share->tdv);
      AppSearchEnd(&search);
      return(APP_MEM_CACHED);
    }
  }

  /* Not here already, check persistent cache */
  *fname = RoverCacheSearchName(name, ApplSession, cache, &type, dv, dvsize);
  if (*fname) {
    if (RoverDebug) fprintf(stderr, "%s: Found %s on disk\n", myAppID, name);
    return(APP_DISK_CACHED);
  } else {
    if (RoverDebug) fprintf(stderr, "%s: %s not found\n", myAppID, name);
    return(APP_NOT_CACHED);
  }
}

/* Invoke the update procedure for matching applications */
PUBLIC void RoverAppDoUpdate(int flag, RoverUpdateProc update)
{
  Tcl_HashSearch *search = NULL;
  HTApp *apentry;
  int appID;

  while ((apentry = AppSearch(&search, &appID)) != NULL)
    if (apentry->updates &flag) update(appID,apentry->topipe,apentry->updates);
}

PUBLIC void RoverUpdateApps(int appID, int toPipe, int watch)
{
  char *rbuf, *rbuf2, buf[1024];
  int size = 0, count = 0, i, j;
  Tcl_HashSearch *search = NULL;
  HTApp *apentry, *list[MAXAPPS];
  RoverErr rv;

  while ((apentry = AppSearch(&search, &i)) != NULL) {
    for (j = 0; j < count; j++)
      if (!strcmp(apentry->url, list[j]->url)) break;
    if (j != count) continue;	/* Found a duplicate name */
    if (RoverDebug) fprintf(stderr, "%s: Found application at %d. %s\n",
			    myAppID, i, apentry->url);
    list[count++] = apentry;
    size += (strlen(apentry->url)+sizeof(int)+12);
  }
  size += sizeof(int);
  if ((rbuf2 = rbuf = malloc(size)) == NULL) {
    fprintf(stderr,"Out of memory! [RoverUpdateApps:HTApplication.c]\n");
    return;
  }
  MarshallInt(&rbuf2, count);
  for (i = 0; i < count; i++) {
    sprintf(buf, "%s (%d) users", list[i]->url, list[i]->share->count);
    MarshallString(&rbuf2, buf);
  }
  ASSERT(size>=(rbuf2-rbuf));
  rv = RoverServerSend(toPipe, 0, ROVER_APP_UPDATE, size, rbuf, 1);
  if (rv != ROVER_OK)
    fprintf(stderr, "%s: Send fail %s [RoverUpdateApps:HTApplication.c]\n",
	    myAppID, RoverErrType(rv));
}

PUBLIC void RoverUpdateAppsAll()
{
  RoverAppDoUpdate(ROVER_WATCH_APPS, RoverUpdateApps);
}

/* Update the dependency vector in the app table and the cache */
PRIVATE RoverErr AppUpdateDependV(char *url, HTApp *apentry, char *tdv,
				  int dvsize)
{
  RoverErr rc;
  unsigned long server, clock;

  if (RoverDebug)
    fprintf(stderr, "%s: New DV for %s is %s (%d)\n", myAppID, url,
	    tdv, dvsize);
  if (apentry) {
    if (apentry->share->tdv) free(apentry->share->tdv);
    apentry->share->tdv = tdv;
    if (dvsize) {
      apentry->share->dvsize = sizeof(dependV);
      sscanf((char *) tdv, "%lu_%lu", &server, &clock);
    }
    if (apentry->share->dv) free(apentry->share->dv);
    if ((apentry->share->dv = (dependV *)malloc(sizeof(dependV))) == NULL) {
      fprintf(stderr, "Out of memory! [App_update:HTApplication.c]\n");
      return(ROVER_NO_MEM);
    }
    apentry->share->dv[0].ServerID = server;
    apentry->share->dv[0].ServerClock = clock;
  }
  rc = RoverCacheUpdateDependV(url, ApplSession, tdv, dvsize);
  if (rc != ROVER_OK) 
    fprintf(stderr, "%s: Cache update dv %d::%s failed %d [AppUpdateDependV:HTApplication.c]\n",
	    myAppID, ApplSession, url, rc);
  return(rc);
}

/* Set and return an application's load delay time and the current time */
PRIVATE int App_setLoadTime(HTStream *me, time_t *now, time_t *delay)
{
  QRPCState *context = (QRPCState *)(me->request->context);
  struct stat sbuf;
  RoverErr rc;
  unsigned long loadTime;

  if (me->cache == ROVER_NO_CACHE) return(HT_OK);

  time(now);
  if (context)
    loadTime = context->loadTime;
  else
    loadTime = *now;

  if (RoverDebug) fprintf(stderr, "Context? %s loadTime %lu now %lu\n",
			  context ? "yes" : "no", loadTime, *now);
  *delay = *now - loadTime;
  /* We are conservative in using the load time for the file's
     timestamp, since it could be an arbitrary amount of time
     between the time the file is requested and it is received.
     Eventually, this will be folded into the general object
     receive procedure and a timestamp vector may be used.
   */	 
  if (stat(me->sourceFileName, &sbuf)) {
    fprintf(stderr,"File %s stat fail: %s [App_setLoadTime:HTApplication.c]\n",
	    me->sourceFileName, strerror(errno));
    return(App_abort(me, (HTError) NULL));
  } 
  rc = RoverCacheUpdate(me->url, ApplSession, me->type, me->cache,
			me->sourceFileName, context->loadTime, *delay,
			*now, sbuf.st_size, NULL, me->tdv, me->dvsize);
  if (rc != ROVER_OK) return(HT_ERROR);
  return(HT_OK);
}

PRIVATE void App_put_block(HTStream *me, CONST char *c, int length)
{
  int len = 0, rc, count = 0;

  if (me->dsize) {	/* Skip data bytes (usually just a separating null) */
    while ((me->dsize > 0) && length) {
      me->dsize--;
      length--;
      c++;
    }
  } else if (me->alloc) {	/* Save the dependency vector */
    while ((me->alloc > 0) && length) {
      me->alloc--;
      length--;
      me->tdv[me->size++] = *(c++);
    }
  }
  if (!length) return;

  /* Otherwise, add it to the file */
  if (me->compress) {
    rc = Uncompress(me->inflate, me->sourceFile, (char *) c, &length, me->cbuf,
		    &len, 65536, &count);
    if ((rc == -1) || length) {
      fprintf(stderr, "Uncompress error [App_put_block:HTApplication.c]\n");
      AppDoAbort(me);
    }
    me->csize += count;
  } else {
    fwrite(c, length, 1, me->sourceFile);
    me->csize += length;
  }
}

PRIVATE void App_put_character(HTStream *me, CONST char c)
{
  App_put_block(me, (CONST char *)&c, 1);
}

PRIVATE void App_put_string(HTStream *me, CONST char *c)
{
  App_put_block(me, c, strlen(c));
}

PRIVATE int App_execute(HTStream *me, int *execute)
{
  QRPCState *context = (QRPCState *)(me->request->context), *next;
  int rc = HT_OK;

  *execute = 0;
  while (context != NULL) {
    next = context->next;
    if (context->execute) {	/* The execute bit is set on if we are
				   connected (or using a localhost server)
				   and want to verify the object */
      *execute = 1;
      rc = App_create(me, (RoverCBList)me->request->callback, context);
      if (rc != HT_OK) break;
    }
    context = next;
  }
  return(rc);
}

PRIVATE int App_free(HTStream *me)
{
  QRPCState *context = (QRPCState *)(me->request->context);
  struct stat sbuf;
  int logID, rc, execute;
  time_t now, delay;

  /* done with receiving the application; now we execute it */
  if (me->sourceFile != NULL) {
    if (RoverDebug)
      fprintf(stderr,
	      "%s: Closing %s [App_free:HTApplication.c]\n", myAppID,
	      me->sourceFileName);
    fflush(me->sourceFile);
    fclose(me->sourceFile);	/* Close the file and make it executable */
    chmod(me->sourceFileName, S_IXUSR|S_IRUSR);
    if (!stat(me->sourceFileName, &sbuf)) {
      if (me->csize != sbuf.st_size) {
	fprintf(stderr, "%s: Binary file size mismatch %d vs %d\n", myAppID,
		(int) sbuf.st_size, me->csize);
	return(App_abort(me, (HTError) NULL));
      }
    } else {
      fprintf(stderr, "%s: File %s stat fail: %s [App_free:HTApplication.c]\n",
	      myAppID, me->sourceFileName, strerror(errno));
      return(App_abort(me, (HTError) NULL));
    }
  }
    
  if (me->compress) free(me->cbuf);
  me->tdv[me->dvsize+1] = '\0';
  if (RoverDebug) {
    fprintf(stderr,"%s: Read %d header bytes `%s'\n",myAppID,me->size,me->tdv);
    fprintf(stderr, "%s: Read %d binary code bytes\n", myAppID,me->csize);
  }
  ASSERT(me->dvsize==me->size);

  if (context) {
    /* Shared contexts share the same log ID; so delete it only once */
    if (LogDelete(logID = context->logID, 0) != ROVER_OK) {
      fprintf(stderr,
	      "%s: Log ID delete failed [App_free:HTApplication.c]\n",myAppID);
      return(App_abort(me, (HTError) NULL));
    }
  }

  while (context != NULL) {
    if (logID != context->logID) {	/* Shared context = shared log ID */
      fprintf(stderr,
	      "%s: WARNING! Got different log ID [App_free:HTApplication.c]\n",
	      myAppID);
    }
    context->logID = 0;
    context = context->next;
  }

  if (me->cache == ROVER_VERIFIED) {
    /* Object is verification message */
    rc = App_process(me, 1);
  } else if (me->type == ROVER_BINARY_UPDATE) {
    /* Object is update message */
    rc = App_process(me, 0);
  } else {
    /* Otherwise, we have a new object */
    if (App_execute(me,&execute) != HT_OK) return(App_abort(me,(HTError)NULL));
    if (execute) return(HT_OK);
    
    rc = AppUpdateDependV(me->url, NULL, me->tdv, me->dvsize);
    if (rc != ROVER_OK) return(App_abort(me, (HTError) NULL));
    rc = App_setLoadTime(me, &now, &delay);
    if (rc != ROVER_OK) return(App_abort(me, (HTError) NULL));
  }

  if (rc != HT_OK) return(App_abort(me, (HTError) NULL));
  free(me);
  return(rc);
}

PRIVATE int App_process(HTStream *me, int verify)
{
  unsigned long dvsize;
  int i, AppID, execute = 0;
  char *old, *tdv;
  RoverCacheable cache;
  HTApp *apentry = NULL;
  RoverErr rc;

  /* Look for the original application */
  if (RoverDebug)
    fprintf(stderr, "%s: Cache %s message for %s\n", myAppID, 
	    (verify ? "verification" : "update"), me->url);

  i = App_CheckCache(me->url, &AppID, &old, &cache, &tdv, &dvsize);
  if (verify) me->cache = cache;/* Overwrite ROVER_VERIFIED w/ correct value */

  if (i == APP_MEM_CACHED) {  /* The app was already in the in-memory cache */ 
    if (RoverDebug) fprintf(stderr, "%s: In-memory cache hit\n", myAppID);
    if (!(apentry = AppGetInfo(AppID))) return(App_abort(me, (HTError) NULL));
    if (verify) {
      me->cache = apentry->share->cache;
      strcpy(me->sourceFileName, apentry->share->sourceFileName);
    } else {
      old = apentry->share->sourceFileName;
      if (old) unlink(old);			/* We control the file */
      strcpy(apentry->share->sourceFileName, me->sourceFileName);
      if (me->cache == ROVER_NO_CACHE)
	rc = RoverCacheFlush(me->url,ApplSession);
      else
	rc = RoverCacheUpdateName(me->url, ApplSession, me->sourceFileName);
      if (rc != ROVER_OK) return(HT_ERROR);

      apentry->share->cache = me->cache;
      if (App_setLoadTime(me, &(apentry->share->lastRef),
			  &(apentry->share->loadDelay)) != HT_OK)
	return(HT_ERROR);
    }

  } else if (i == APP_DISK_CACHED) {/* The app wasn't in the in-memory cache */
    if (RoverDebug) {
      fprintf(stderr, "%s: Persistent cache hit\n", myAppID);
      fprintf(stderr,"%s: Got cache %s msg for unloaded app %s\n", myAppID,
	      (verify ? "verify" : "update"), me->url);
    }
    if (verify) {
      strcpy(me->sourceFileName, old);
      me->dvsize = dvsize;
      if (tdv) {
	if (me->tdv) free(me->tdv);
	me->tdv = strdup(tdv);
      } else
	me->tdv = NULL;
    } else {
      if (old) unlink(old);			/* We control the file */
      if (me->cache != ROVER_NO_CACHE)
	rc = RoverCacheUpdateName(me->url, ApplSession, me->sourceFileName);
    }
    if (me->cache == ROVER_NO_CACHE) rc = RoverCacheFlush(me->url,ApplSession);

  } else if (i == APP_NOT_CACHED) {
    if (RoverDebug)
      fprintf(stderr, "%s: Got bad cache update message for %d\n", myAppID,
	      AppID);
    return(HT_ERROR);
  }

  if (App_execute(me, &execute) != HT_OK) return(App_abort(me, (HTError)NULL));
  if (execute) return(HT_OK);

  if (!verify)
    if (AppUpdateDependV(me->url, apentry, me->tdv, me->dvsize) != ROVER_OK)
      return(HT_ERROR);

  return(HT_OK);
}

/* Initialize the apentry data structure */
PRIVATE RoverErr App_init(char *url, int AppID, HTApp *apentry, HTStream *me,
			  int topipe, int frompipe, int reuseID, int pid)
{
  HTApp *apentry2;
  RoverErr rc;

  apentry->url = strdup(url);
  apentry->updates = ROVER_WATCH_NONE;
  apentry->pid = pid;
  apentry->topipe = topipe;
  apentry->frompipe = frompipe;
  apentry->Rstate.RcvState = ROVER_IDLE;
  apentry->Rstate.RcvCount = 0;

  if (reuseID >= 0) {	/* Share the existing cache entry */
    if (!(apentry2 = AppGetInfo(reuseID))) return(ROVER_ERR);
    apentry->share = apentry2->share;
    apentry->share->count++;
    time(&(apentry->share->lastRef));
    if (me && (me->cache != ROVER_NO_CACHE)) {
      rc = RoverCacheRef(url, ApplSession, apentry->share->lastRef);
      if (rc != ROVER_OK) return(rc);
    }
    if (RoverDebug)
      fprintf(stderr, "%s: Reusing application (%d users)\n", myAppID,
	      apentry->share->count);
  } else {
    apentry->share = (AppShare *)calloc(sizeof(AppShare),1);
    apentry->share->count = 1;

    if (me && App_setLoadTime(me, &(apentry->share->lastRef),
			      &(apentry->share->loadDelay)) != HT_OK)
      return(ROVER_ERR);

    if (RoverDebug)
      fprintf(stderr,"%s: New app %s (%d users), loaded in %lu secs\n",myAppID,
	      url, apentry->share->count, apentry->share->loadDelay);
  }

  if (me) {
    strcpy(apentry->share->sourceFileName, me->sourceFileName);
    apentry->share->cache = me->cache;
    rc = AppUpdateDependV(me->url, apentry, me->tdv, me->dvsize);
    if (rc != ROVER_OK) return(HT_ERROR);
  } else
    apentry->share->cache = ROVER_NO_CACHE;


  /* Enter the pipes into tables (for later fast lookup) */
  PipeTable[apentry->frompipe] = apentry->topipe;
  Pipe2App[apentry->frompipe] = -AppID;
  Pipe2App[apentry->topipe] = AppID;
  
  apentry->user = calloc(sizeof(HTEventCallBack),1);
  apentry->user->sockfd = apentry->frompipe;
  apentry->user->callback = RoverEventHandler;
  HTEventRegister(apentry->user, EVENT_READ);		  /* Register client */
  
  Qhandlers[apentry->topipe] = calloc(sizeof(HTEventCallBack),1);
  Qhandlers[apentry->topipe]->sockfd = apentry->topipe;
  Qhandlers[apentry->topipe]->callback = RoverEventHandler;
  
  HTAppCount++;
  return(ROVER_OK);
}

PUBLIC RoverErr RoverAppRemoteCreate(char *url, int sock, int pid,
				     char *hostname, char *username)
{
  HTApp *apentry;
  RoverErr rc;
  int AppID;
  char line[2048];

  apentry = AppCreate(&AppID);
  if (AppID < 0) {
    fprintf(stderr,"%s: App slot create error [RoverAppRemoteCreate:HTApplication.c]\n",
	    myAppID);
    return(ROVER_ERR);
  }
  
  sprintf(line, "%s:%d [%s@%s]", url, pid, username, hostname);
  if (RoverDebug) fprintf(stderr, "Creating app entry for %s\n", line);
  rc = App_init(line, AppID, apentry, NULL, sock, sock, -1, pid);
  if (rc != ROVER_OK) return(ROVER_ERR);

  RoverUpdateAppsAll();
  RoverServerHandlerRcv();	/* Perform a quick check */
  return(ROVER_OK);
}


#ifdef DEBUG_FDS
PRIVATE void RoverShowFDs(char *string)
{
  int fd, rv;

  fprintf(stderr, "%s\n", string);
  for (fd = 0; fd < 32; fd++)
    if ((rv = fcntl(fd, F_GETFD)) == -1) 
      fprintf(stderr, "Bad fd %d: %s\n", fd, strerror(errno));
    else 
      fprintf(stderr, "Good fd %d: %s\n", fd, ((rv&1) ? "close-on-exec" :
					       "leave open"));
}
#endif /* DEBUG_FDS */

PRIVATE int App_create(HTStream *me, RoverCBList callback, QRPCState *state)
{
  int i, app, status, fildes1[2], fildes2[2], pid, AppID, rc1, rc2, argc = 0;
  char *buf, **argv = NULL, *name, *tdv;
  HTApp *apentry;
  RoverErr rc;
  unsigned long dvsize;
  RoverCacheable cache;

  /* Create two pipes for bidirectional communication */
  rc1 = rc2 = 0;
  if (((rc1 = pipe(fildes1)) != 0) || ((rc2 = pipe(fildes2)) != 0)) {
    fprintf(stderr,
	    "%s: Pipe opens failed: %s (%d %d) [App_create:HTApplication.c]\n",
	    myAppID, strerror(errno), rc1, rc2);
    return(HT_ERROR);
  }
  
  /* Set close-on-exec flags */
  RoverSetCloseExec(fildes1[1]);
  RoverSetCloseExec(fildes2[0]);

  /* Don't want to block on I/O operations */
  RoverMakeSockNB(fildes1[0], 1);
  RoverMakeSockNB(fildes1[1], 1);
  RoverMakeSockNB(fildes2[0], 1);
  RoverMakeSockNB(fildes2[1], 1);

  /* Stuff the child's pipes into the environment we're going to pass to it */
  buf = (char *)malloc(sizeof(char)*17);
  sprintf(buf,"WRITEPIPE=%d", fildes2[1]);
  putenv(buf);

  buf = (char *)malloc(sizeof(char)*17);
  sprintf(buf,"READPIPE=%d", fildes1[0]);
  putenv(buf);

  status = App_CheckCache(me->url, &app, &name, &cache, &tdv, &dvsize);
  apentry = AppCreate(&AppID);
  if (AppID < 0) {
    fprintf(stderr,"%s: App slot create error [App_create:HTApplication.c]\n",
	    myAppID);
    return(HT_ERROR);
  }

  buf = (char *)malloc(sizeof(char)*17);
  sprintf(buf,"APP_ID=%d", AppID);
  putenv(buf);

  if (RoverDebug)
    fprintf(stderr, 
	    "%s: Client.%d's pipes: Read %d, Write %d\n", myAppID, AppID,
	    fildes1[0], fildes2[1]);

  /* Flush any open files before forking (to avoid duplicate buffer flushes) */
  fflush(NULL);
  fflush(RoverLogFile);

#ifdef DEBUG_FDS
  RoverShowFDs("before fork");
#endif /* DEBUG_FDS */

  /* Now, create the child */
  if ((pid = fork ()) == 0) { 
    /* All Access Manager sockets should be marked as close on exec,
     *   so there should not be any sockets left open after the exec.
     */

#ifdef DEBUG_FDS
    RoverShowFDs("after fork");
#endif /* DEBUG_FDS */

    /* Now exec the application */
    if (state && state->argc) {
      argc = state->argc;
      argv = state->argv;
      argv[0] = me->sourceFileName;
      if (RoverDebug) {
	fprintf(stderr, "%s: App got %d args:", myAppID, argc);
	for (rc1 = 0; rc1 < argc; rc1++)
	  fprintf(stderr, " |%s| ", argv[rc1]);
	fprintf(stderr, "\n");
      }
    }
    for (i = 2, rc2 = 0; rc2 < 4; rc2++, i <<= 1) {
      /* For the occasional "Text file busy" error, we loop */
      rc1 = execve(me->sourceFileName, argv, environ);
      if (errno != ETXTBSY) {
	fprintf(stderr,"%s: FATAL ERROR: Execve(%s) failed: %s (%d)\n",myAppID,
		me->sourceFileName, strerror(errno), errno);
	break;
      }
      fprintf(stderr,"%s: WARNING: Execve(%s) failed: %s (%d)\n",myAppID,
	      me->sourceFileName, strerror(errno), errno);
      sync();		/* Try forcing the filesystem to flush buffers */
      fprintf(stderr,"%s: Retrying after sleep(%d) seconds...\n",myAppID, i);
      sleep(i);
    }
    ASSERT(rc1=0);
    /* Never returns */
  }
  
  /* Close the child's pipes so we don't run out of file descriptors */
  close(fildes1[0]);
  close(fildes2[1]);

  rc = App_init(me->url, AppID, apentry, me, fildes1[1], fildes2[0], app, pid);
  if (rc != ROVER_OK) return(HT_ERROR);

  /* Perform a callback on the request */
  if (callback) {
    state->type = AppID;
    rc1 = (HTCallbackTable[callback])(me->request, state);
    if (rc1 < 0) {
      fprintf(stderr,
	      "%s: Callback error %d [App_create:HTApplication.c]\n", myAppID,
	      rc1);
      return(App_abort(me, (HTError) NULL));
    }
  }
  RoverUpdateAppsAll();
  RoverServerHandlerRcv();	/* Perform a quick check */
  return(HT_OK);
}

PUBLIC int RoverAMAppInit()
{
  int AppID = 0, nfds, rc1;
  HTApp *apentry;

  Tcl_InitHashTable(&AppTable, TCL_ONE_WORD_KEYS);

  /* Initialize entry for Rover Access Manager */
  apentry = AppCreate(&AppID);
  if (AppID != 0) {
    fprintf(stderr,
	    "%s: App slot create error [HTAppInit:HTApplication.c]\n",myAppID);
    return(ROVER_ERR);
  }

  apentry->share = (AppShare *)calloc(sizeof(AppShare),1);
  apentry->share->cache = 0;
  apentry->url = "Rover Access Manager";
  apentry->share->tdv = NULL;
  strcpy(apentry->share->sourceFileName, RoverArgv[0]);
  apentry->share->dvsize = 0;
  apentry->share->dv = NULL;
  apentry->updates = ROVER_WATCH_NONE;
  apentry->pid = 0;
  apentry->topipe = 0;
  apentry->frompipe = 0;
  apentry->Rstate.RcvState = ROVER_IDLE;
  apentry->Rstate.RcvCount = 0;
  apentry->share->count = 1;
  time(&(apentry->share->lastRef));
  apentry->user = NULL;

  /* Alloc and init PipeTable, Pipe2App, and ServerWriteQ */
  nfds = getdtablesize();
  PipeTable = (int *) calloc(nfds,sizeof(int));
  Pipe2App = (int *) calloc(nfds,sizeof(int));
  ServerWriteQ = (RoverQ **) calloc(nfds, sizeof(RoverQ *));
  Qhandlers = (HTEventCallBack **)calloc(nfds, sizeof(HTEventCallBack *));
  for (rc1 = 0; rc1 < nfds; PipeTable[rc1++] = -1);

  if (ApplSession == -1)
    ApplSession = RoverSessionServerCreate("appl",
					   (ROVER_NO_CONSIST|ROVER_ACCEPT_TENT),
					   ROVER_MH, 0, RoverNobodyUser, 10,
					   "PassWord!");

  if (ApplSession < 0) {
    fprintf(stderr,
	    "%s: Unable to locate server for `appl' session\n", myAppID);
    return(ROVER_INVLD_SES);
  }
  apentry->user = NULL;
  return(ROVER_OK);
}

PRIVATE int RoverCleanupChild(int pid)
{
  Tcl_HashSearch *search = NULL;
  HTApp *apentry;
  int appID;

  while ((apentry = AppSearch(&search, &appID)) != NULL) {
    if (apentry->pid == pid) {
      /* Clean up after the child */
      fprintf(stderr, "%s: App %s [pid: %d] (File: %s) ended\n", myAppID,
	      apentry->url, pid,(apentry->share->sourceFileName ?
				 apentry->share->sourceFileName : "<remote>"));
      --apentry->share->count;
      fprintf(stderr, "%s: Application now has %d users\n", myAppID,
	      apentry->share->count);
      if (apentry->share->count == 0) {
	if (RoverDebug)
	  fprintf(stderr, "%s: Deleting shared application state\n", myAppID);
	if (apentry->share->cache == ROVER_NO_CACHE) 
	  unlink(apentry->share->sourceFileName); /* We control the file */
	apentry->share->sourceFileName[0] = '\0';
	if (apentry->share->dv) free(apentry->share->dv);
	if (apentry->share->tdv) free(apentry->share->tdv);
	free(apentry->share);
      }
      close(apentry->topipe);
      PipeTable[apentry->frompipe] = -1;
      Pipe2App[apentry->frompipe] = 0;
      Pipe2App[apentry->topipe] = 0;
      close(apentry->frompipe);
      apentry->url = NULL;
      apentry->pid = 0;
      apentry->frompipe = -1;
      HTEventDeregister(apentry->user, EVENT_READ);
      if (apentry->user) free(apentry->user);
      if (ServerWriteQ[apentry->topipe])
	HTEventDeregister(Qhandlers[apentry->topipe], EVENT_WRITE);
      if (Qhandlers[apentry->topipe]) free(Qhandlers[apentry->topipe]);
      apentry->topipe = -1;
      apentry->Rstate.RcvCount = 0;
      apentry->Rstate.RcvState = ROVER_IDLE;
      HTAppCount--;
      AppDelete(appID);
      AppSearchEnd(&search);
      RoverUpdateAppsAll();
      return(1);	
    }
  }
  return(0);
}

PUBLIC void RoverAppCloseAll()
{
  HTApp *apentry;
  Tcl_HashSearch *search = NULL;
  int appID;
  
  /* Clean up after the children */
  while ((apentry = AppSearch(&search, &appID)) != NULL) {
    close(apentry->topipe);
    close(apentry->frompipe);
  }
}

PUBLIC void RoverAppPipeFlush()
{
  Tcl_HashSearch *search = NULL;
  HTApp *apentry;
  int appID;

  /* Flush any pending writes to client applications */
  while ((apentry = AppSearch(&search, &appID)) != NULL) {
    if ((ServerWriteQ[apentry->topipe] != NULL) &&
	(RoverFlushQ(&(ServerWriteQ[apentry->topipe])) != 0)) {
      if (RoverDebug)
	fprintf(stderr, "%s: Deregistering for write on %d [RoverAppPipeFlush:HTApplication.c]\n",
		myAppID, apentry->topipe);
      HTEventDeregister(Qhandlers[apentry->topipe], EVENT_WRITE);
    }
  }
}

PUBLIC HTEventState RoverServerHandlerRcv()
{
  RoverErr rc;
  int pid;
#ifdef decstation
  union wait status;
#else
  int status;
#endif

  pid = waitpid(-1, &status, WNOHANG);	/* Get children's status */
  if (RoverDebug) {
    fprintf(stderr, "%s: Childrens' status is %d ", myAppID, pid);
    if (pid > 0) {
      if (WIFEXITED(status))
	fprintf(stderr, "(normal exit: %d)\n", WEXITSTATUS(status));
      else if (WIFSIGNALED(status))
	fprintf(stderr, "(killed by signal: %d/%s)\n",
		WTERMSIG(status), signals[WTERMSIG(status)]);
      else
	fprintf(stderr, "(no information)\n");
    } else
      fprintf(stderr, "\n");
  }
  if (pid == -1) {				/* No one's alive */
    fprintf(stderr, "%s: No one's alive...exiting (%d left)\n", myAppID,
	    HTAppCount);
    RoverSessionSave();
    if ((rc = RoverCacheSave()) != ROVER_OK) 
      fprintf(stderr, "Cache save failed during shutdown\n");
    LogClose();
    return(EVENT_QUIT);
  } else if (pid > 0) {			/* Someone ended */
    if (!RoverCleanupChild(pid)) {
      fprintf(stderr,
	      "%s: WARNING! Child pid %d not found [RoverServerHandlerRcv:HTApplication.c]\n", myAppID, pid);
    }
    if (HTAppCount <= 0) {	/* Everyone's gone */
      fprintf(stderr,"%s: No applications left...exiting\n", myAppID);
      RoverSessionSave();
      if ((rc = RoverCacheSave()) != ROVER_OK) 
	fprintf(stderr, "Cache save failed during shutdown\n");
      LogClose();
      return(EVENT_QUIT);
    }
    RoverUpdateAppsAll();
  } else {				/* All OK */ 
    RoverServerRcvCheck();
  }
  return(EVENT_OK);
}

/* This procedure is called to check for pending data from any ROVER clients.
   It returns immediately if no messages are pending.
 */   
PUBLIC void RoverServerRcvCheck()
{
  RoverInRcvCheck = 1;		/* Prevent recursion */
  RoverServerRcvProc(0);
  RoverInRcvCheck = 0;
}

/* This procedure is called to check for pending data from any ROVER clients.
   It blocks until a message is ready.
 */   
PUBLIC void RoverServerRcvWait()
{
  RoverInRcvCheck = 1;		/* Prevent recursion */
  RoverServerRcvProc(1);
  RoverInRcvCheck = 0;
}

PUBLIC void RoverServerRcvProc(int dowait)
{
  int i, j, ready = 0, nfds = 0;
  fd_set readfds, writefds;
  struct timeval timeout;
  Tcl_HashSearch *search = NULL;
  HTApp *apentry;
  int appID;

  if (RoverDebug)
    fprintf(stderr, "%s: Preparing to select...\n", myAppID);
  FD_ZERO(&readfds);
  FD_ZERO(&writefds);

  while ((apentry = AppSearch(&search, &appID)) != NULL) {
    if (appID == 0) continue;	/* Skip ourselves */

    /* Test the read pipe to make sure it is still valid */
    if (!RoverTCPValidSocket(apentry->frompipe)) {
      if (!RoverCleanupChild(apentry->pid)) {
	fprintf(stderr,
		"%s: WARNING! Child pid %d not found [RoverServerRcvProc:HTApplication.c]\n",
		myAppID, apentry->pid);
      }
      continue;
    }

    /* Add from pipe to select read set */
    if (RoverDebug) fprintf(stderr, "%s: Adding %d to wait read sets\n",
			    myAppID, apentry->frompipe);
    FD_SET(apentry->frompipe, &readfds);
    if (nfds <= apentry->frompipe) nfds = apentry->frompipe + 1;

    if (ServerWriteQ[apentry->topipe] != NULL) {
      if (RoverDebug) fprintf(stderr, "%s: Adding %d to wait write sets\n",
			      myAppID, apentry->topipe);
      FD_SET(apentry->topipe, &writefds);
      if (nfds <= apentry->topipe) nfds = apentry->topipe + 1;
    }
  }
  if (RoverDebug)
    fprintf(stderr, "%s: Calling select on %d fds (%s)...\n", myAppID,
	    nfds, (dowait) ? "waiting" : "polling");
  if (dowait) {
    timeout.tv_sec = 0L;		/* 0 seconds */
    timeout.tv_usec = 200L;		/* and 200 microseconds */
    ready = select(nfds, &readfds, &writefds,NULL,&timeout); /* Short wait */
  } else {
    timeout.tv_sec = 0L;		/* 0 seconds */
    timeout.tv_usec = 0L;		/* and 0 microseconds */
    ready = select(nfds,&readfds,&writefds,NULL,&timeout); /* Poll children */
  }
  if (RoverDebug) fprintf(stderr, "%s: Done selecting: %d\n", myAppID, ready);
  if (ready < 0) {
    fprintf(stderr,
	    "%s: Select error %d: %s (%d) [RoverServerRcvProc:HTApplication.c]\n", myAppID,
	    ready, strerror(errno), errno);
    return;
  } else if (!ready) return;
      
  for (j = 0; j < nfds; j++) {
    /* Check the read pipe */
    if (FD_ISSET(j, &readfds)) {
      i = abs(Pipe2App[j]);
      if ((apentry = AppGetInfo(i)) == NULL) {
	fprintf(stderr, "%s: AppGetInfo failed for appID %d [RoverServerRcvProc:HTApplication.c]\n",
		myAppID, i);
	continue;
      }
      if (RoverDebug)
	fprintf(stderr,
		"%s: File descriptor %d (AppID %d/%d) ready for reading\n", myAppID, j, Pipe2App[j], i);
      RoverServerRcv(j, PipeTable[j], &(apentry->Rstate));
      if (apentry->Rstate.RcvState == ROVER_PROCESS_DIED) {
	if (!RoverCleanupChild(apentry->pid)) {
	  fprintf(stderr,
		  "%s: WARNING! Child pid %d for entry %d  not found [RoverServerRcvProc:HTApplication.c]\n", myAppID,
		  apentry->pid, i);
	}
	if (HTAppCount <= 0) 	/* Everyone's gone */
	  return;
      }
      ready--;
    }
    /* Check the write pipe */
    if (FD_ISSET(j, &writefds)) {
      i = abs(Pipe2App[j]);
      if (RoverDebug)
	fprintf(stderr,
		"%s: File descriptor %d (AppID %d/%d) ready for writing\n", myAppID, j, Pipe2App[j], i);
      if (RoverFlushQ(&(ServerWriteQ[j])) != 0) {
	if (RoverDebug)
	  fprintf(stderr,
		  "%s: Deregistering for write on %d [RoverServerRcvProc:HTApplication.c]\n", myAppID, j);
	HTEventDeregister(Qhandlers[j], EVENT_WRITE);
      }
      ready--;
    }
  }
  if (ready) {
    fprintf(stderr,
	    "%s: ready != 0 %d [RoverServerRcvProc:HTApplication.c]\n", myAppID,
	    ready);
    return;
  }
}

PRIVATE int App_abort(HTStream *me, HTError e)
{
  /* Abort for some reason. Clean up by closing file and unlinking it */
  fprintf(stderr, "%s: Aborting application load\n", myAppID);
  if (me->sourceFile) {
    fclose(me->sourceFile);
    unlink(me->sourceFileName);
  }
  if (me->cache != ROVER_NO_CACHE) RoverCacheFlush(me->url, ApplSession);
  if (me->tdv) free(me->tdv);
  if (me->compress) free(me->cbuf);
  free(me);
  return EOF;
}

PRIVATE void App_putcAbort(HTStream *me, CONST char c) {
  return;
}

PRIVATE void App_putsAbort(HTStream *me, CONST char *c) {
  return;
}

PRIVATE void App_putbAbort(HTStream *me, CONST char *c, int length) {
  return;
}

PRIVATE int App_freeAbort(HTStream *me)
{
  /* Abort for some reason. Clean up by closing file and unlinking it */
  fprintf(stderr, "%s: Finishing aborted application load\n", myAppID);
  return(App_abort(me, (HTError) NULL));
}

/*		Structured Object Class
**		-----------------------
*/
PRIVATE CONST HTStreamClass HTApplicationAbort =
{		
	"Application Runner Abort",
	App_freeAbort,
	App_abort,
	App_putcAbort, App_putsAbort, App_putbAbort
}; 

PUBLIC int AppDoAbort(HTStream *me)
{
  /* Abort for some reason. Clean up by closing file and unlinking it */
  fprintf(stderr, "%s: Aborting application load\n", myAppID);
  me->isa = &HTApplicationAbort;
  return EOF;
}

/*		Structured Object Class
**		-----------------------
*/
PUBLIC CONST HTStreamClass HTApplication =
{		
	"Application Runner",
	App_free,
	App_abort,
	App_put_character, App_put_string, App_put_block
}; 


/*		New object
**		----------
*/
PUBLIC HTStream* RoverAppRun ARGS5(
	HTRequest *,		request,
	void *,			param,
        HTFormat,               input_format,
        HTFormat,               output_format,
	HTStream *,		output_stream)
{
  fprintf(stderr, 
	  "Warning! Application received as application/x-application!\n");
  return(HTStream*) NULL;
}

PRIVATE int App_TableExec(int idx, RoverCBList callback, QRPCState *state)
{
  HTStream* me;
  HTRequest *req;
  HTApp *apentry;

  if (RoverDebug) fprintf(stderr, "Cache hit in AppTable\n");

  /* Allocate a new entry and copy over the old one */
  if ((apentry = AppGetInfo(idx)) == NULL) return(ROVER_ERR);
  me = (HTStream*)calloc(sizeof(*me),1);
  me->isa = &HTApplication;
  me->dvsize = strlen(apentry->share->tdv)+1;
  me->tdv = strdup(apentry->share->tdv);
  /* Get a request struct */
  req = Thread_new();
  /* Fill in the important info */
  req->method = METHOD_GET;		/* Default, but set it anyway */
  req->callback = (void *) callback;
  req->context = (void *) state;	/* Save for later use */
  state->logID = 0;
  state->IDsent = 0;
  state->marshallObj = NULL;
  state->marshallSize = 0;
  me->request = req;
  me->anchor = NULL;
  me->url = apentry->url;
  me->sourceFile = NULL;
  strcpy(me->sourceFileName, apentry->share->sourceFileName);
  return(App_create(me, callback, state));
}

PRIVATE int App_CacheExec(char *name, char *fname, RoverCBList callback,
			    QRPCState *state, RoverCacheable cache, char *dv,
			    int dvsize)
{
  HTStream* me;
  HTRequest *req;

  /* Allocate a new entry and copy over the important info */
  if (RoverDebug)
    fprintf(stderr, "Persistent cache hit\n");
  
  me = (HTStream*)calloc(sizeof(*me),1);
  me->isa = &HTApplication;
  me->tdv = dv;
  me->dvsize = dvsize;
  me->cache = cache;

  /* Get a request struct */
  req = Thread_new();
  /* Fill in the important info */
  req->method = METHOD_GET;		/* Default, but set it anyway */
  req->callback = (void *) callback;
  req->context = (void *) state;	/* Save for later use */
  state->logID = 0;
  state->IDsent = 0;
  state->marshallObj = NULL;
  state->marshallSize = 0;
  me->request = req;
  me->anchor = NULL;	/* Hope this doesn't cause problems */
  me->url = name;
  strcpy(me->sourceFileName, fname);
  me->sourceFile = NULL;
  return(App_create(me, callback, state));
}

/* Lookup a url in the AppTable and fork off a new child if found */
PUBLIC int App_LocalExec(char *url, RoverCBList callback, QRPCState *state)
{
  int idx, Qid, i;
  char *tdv, *dv = NULL, *fname = NULL, *buf;
  RoverCacheable cache;
  unsigned long dvsize;
  QRPCState *state2;
  RoverSession *sentry;
  HTApp *apentry;

  if ((buf = strdup(url)) == NULL) {
    fprintf(stderr, "Out of memory [App_LocalExec:HTApplication.c]\n");
    return(ROVER_NO_MEM);
  }
  url = buf;	/* Copy this for our invocation below in case url is freed by
		   the caller */
  state->clientDvsize = 0;
  state->clientDv = NULL;
  state->obj = NULL;
  state->IDsent = 0;
  state->marshallObj = NULL;
  state->marshallSize = 0;
  state->execute = 1;
  i = App_CheckCache(url, &idx, &fname, &cache, &tdv, &dvsize);
  if (tdv) dv = strdup(tdv);
  if (i == APP_NOT_CACHED) {	/* Not here? */
    /* Go fetch the object */
    return(RoverSessionLoadApp(url, callback, state, MAX_NET_PRIORITY));
  }

  switch(cache) {
  case ROVER_NO_CACHE:		/* Do not cache */
    fprintf(stderr,
	    "App %s is NO_CACHE in cache [App_LocalExec:HTApplication.c]\n", 
	    url);
    return(ROVER_ERR);
    break;
    
  case ROVER_CACHE_FOREVER:	/* Can be cached forever (immutable) */
    /* Away we go */
    if (idx >= 0) {
      return(App_TableExec(idx, callback, state));
    } else
      return(App_CacheExec(url, fname, callback, state, cache, dv, dvsize));
    break;
    
  case ROVER_VERIFY:		/* Verify object before returning from cache */
  case ROVER_BEST_EFFORT:	/* Periodically check object is up-to-date  */
    if (!(sentry = RoverSessionLookup(ApplSession))) {
      fprintf(stderr,
	      "ApplSession is invalid! %d [App_LocalExec:HTApplication.c]\n",
	      ApplSession);
      return(ROVER_INVLD_SES);
    }
    if ((cache == ROVER_VERIFY) ||
	(RoverNetState.NetAvail == IP) ||
	!strncasecmp(sentry->CurrentServer, "localhost", 9)) {
      /* If there's an IP network interface available OR the service
	 is locally available, verify the object before returning it.
	 Otherwise, enqueue a verify request and return the object.
	 */
      if (idx >= 0) {
	if ((apentry = AppGetInfo(idx)) == NULL) return(ROVER_ERR);
  	if (RoverDebug)
	  fprintf(stderr, "Found binary in cache at %d, verifying %s\n", idx,
		  apentry->share->tdv);
	state->obj = NULL;
	state->dvsize = strlen(apentry->share->tdv)+1;

	if ((state->dv = strdup(apentry->share->tdv)) == NULL) {
	  fprintf(stderr,"Out of memory! [App_LocalExec:HTApplication.c]\n");
	  return(ROVER_NO_MEM);
	}
      } else {
	state->obj = NULL;
	if (RoverDebug)
	  fprintf(stderr,
		  "Verifying binary found in persistent cache %s, %ld bytes\n",
		  dv, dvsize);
	state->dvsize = dvsize;
	if ((state->dv = strdup(dv)) == NULL) {
	  fprintf(stderr, "Out of memory! [App_LocalExec:HTApplication.c]\n");
	  return(ROVER_NO_MEM);
	}
      }
      return(RoverSessionLoadApp(url, callback, state, MAX_NET_PRIORITY));
    }
    if ((state2 = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
      fprintf(stderr,"Out of memory! [App_LocalExec:HTApplication.c]\n");
      return(ROVER_NO_MEM);
    }
    if ((state2->url = strdup(url)) == NULL) {
      fprintf(stderr, "Out of memory [App_LocalExec:HTApplication.c]\n");
      return(ROVER_NO_MEM);
    }
    state2->session = ApplSession;
    state2->execute = state2->clientDvsize = state2->argc = state->IDsent = 0;
    state->marshallObj = NULL;
    state->marshallSize = 0;
    state2->next = NULL;
    state2->clientDv = NULL;
    state2->argv = NULL;
    if (idx >= 0) {
      if ((apentry = AppGetInfo(idx)) == NULL) return(ROVER_ERR);
      if (RoverDebug)
	fprintf(stderr,
		"Found binary in cache at %d, best effort verifying %s\n", idx,
		apentry->share->tdv);
      state2->obj = NULL;
      state2->dvsize = strlen(apentry->share->tdv)+1;
      state2->dv = strdup(apentry->share->tdv);
      if (!state2->dv) {
	fprintf(stderr, "Out of memory! [App_LocalExec:HTApplication.c]\n");
	return(ROVER_NO_MEM);
      }
    } else {
      state2->obj = NULL;
      if (RoverDebug)
	fprintf(stderr,
		"Found binary in persistent cache, best effort verifying %s (%ld bytes)\n",
		dv, dvsize);
      state2->dvsize = dvsize;
      if ((state2->dv = malloc(state2->dvsize)) == NULL) {
	fprintf(stderr, "Out of memory! [App_LocalExec:HTApplication.c]\n");
	return(ROVER_NO_MEM);
      }
      strcpy(state2->dv, dv);
    }
    Qid = RoverSessionLoadApp(url, VerifyCB, state2, MIN_NET_PRIORITY);
    if (Qid < 0) {
      fprintf(stderr,
	      "Session read failed %s [App_LocalExec:HTApplication.c]\n",
	      RoverErrType(Qid));
      return(Qid);
    }
    if (idx >= 0) 
      return(App_TableExec(idx, callback, state));
    else
      return(App_CacheExec(url, fname, callback, state, cache, dv, dvsize));
    break;
    
  case ROVER_VERIFIED:		/* Verification of client's copy of object */
  case ROVER_CALLBACK:		/* Object registered for server callback */
  case ROVER_LEASE:		/* Object valid for specific time period */
  default:
    fprintf(stderr,
	    "Found invld/unimpl'd cache type %d (appl::%s) [App_LocalExec:HTApplication.c]\n",
	    (int) cache, url);
    return(ROVER_ERR);
    break;
  }
  return(ROVER_ERR);
}

/* ------------------------------------------------------------------------- */
/*				EVENT FUNCTIONS				     */
/* ------------------------------------------------------------------------- */

/*
**  This function creates a new request structure and adds it to the global
**  list of active threads
*/
PUBLIC HTRequest *Thread_new()
{
  HTRequest *newreq = HTRequest_new();		/* Set up a new request */
  if (!reqlist)
    reqlist = HTList_new();
  HTList_delete(newreq->conversions);
  if (!RoverConversions) RoverConversionInit();
  newreq->conversions = RoverConversions;	/* Take from global list */
  HTList_addObject(reqlist, (void *) newreq);
  return newreq;
}


/*
**  This function deletes a request structure and takes it out of the list
**  of active threads.
*/
PUBLIC void Thread_delete(HTRequest *oldreq)
{
  if (oldreq) {
    if (reqlist)
      HTList_removeObject(reqlist, (void *) oldreq);
    if (oldreq->conversions == RoverConversions)
      oldreq->conversions = NULL;	    /* We keep them in a global list */
    HTRequest_delete(oldreq);
  }
}


/*
**  This function deletes the whole list of active threads.
*/
PUBLIC void Thread_deleteAll()
{
  BOOL first=YES;	      /* We only have one global list of conversions */
  if (reqlist) {
    HTList *cur = reqlist;
    HTRequest* pres;
    while ((pres = (HTRequest *) HTList_nextObject(cur)) != NULL) {
      if (first) {
	HTRequest_delete(pres);
	first = NO;
      } else {
	pres->conversions = NULL;
	HTRequest_delete(pres);
      }
    }
    HTList_delete(reqlist);
    reqlist = NULL;
  }
}


/*						        HTEventRequestTerminate
**
**	React to the status of the terminated request
*/
PUBLIC HTEventState HTEventRequestTerminate(HTRequest *actreq, int status) 
{
  Thread_delete(actreq);
  return EVENT_OK;
}


PUBLIC void RoverConversionInit()
{
  /* Initialize conversions */
  RoverConversions = HTList_new();	
  HTSetConversion(RoverConversions, "www/mime", "*/*", HTMIMEConvert, 1.0,
		  0.0, 0.0);
  HTSetConversion(RoverConversions, "application/x-object", "www/present",
		  ObjectReceive, 1.0, 0.0, 0.0);
  /*  HTFormatInit(RoverConversions); */
  HTTimerRegister(RoverTimer, 1);

}

/*								   EventHandler
**
**	Read in the user's input, and deal with it as necessary.
**
**	Any Command which works returns from the routine. If nothing
**	works then a search or error message down at the bottom.
*/
PUBLIC HTEventState RoverEventHandler(HTRequest **actreq)
{ 
  HTEventState status = EVENT_OK;
  status = RoverServerHandlerRcv();
  return status;
}



