/**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTApplication.h
 *
 * Description: Receives, forks, and executes a binary application
 *
 * Last Modified:  8-Feb-95 (adj)
 *
 * Global Functions:
 *    PUBLIC HTStream *Rover(HTRequest *request,
 *			     void *param,
 *			     HTFormat input_format,
 *			     HTFormat output_format,
 *			     HTStream *output_stream)
 *        Converter for application/x-application MIME stream. Downloads
 *        application binary code, creates pipes for IPC, and forks off
 *        a separate process.
 *
 * Global Variables: PUBLIC CONST HTStreamClass HTApplication
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#ifndef HTAPP_H
#define HTAPP_H

#include "RoverErr.h"
#include <stdio.h>
#include "HTUtils.h"
#include "tcp.h"
#include "HTStream.h"
#include "HTAnchor.h"
#include "HTString.h"
#include "HTError.h"
#include "RoverTypes.h"
#include "time.h"
#include "RoverEvent.h"

extern HTStream * RoverAppRun PARAMS((HTRequest *     request,
				      void *          param,
				      HTFormat        input_format,
				      HTFormat        output_format,
				      HTStream *      output_stream));

extern HTRequest *Thread_new();
extern void Thread_delete(HTRequest *oldreq);
extern void Thread_deleteAll();
extern void RoverConversionInit();
extern HTList *RoverConversions;	       /* Format conversions */
extern HTEventState RoverEventHandler(HTRequest **actreq);

#ifdef sparc
#define memmove	memcpy
#endif

#if defined(ROVER_CLIENT) || defined(ROVER_ACCESSMGR)

#if defined(ultrix)
#include <tcl.h>
#include <mit/X11/Xlib.h>
#include <tk.h>
#else
#include <tcl.h>
#include <tk.h>
extern void bzero(void *b1, int length);
extern int strcasecmp(const char *s1, const char *s2);
#endif /* linux */

#if 0
#ifndef DEBUG
#define DEBUG
#endif /* DEBUG */
#endif

typedef void (*RoverUpdateProc)(int appID, int toPipe, int watch);

typedef struct _RoverQ RoverQ;
struct _RoverQ {
  void (*rcv)();
  int size;
  char *data;
  char *ptr;
  int left;
  int tp;
  char *err;
  int dofree;
  RoverQ *next;
};

typedef enum _RoverType {
  ROVER_HELLO = 1,		/* Initial client->server hello msg */
  ROVER_HELLO_1,		/* Initial client ->server hello msg (v1.0) */
  ROVER_REPLY,			/* Generic reply msg */
  ROVER_ERROR_REPLY,		/* Error reply object */
  ROVER_NEW_SESSION,		/* Create a new session */
  ROVER_END_SESSION,		/* Delete a session */
  ROVER_INV,			/* Perform local (ROVER space) invocation */
  ROVER_RINV,			/* Perform remote (server space) invocation */
  ROVER_IMPORT_OBJ,		/* Import an object into client space */
  ROVER_IMPORT_DATA,		/* Import a data object into client space */
  ROVER_EXPORT_OBJ,		/* Export an object back to a server */
  ROVER_OBJECT,  		/* Message containing an object from ROVER */
  ROVER_CACHE_OBJECT,		/* Local cache response */
  ROVER_FLUSH_HTTP,		/* Flush an http object from server's cache */
  ROVER_FLUSH_CACHE,		/* Flush an object from the server's cache */
  ROVER_UPDATE_CACHE,		/* Update an object in the server's cache */
  ROVER_INSERT_CACHE,		/* Insert an object into the server's cache */
  ROVER_LOAD_APP,		/* Load an application */
  ROVER_SET_NET,		/* Set network parameters */
  ROVER_SET_WATCH,		/* Set parameters to watch */
  ROVER_LOG_UPDATE,		/* Notification of log change */
  ROVER_NET_UPDATE,		/* Notification of network change */
  ROVER_APP_UPDATE,		/* Notification of application change */
  ROVER_HTTPCACHE_UPDATE,	/* Notification of HTTP cache change */
  ROVER_GET_PID,		/* Get application ID -> pid mapping */
  ROVER_SEND_SMTP,		/* Send request via SMTP */
  ROVER_LOGID,			/* Notification of export operation Log ID */
  ROVER_LOGSUFFIX,		/* Log suffix from Access Manager */
  ROVER_RESET_COMMAND,		/* Send a reset cmd on a session to a server */
  ROVER_SET_COMPRESS,		/* Set compression parameter */
  ROVER_SET_BATCH,		/* Set batching parameter */
  ROVER_COMPRESS_UPDATE,	/* Notification of compression change */
  ROVER_BATCH_UPDATE,		/* Notification of batching change */
  ROVER_STATUS,			/* Notification of QRPC status change */
} RoverType;

typedef enum _RoverRcvState {
  ROVER_IDLE = 0,		/* Idle state */
  ROVER_NEED_HDR,		/* Waiting for full header */
  ROVER_GOT_HDR,		/* Got full header */
  ROVER_NEED_DATA,		/* Waiting for full data */
  ROVER_GOT_DATA,		/* Got full data */
  ROVER_PROCESS_DIED		/* Remote process died */
} RoverRcvState;

typedef enum _RoverStatus {
  ROVER_NOTQUEUED = 0,		/* QRPC waiting to be sent to AM */
  ROVER_PENDING,		/* QRPC waiting to be sent to server */
  ROVER_SENT,			/* QRPC sent */
  ROVER_RESEND,			/* Send failed, to be resent */
  ROVER_SERVER_WAIT,		/* Reply failed, wait for server to resend */
  ROVER_ERROR,			/* QRPC failed */
} RoverStatus;

struct _RoverHdr {
  RoverType type;
  int appID;
  int msgID;
  int size;
};
typedef struct _RoverHdr RoverHdr;  

struct _RoverRcv {
  RoverRcvState RcvState;
  RoverHdr RcvHdr;
  int RcvCount;
  char *RcvPtr;
  char *RcvData;
};
typedef struct _RoverRcv RoverRcv;

struct _dependV {
  int ServerID;
  int ServerClock;
};
typedef struct _dependV dependV;

typedef struct _RoverObj RoverObj;
typedef struct _HTOper HTOper;

struct _HTOper {
  int session;			/* Object's session ID */
  char *url;			/* Object's url */
  int import;			/* Operation is an import */
  int size;			/* Operation's size */
  char *oper;			/* The operation */
  char *callback;		/* Operation's callback */
  int logID;			/* Log ID for exported operation */
  RoverStatus status;		/* Operation's status */
  HTOper *next;			/* Next in linked list */
  HTOper *prev;			/* Previous in linked list */
};

typedef enum _HTAllocType {
  Unused = 0,		/* Not in use */
  CodeOnly = 1,		/* Only code field is allocated (not data or log) */
  AllFields = 2		/* Code, data, and log fields are allocated */
} HTAllocType;

struct _RoverObj {
  int deleted;			/* Object deleted bit */
  HTAllocType inuse;		/* Structure in use field */
  RoverType type;		/* Type of object (typ. ordinary or error) */
  char *code;			/* Pointer to object's code */
  int csize;			/* Size of object's code */
  char *data;			/* Pointer to object's data */
  int dsize;			/* Size of object's data */
  char *tdata;			/* Pointer to object's tentative data */
  int tdsize;			/* Size of object's tentative data */
  char *log;			/* Pointer to object's log suffix */
  int lsize;			/* Size of object's log suffix */
  int session;			/* Session ID */
  char *url;			/* Relative url for object */
  HTParentAnchor *anchor;	/* Object's parent anchor */
  dependV *dv;			/* Dependency vector */
  int dvsize;			/* Dependency vector size */
  Tcl_Interp *interp;		/* Interpreter for object */
  HTOper *oper;			/* Linked list of pending operations */
  int operCount;		/* Size of pending operation list */
  time_t loadDelay;		/* Time to load object */
  time_t lastRef;		/* Time object was last referenced */
  int refCount;			/* Number of applications using object? */
  RoverCacheable cache;		/* Cacheable? flag */
  char *unmarshall;		/* Procedure name for unmarshall operation */
  char *name;			/* Tcl name for object */
};

typedef struct _QRPCState QRPCState;
struct _QRPCState {
  int execute;			/* Create new process flag */
  time_t loadTime;		/* Time request was loaded */
  int appID;			/* Application ID of requesting application */
  int toFD;			/* File descriptor of requesting application */
  int msgID;			/* Message ID of requesting application */
  int type;			/* Type of response */
  int session;			/* Object's session ID */
  char *url;			/* Object's URL */
  RoverObj *obj;		/* Pointer to object (optional) */
  char *marshallObj;		/* Pointer to marshalled object (optional) */
  int marshallSize;		/* Size of marshalled object (optional) */
  int argc;			/* Number of args for new process (optional) */
  char **argv;			/* Arguments for new process (optional) */
  int logID;			/* Request's log ID */
  char *dv;			/* Dependency vector (optional) */
  int dvsize;			/* Size of dependency vector (optional) */
  char *clientDv;		/* Client dependency vector (optional) */
  int clientDvsize;		/* Size of client dep vector (optional) */
  int IDsent;			/* Flag indicating log ID was sent to client */
  QRPCState *next;		/* Next state structure in chain (optional) */
};

/*
 *
 * WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING 
 *
 * DO NOT change the following structure without changing the Callback ID
 *   table and deleting the persistent cache. These IDs are stored in the
 *   persistent cache.
 *
 * WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING 
 *
 */
typedef enum _RoverCBList {
  NullCB = 0,		/* No callback */
  ImportObjCB,		/* RoverImportObj */
  ExportObjCB,		/* RoverExportObj */ 
  LoadAppCB,		/* RoverLoadApp */ 
  LocalInvCB,		/* RoverLocalInvocation */
  QRPCTimeCB,		/* RoverQRPCTime */
  ResetCB,		/* RoverReset */
  VerifyCB		/* RoverVerify */
} RoverCBList;

typedef BOOL (*HTCallback)(HTRequest *request, QRPCState *state);

#ifdef DEFINE_CBTABLE
PRIVATE BOOL RoverImportObj(HTRequest *req, QRPCState *state);
PRIVATE BOOL RoverExportObj(HTRequest *req, QRPCState *state);
PRIVATE BOOL RoverLoadApp(HTRequest *req, QRPCState *state);
PRIVATE BOOL RoverLocalInvocation(HTRequest *req, QRPCState *state);
PUBLIC BOOL RoverQRPCTime(HTRequest *req, QRPCState *state);
PUBLIC BOOL RoverReset(HTRequest *req, QRPCState *state);
PRIVATE BOOL RoverVerify(HTRequest *req, QRPCState *state);

PUBLIC HTCallback HTCallbackTable[] = {
  (HTCallback) NULL,
  RoverImportObj,
  RoverExportObj,
  RoverLoadApp,
  RoverLocalInvocation,
  RoverQRPCTime,
  RoverReset,
  RoverVerify
};
#endif /* DEFINE_CBTABLE */


typedef void (*RoverCallback)(int appID, int toFD, int cbID, RoverType type,
			      int size, char *buf, char *data);

typedef void (*HTAppInitializer)(int appID);

struct _state {
  int topipe;		/* Pipes for communication with Rover server */
  int frompipe;
  int ApplicationID;
  RoverQ *writeQ;
  RoverRcv Rstate;
  char *logWidget;
  char *appWidget;
  int watch;
};
typedef struct _state RoverState;

#define ROVER_WATCH_NONE	0
#define ROVER_WATCH_NET		(1 << 0)
#define ROVER_WATCH_LOG		(1 << 1)
#define ROVER_WATCH_APPS	(1 << 2)
#define ROVER_WATCH_CACHE	(1 << 3)
#define ROVER_WATCH_HTTP	(1 << 4)
#define ROVER_WATCH_BATCH	(1 << 5)
#define ROVER_WATCH_COMPRESS	(1 << 6)

#define ROVER_NO_CONSIST	(1 << 1)	/* No server consistency req */
#define ROVER_FULL_CONSIST	(1 << 2)	/* Server consistency req */
#define ROVER_ACCEPT_TENT	(1 << 3)	/* Accept tentative data */
#define ROVER_ACCEPT_PERM	(1 << 4)	/* Accept only permanent data*/


extern RoverQ **ServerWriteQ;
extern HTEventCallBack **Qhandlers;
extern CONST HTStreamClass HTApplication;
extern int *Pipe2App;
extern int HTAppCount;

extern void RoverAppPipeFlush();
extern void RoverAppCloseAll();
extern void RoverAppDoUpdate(int flag, RoverUpdateProc update);
extern int AppDoAbort(HTStream *me);
extern void RoverServerRcvCheck();
extern void RoverServerRcvWait();
extern void RoverServerRcvProc(int wait);
extern int App_LocalExec(char *url, RoverCBList callback, QRPCState *state);

#ifdef ROVER_ACCESSMGR
extern time_t RoverIncarnation;
extern int RoverArgc;
extern char **RoverArgv;
extern FILE *RoverLogFile;
extern int ResetWaitCount;

extern int RoverAMAppInit();
extern char *RoverAppGetURL(int appID);
extern int RoverAppGetPID(int appID);
extern RoverErr RoverAppSetWatch(int appID, int watch);
extern void RoverUpdateApps(int appID, int toPipe, int watch);
extern RoverErr RoverAppRemoteCreate(char *url, int sock, int pid,
				     char *hostname, char *username);
extern int RoverAppPipe2AppID(int toPipe);

#endif /* ROVER_ACCESSMGR */
#endif /* defined(ROVER_CLIENT) || defined(ROVER_ACCESSMGR) */
#endif /* HTAPP_H */
