#if 0
  Should schedule files for periodic stat to make sure they are consistent

#endif


 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: RoverCache.c
 *
 * Description: Implements persistent cache/log
 *
 * Last Modified:  14-Jun-95 (adj)
 *
 * Global Functions:
 *    void set_prof_(pc, profileInfo *, const char *)
 *        Record address of profileInfo for procedure at pc.
 *        Called only from InitProfile_
 *
 *
 * Global Variables: none
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define ROVER_ACCESSMGR

#include "RoverBase.h"
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include "tcl.h"
#include "RoverFile.h"
#include "RoverTypes.h"
#include "RoverErr.h"
#include "HTCache.h"


PUBLIC char *RoverCacheDir;
PUBLIC int RoverCacheMemLimit = 2048*1024;

struct _CacheEntry {
  int valid;			/* Valid bit (0 if req', but !initialized) */
  int dirty;			/* Save object (and meta data) to disk */
  char *url;			/* URL of cached object */
  RoverObjType type;		/* Type of cached object */
  RoverCacheable cache;		/* Cacheable? flag */
  unsigned long size;		/* Object's size */
  time_t objTime;		/* Load time for object */
  time_t loadDelay;		/* Time to load object */
  time_t lastRef;		/* Time object was last referenced */
  char imageName[256];		/* Name of cache file for object */
  char *data;			/* Object data */
  char *dv;			/* Object's dependency vector */
  unsigned long dvsize;		/* Object's dv size */
};
typedef struct _CacheEntry CacheEntry;

PRIVATE Tcl_HashTable CacheHash;
PRIVATE char *CacheInfo;
PRIVATE char *CacheInfo2;
PRIVATE int CacheDirty = 0;
PRIVATE int CacheSize = 0;
PRIVATE int CacheInitMem = 0;
PRIVATE int CacheMemSize = 0;
PRIVATE CacheEntry **CacheSortList = NULL;
PRIVATE int CacheSortListSize = 0;
PRIVATE int CacheSortListMax = 0;
PRIVATE char TempPtr[1024];
PRIVATE char writebuf[2048];

PRIVATE int CacheEvictCheck(int size);
PRIVATE CacheEntry *CacheSearchInfo(char *url, int session);
PRIVATE CacheEntry *CacheInsert(char *url, char *name, RoverObjType type,
				RoverCacheable cache, char *data,
				unsigned long size, time_t objTime,
				time_t delay, time_t last, char *dv,
				unsigned long dvsize);

PRIVATE void CacheInsertSortList(CacheEntry *centry)
{
  if ((CacheSortListSize + 1) >= CacheSortListMax) {
    CacheSortListMax <<= 1;
    CacheSortList=(CacheEntry **)realloc(CacheSortList, 
					 CacheSortListMax*sizeof(CacheEntry*));
  }
  CacheSortList[CacheSortListSize++] = centry;
}

PRIVATE void CacheDeleteSortList(CacheEntry *centry)
{
  int i;
 
  for (i = 0; i < CacheSortListSize; i++) {
    if (CacheSortList[i] == centry) {
      CacheSortList[i] = CacheSortList[--CacheSortListSize];
      return;
    }
  }
  fprintf(stderr, "WARNING! Cache sort list delete failed: %s\n", centry->url);
}

PUBLIC int RoverCacheInit()
{
  struct stat sbuf;
  unsigned long size, cache, dvsize, version, type;
  time_t objTime, loadDelay, lastRef;
  char Name[512], Url[1024], tmp[32];
  int rc;
  char *dv;
  FILE *file;
  CacheEntry *centry;

  Tcl_InitHashTable(&CacheHash, TCL_STRING_KEYS);

  CacheSortListMax = 512;
  if (!(CacheSortList =
	(CacheEntry **)malloc(CacheSortListMax*sizeof(CacheEntry *))))
    return(ROVER_NO_MEM);

  fprintf(stderr, "Initializing persistent cache\n");
  /* Rebuild the in-memory cache here */
  if (!(RoverCacheDir = (char *)malloc(strlen(ROVER_CACHE_DIRECTORY)+
				       strlen(RoverHome)+2)))
    return(ROVER_NO_MEM);
  sprintf(RoverCacheDir, "%s/%s", RoverHome, ROVER_CACHE_DIRECTORY);
  if (stat(RoverCacheDir, &sbuf)) {
    fprintf(stderr, "Creating Rover cache directory (%s)\n", RoverCacheDir);
    if ((rc = mkdir(RoverCacheDir, 0777)) != 0) {
      fprintf(stderr,
	      "Cache dir create err: %s (%d) [RoverCacheInit:RoverCache.c]\n",
	      RoverCacheDir, rc);
      return(ROVER_ERR);
    }
  } else if (!S_ISDIR(sbuf.st_mode)) {
    fprintf(stderr,
	    "Cache dir file isn't a dir [RoverCacheInit:RoverCache.c]\n");
    return(ROVER_ERR);
  }
  
  /* Allocate primary cache information file */
  if (!(CacheInfo = (char *)malloc(strlen(ROVER_CACHE_INFO)+
					strlen(RoverCacheDir)+2)))
    return(ROVER_NO_MEM);
  sprintf(CacheInfo, "%s/%s", RoverCacheDir, ROVER_CACHE_INFO);

  /* Allocate backup cache information file */
  if (!(CacheInfo2 = (char *)malloc(strlen(ROVER_CACHE_INFO)+
				    strlen(RoverCacheDir)+3)))
    return(ROVER_NO_MEM);
  sprintf(CacheInfo2, "%s/%s2", RoverCacheDir, ROVER_CACHE_INFO);

  /* Now read the cache file */
  if (stat(CacheInfo, &sbuf)) {
    /* No info file, flush directory */
  } else {
    /* 1. Read the info from the file
       2. Flush any files from the directory that aren't in the info file
     */
    
    if ((file = fopen(CacheInfo, "r")) == NULL) {
      fprintf(stderr, "Rover cache file (%s) open failed: %s\n",
	      CacheInfo, strerror(errno));
      return(ROVER_ERR);
    }
  
    /* Read the cache version info */
    if (!feof(file)) {
      rc = fscanf(file, "ROVER_CACHE_VERSION %lu\n", &version);
      if (rc != 1) {
	fprintf(stderr,
		"Cache version id read fail [RoverCacheInit:RoverCache.c]\n");
	return(ROVER_ERR);
      }
    }
    if (version != (unsigned long) ROVER_CACHE_VERSION) {
      fprintf(stderr,
	      "Cache version %lu, expected %lu [RoverCacheInit:RoverCache.c]\n",
	      version, (unsigned long) ROVER_CACHE_VERSION);
      /* We can read version 2 files */
      if (version != 2) return(ROVER_ERR);
      fprintf(stderr,
	      "Updating cache to version %lu [RoverCacheInit:RoverCache.c]\n",
	      (unsigned long) ROVER_CACHE_VERSION);
    }

    /* Now, read the directory */
    while (1) {
      if (feof(file)) break;
      rc = fscanf(file, "%lu %lu %lu %lu %lu %lu %lu %s %s ",
		  &type, &cache, &size, &dvsize, &objTime, &loadDelay,
		  &lastRef, Name, Url);
      if (rc != 9) {
	if (!rc || feof(file)) break;
	fprintf(stderr,
		"Read bad cache line %d/9 items [RoverCacheInit:RoverCache.c]\n",
		rc);
	continue;
      }
      if (RoverDebug)
	fprintf(stderr,"Read info for URL: %s [RoverCacheInit:RoverCache.c]\n",
		Url);
      if (dvsize) {
	if ((dv = (char *)malloc(dvsize)) == NULL) return(ROVER_NO_MEM);
	rc = fread(dv, sizeof(char), dvsize, file);
	if (rc != dvsize) {
	  fprintf(stderr,
		  "Bad cache dv read size %d [RoverCacheInit:RoverCache.c]\n",
		  rc);
	  free(dv);	/* Blow away the storage */
	  continue;
	}
      } else {
	dv = NULL;
      }
      if (dvsize && (version == 3)) {
	/* Read the \n after the dependency vector */
	rc = fread(tmp, sizeof(char), 1, file);
	if (rc != 1) {
	  fprintf(stderr,
		  "Bad cache \\n read size %d [RoverCacheInit:RoverCache.c]\n",
		  rc);
	  free(dv);	/* Blow away the storage */
	  continue;
	}
      }
      if (stat(Name, &sbuf)) {	/* Opps! We found a non-existent entry */
	fprintf(stderr,"Data file %s for cache entry %s not found\n",Url,Name);
	if (dvsize) free(dv);	/* Blow away the storage */
	continue;
      }
      centry = CacheInsert(Url, Name, (RoverObjType) type,
			   (RoverCacheable) cache, NULL, size, objTime,
			   loadDelay, lastRef, dv, dvsize);
      if (!centry)
	fprintf(stderr,
		"Cache insert %s failed [RoverCacheInit:RoverCache.c]\n", Url);
      else
	centry->dirty = 0;
    }
    fclose(file);
  }
  CacheInitMem = 1;
  fprintf(stderr, "Rover in-memory cache initialized\n");
  RoverCacheSchedStat();/* Schedule ROVER_VERIFY & ROVER_BEST_EFFORT entries */
  return(ROVER_OK);
}

PRIVATE RoverErr CacheSaveEntry(CacheEntry *centry)
{
  int rc;
  FILE *file;

  if (!centry->valid || (centry->type == ROVER_BINARY) ||
      (centry->type == ROVER_BINARY_UPDATE))
    return(ROVER_OK);
  if (centry->imageName[0] == '\0') {
    sprintf(centry->imageName, "%s/RoverObjXXXXXXXX", RoverCacheDir);
    if ((file = RoverGetUniqFile(centry->imageName)) == NULL) {
      fprintf(stderr,
	      "Cache obj %s open error: %s [CacheSaveEntry:RoverCache.c]\n",
	      centry->imageName, strerror(errno));
      return(ROVER_ERR);
    }
  } else {
    if ((file = fopen(centry->imageName, "w")) == NULL) {
      fprintf(stderr,
	      "Error opening cache obj %s: %s [CacheSaveEntry:RoverCache.c]\n",
	      centry->imageName, strerror(errno));
      return(ROVER_ERR);
    }
  }
  rc = fwrite(centry->data, sizeof(char), centry->size, file);
  if (rc != centry->size) {
    fprintf(stderr,
	    "Cache obj %s write err %d != %ld: %s [CacheSaveEntry:RoverCache.c]\n",
	    centry->imageName, rc, centry->size, strerror(errno));
    fclose(file);
    return(ROVER_ERR);
  }
  fclose(file);
  centry->dirty = 0;
  return(ROVER_OK);
}
    
PUBLIC RoverErr RoverCacheSave()
{
  int rc, size;
  Tcl_HashSearch search;
  Tcl_HashEntry *hentry;
  CacheEntry *centry;
  RoverErr rv;
  FILE *file;

  if (!CacheInitMem) {
    fprintf(stderr, "Rover in-memory cache NOT initialized, not saving\n");
    return(ROVER_OK);
  }

  /* Only save the cache if it's changed */
  if (!CacheDirty) return(ROVER_OK);

  /* Save the in-memory cache */
  if (!(file = fopen(CacheInfo2, "w"))) {
    fprintf(stderr, "Rover cache file (%s) open failed: %s\n", CacheInfo2,
	    strerror(errno));
    CacheDirty = 0;	/* Avoid recursive aborts */
    return(ROVER_ERR);
  }

  if (RoverDebug) {
    fprintf(stderr, "Saving in-core cache (%d bytes) to %s\n", CacheMemSize,
	    CacheInfo);
    fprintf(stderr, "In-core cache size %d bytes\n", CacheMemSize);
  }
  fprintf(file, "ROVER_CACHE_VERSION %d\n", ROVER_CACHE_VERSION);

  for (hentry = Tcl_FirstHashEntry(&CacheHash, &search); hentry != NULL;
       hentry = Tcl_NextHashEntry(&search)) {
    centry = (CacheEntry *) Tcl_GetHashValue(hentry);
    if (!centry || !centry->valid || !centry->size ||
	(centry->cache == ROVER_NO_CACHE))
      continue;

    if (RoverDebug) fprintf(stderr,"Saving cache entry for %s\n", centry->url);
    if (centry->dirty) {
      rv = CacheSaveEntry(centry); /* Save object data to a file */
      if (rv != ROVER_OK) continue;
    }
    sprintf(writebuf, "%u %u %lu %lu %lu %lu %lu %s %s ",
	    centry->type, centry->cache, centry->size, centry->dvsize,
	    centry->objTime, centry->loadDelay, centry->lastRef,
	    centry->imageName, (char *) Tcl_GetHashKey(&CacheHash, hentry));
    size = strlen(writebuf);
    ASSERT((size+1)<=2048);
    rc = fwrite(writebuf, sizeof(char), size, file);
    if (rc != size) {
      fprintf(stderr, 
	      "Cache %s write err %d != %d: %s [1] [RoverCacheSave:RoverCache.c]\n",
	      (char *) Tcl_GetHashKey(&CacheHash, hentry), rc, size,
	      strerror(errno));
      fclose(file);
      unlink(CacheInfo2);
      return(ROVER_ERR);
    }
    rc = fwrite(centry->dv, sizeof(char), centry->dvsize, file);
    if (rc != centry->dvsize) {
      fprintf(stderr, 
	      "Cache %s dv write err %d != %ld: %s [2] [RoverCacheSave:RoverCache.c]\n",
	      (char *) Tcl_GetHashKey(&CacheHash, hentry), rc, centry->dvsize,
	      strerror(errno));
      fclose(file);
      unlink(CacheInfo2);
      return(ROVER_ERR);
    }
    /* Version 3 files have a \n following the dependency information */
    fputs("\n", file);
  }
  fclose(file);
  
  /* Replace the cache file with the backup file */
  if ((rc = rename(CacheInfo2, CacheInfo)) != 0) return(ROVER_ERR);
  CacheDirty = 0;
  return(ROVER_OK);
}


PRIVATE CacheEntry *CacheInsert(char *url, char *name, RoverObjType type,
				RoverCacheable cache, char *data,
				unsigned long size, time_t objTime,
				time_t delay, time_t last, char *dv,
				unsigned long dvsize)
{
  Tcl_HashEntry *hentry;
  CacheEntry *centry = NULL;
  int new, rc;
  
  url = strdup(url);
  hentry = Tcl_CreateHashEntry(&CacheHash, url, &new);
  if (!new) {				/* Check for duplicate urls */
    if (RoverDebug)
      fprintf(stderr, "Updating url: %s [CacheInsert:RoverCache.c]\n", url);
    centry = (CacheEntry *) Tcl_GetHashValue(hentry);

    /* Delete the old image, if one exists */
    if (*centry->imageName && ((rc = unlink(centry->imageName)) != 0)) {
      fprintf(stderr,
	      "Updated cache entry delete failed %s [CacheInsert:RoverCache.c]\n",
	      strerror(errno));
      return(NULL);
    }

    /* Delete old data, if it exists */
    if (centry->dvsize) free(centry->dv);
    if (centry->url) free(centry->url);
    if ((centry->type != ROVER_BINARY)&&(centry->type != ROVER_BINARY_UPDATE)){
      if (centry->data) {
	free(centry->data);
	CacheMemSize -= centry->size;
      }
    }
  } else {
    if (RoverDebug)
      fprintf(stderr, "New url: %s [CacheInsert:RoverCache.c]\n", url);
    CacheSize++;
    if (!(centry = (CacheEntry *)malloc(sizeof(CacheEntry)))) {
      fprintf(stderr, "Out of memory [CacheInsert:RoverCache.c]\n");
      return(NULL);
    }

    if ((type != ROVER_BINARY) && (type != ROVER_BINARY_UPDATE))
      CacheInsertSortList(centry);
  }

  if ((type != ROVER_BINARY) && (type != ROVER_BINARY_UPDATE)) {
    CacheEvictCheck(size);
    CacheMemSize += size;
  }
  centry->dirty = centry->valid = 1;
  centry->url = url;
  centry->type = type;
  centry->cache = cache;
  centry->size = size;
  centry->loadDelay = delay;
  centry->objTime = objTime;
  centry->lastRef = last;
  if (name) 
    strcpy(centry->imageName, name);
  else 
      centry->imageName[0] = '\0';
  centry->data = data;
  if (dv) {
    if (!(centry->dv = malloc(dvsize)))  {
      fprintf(stderr, "Out of memory [CacheInsert:RoverCache.c]\n");
      return(NULL);
    }
    memcpy(centry->dv, dv, dvsize);
    centry->dvsize = dvsize;
  } else {
    centry->dv = NULL;
    centry->dvsize = 0;
    if (dvsize) {
      fprintf(stderr, "%s: dv = NULL but dvsize = %lu\n", myAppID, dvsize);
      return(NULL);
    }
  }
  if (RoverDebug)
    fprintf(stderr, "Url: %s has dv `%s', size %ld\n", url, dv, dvsize);
  
  /* Save cache entry */
  Tcl_SetHashValue(hentry, (char *) centry);
  CacheDirty = 1;
  return(centry);
}

PUBLIC RoverErr RoverCacheInsertData(char *url, int session,
				     RoverObjType type, RoverCacheable cache,
				     char *data, unsigned long size,
				     time_t objTime, time_t delay,
				     time_t last, char *dv,
				     unsigned long dvsize)
{
  sprintf(TempPtr, "%d::%s", session, url);
  ASSERT((strlen(TempPtr)+1)<=1024);
  if (RoverDebug) fprintf(stderr, "Time for %s is %lu\n", TempPtr, objTime);

  if (CacheInsert(TempPtr, NULL, type, cache, data, size, objTime, delay,
		       last, dv, dvsize))
    return(ROVER_OK);
  else
    return(ROVER_ERR);
}

PRIVATE CacheEntry *CacheSearchInfo(char *url, int session)
{
  Tcl_HashEntry *hentry;

  sprintf(TempPtr, "%d::%s", session, url);
  ASSERT((strlen(TempPtr)+1)<=1024);
  if (!(hentry = Tcl_FindHashEntry(&CacheHash, TempPtr))) return(NULL);
  return((CacheEntry *) Tcl_GetHashValue(hentry));
}

PUBLIC RoverErr RoverCacheRef(char *url, int session, time_t last)
{
  CacheEntry *centry;

  if (!(centry = CacheSearchInfo(url, session))) return(ROVER_ERR);
  centry->lastRef = last;
  CacheDirty = 1;
  return(ROVER_OK);
}

PUBLIC RoverErr RoverCacheUpdateName(char *url, int session, char *name)
{
  CacheEntry *centry;

  if (!(centry = CacheSearchInfo(url, session))) return(ROVER_ERR);
  strcpy(centry->imageName, name);
  CacheDirty = 1;
  return(ROVER_OK);
}

PUBLIC RoverErr RoverCacheUpdate(char *url, int session, RoverObjType type,
				 RoverCacheable cache, char *name,
				 time_t objTime, time_t delay, time_t last,
				 int size, char *data, char *dv,
				 unsigned long dvsize)
{
  CacheEntry *centry;

  if (!(centry = CacheSearchInfo(url, session))) return(ROVER_ERR);
  centry->type = type;
  centry->cache = cache;
  if (data && (type != ROVER_BINARY) && (type != ROVER_BINARY_UPDATE)) {
    if (centry->data) {
      free(centry->data);
      CacheMemSize -= centry->size;
    }
    CacheEvictCheck(size);
    CacheMemSize += size;
    centry->data = data;
  }
  centry->size = size;
  if (delay) centry->loadDelay = delay;
  if (objTime) centry->objTime = objTime;
  if (last) centry->lastRef = last;
  if (name) strcpy(centry->imageName, name);
  if (centry->dvsize) free(centry->dv);
  if (dv) {
    if ((centry->dv = malloc(dvsize)) == NULL) {
      fprintf(stderr, "Out of memory! [RoverCacheUpdate:RoverCache.c]\n");
      return(ROVER_NO_MEM);
    }
    memcpy(centry->dv, dv, dvsize);
    centry->dvsize = dvsize;
  } else {
    centry->dv = NULL;
    centry->dvsize = 0;
    if (dvsize) {
      fprintf(stderr, "%s: dv = NULL but dvsize = %lu\n", myAppID, dvsize);
      return(ROVER_ERR);
    }
  }
  centry->valid = centry->dirty = CacheDirty = 1;
  return(ROVER_OK);
}  

PUBLIC RoverErr RoverCacheUpdateData(char *url, int session, char *data,
				     int size, char *dv, unsigned long dvsize)
{
  CacheEntry *centry;

  if (!(centry = CacheSearchInfo(url, session))) return(ROVER_ERR);
  return(RoverCacheUpdate(url, session, centry->type, centry->cache, NULL,
			  (time_t) 0, (time_t) 0, (time_t) 0, size, data, dv,
			  dvsize));
}

PUBLIC RoverErr RoverCacheUpdateDependV(char *url, int session, char *dv,
					unsigned long dvsize)
{
  CacheEntry *centry;

  if (!(centry = CacheSearchInfo(url, session))) return(ROVER_ERR);
  return(RoverCacheUpdate(url, session, centry->type, centry->cache, NULL,
			  (time_t) 0, (time_t) 0, (time_t) 0, centry->size, 
			  NULL, dv, dvsize));
}

PUBLIC char *RoverCacheSearchName(char *url, int session, RoverCacheable *cache,
				  RoverObjType *type, char **dv,
				  unsigned long *dvsize)
{
  CacheEntry *centry;

  if (!(centry = CacheSearchInfo(url, session))) return(NULL);
  time(&(centry->lastRef));
  *cache = centry->cache;
  *type = centry->type;
  *dv = centry->dv;
  *dvsize = centry->dvsize;
  CacheDirty = 1;
  return(centry->imageName);
}

PRIVATE RoverErr CacheLoadData(CacheEntry *centry)
{
  int rc;
  FILE *file;

  if (RoverDebug)
    fprintf(stderr, "Loading %s [CacheLoadData:RoverCache.c]\n", centry->url);
  CacheEvictCheck(centry->size);
  CacheMemSize += centry->size;
  if ((centry->data = malloc(centry->size)) == NULL) {
    fprintf(stderr, "Out of memory [CacheLoadData:RoverCache.c]\n");
    return(ROVER_NO_MEM);
  }
  if ((file = fopen(centry->imageName, "r")) == NULL) {
    fprintf(stderr, "Cache file open error %s [CacheLoadData:RoverCache.c]\n",
	    centry->imageName);
    return(ROVER_ERR);
  }
  rc = fread(centry->data, sizeof(char), centry->size, file);
  if (rc != centry->size) {
    fprintf(stderr,
	    "Cache read err %s: %d; not %ld [CacheLoadData:RoverCache.c]\n",
	    centry->imageName, rc, centry->size);
    return(ROVER_ERR);
  }
  fclose(file);
  return(ROVER_OK);
}

PUBLIC char *RoverCacheSearchData(char *url, int session, int *size)
{
  CacheEntry *centry;
  RoverErr rc;

  if (!(centry = CacheSearchInfo(url, session))) {
    *size = 0;
    return(NULL);
  }

  *size = centry->size;
  time(&(centry->lastRef));
  if (!centry->data) {
    rc = CacheLoadData(centry);
    if (rc != ROVER_OK) return(NULL);
  }
  CacheDirty = 1;
  return(centry->data);
}

PUBLIC RoverErr RoverCacheFlush(char *url, int session)
{
  Tcl_HashEntry *hentry;
  CacheEntry *centry;
  int rc;
  RoverErr rv = ROVER_OK;

  sprintf(TempPtr, "%d::%s", session, url);
  ASSERT((strlen(TempPtr)+1)<=1024);
  if (RoverDebug) fprintf(stderr, "Flushing %s [RoverCacheFlush:HTCache.c]\n",
			  TempPtr);

  if (!(hentry = Tcl_FindHashEntry(&CacheHash, TempPtr))) return(ROVER_OK);

  centry = (CacheEntry *) Tcl_GetHashValue(hentry);
  if (centry->dvsize) free(centry->dv);
  if ((centry->type != ROVER_BINARY)&&(centry->type != ROVER_BINARY_UPDATE)){
    if (centry->data) {
      free(centry->data);
      CacheMemSize -= centry->size;
    }
    CacheDeleteSortList(centry);
  }
  if (*centry->imageName && ((rc = unlink(centry->imageName)) != 0)) {
    fprintf(stderr,
	    "Cache file %s delete failed [RoverCacheFlush:RoverCache.c]\n",
	    centry->imageName);
    rv = ROVER_ERR;
  }
  free(centry->url);
  free(centry);
  Tcl_DeleteHashEntry(hentry);
  CacheSize--;
  CacheDirty = 1;
  return(rv);
}

PUBLIC RoverErr RoverCacheRequest(char *url, int session,
				    RoverObjType type)
{
  CacheEntry *centry;

  sprintf(TempPtr, "%d::%s", session, url);
  ASSERT((strlen(TempPtr)+1)<=1024);
  centry = CacheInsert(TempPtr, NULL, type, (RoverCacheable) 0, NULL, 0L,
		       (time_t) 0, (time_t) 0, (time_t) 0, NULL, 0L);
  if (!centry) return(ROVER_ERR);
  
  centry->dirty = 0;	/* Mark clean so that it doesn't get written to disk */
  centry->valid = 0;	/* Not yet valid */
  return(ROVER_OK);
}

PUBLIC char *RoverCacheSearchByType(RoverObjType type, int *size,
				    time_t *objTime, time_t *lastRef,
				    Tcl_HashSearch **search)
{
  Tcl_HashEntry *hentry;
  CacheEntry *centry;
  RoverErr rc;

  *size = 0;
  *objTime = *lastRef = (time_t) 0;
  if (!*search) {
    if (!(*search = (Tcl_HashSearch *)malloc(sizeof(Tcl_HashSearch)))) {
      fprintf(stderr,"Out of memory! [RoverCacheSearchByType:RoverCache.c]\n");
      return(NULL);
    }
    hentry = Tcl_FirstHashEntry(&CacheHash, *search); 
  } else 
    hentry = Tcl_NextHashEntry(*search);

  for (; hentry != NULL; hentry = Tcl_NextHashEntry(*search)) {
    centry = (CacheEntry *) Tcl_GetHashValue(hentry);
    if (!centry || !centry->valid || !centry->size ||
	(centry->cache == ROVER_NO_CACHE) || (centry->type != type))
      continue;
    time(&(centry->lastRef));
    if (!centry->data) {
      rc = CacheLoadData(centry);
      if (rc != ROVER_OK) {
	free(*search);
	*search = NULL;
	return(NULL);
      }
    }
    CacheDirty = 1;
    *size = centry->size;
    *objTime = centry->objTime;
    *lastRef = centry->lastRef;
    return(centry->data);
  }
  free(*search);
  *search = NULL;
  return(NULL);
}



PRIVATE int CacheComp(const void *a, const void *b)
{
  CacheEntry *c1 = *((CacheEntry **) a);
  CacheEntry *c2 = *((CacheEntry **) b);

  return(c1->lastRef - c2->lastRef);
}

PRIVATE int CacheEvictCheck(int size)
{
  CacheEntry *centry;
  char *ref;
  int i, target;
  RoverErr rv;

  if (RoverDebug)
    fprintf(stderr, "Evict check %d+%d <? %d == %s\n", size, CacheMemSize,
	    RoverCacheMemLimit,
	    (((size + CacheMemSize) < RoverCacheMemLimit) ? "YES" :"NO"));
  if ((size + CacheMemSize) < RoverCacheMemLimit) return(0);
  if (size > CacheMemSize) {
    fprintf(stderr,
	    "%s: WARNING! Resetting cache memory limit to %d\n", myAppID,
	    size);
    RoverCacheMemLimit = size;
  }

  target = RoverCacheMemLimit - size;
  if (RoverDebug) 
    for (i = 0; i < CacheMemSize; i++) {
      centry = CacheSortList[i];
      if (!centry->valid || !centry->data || !centry->size) continue;
      ref = ctime(&(centry->lastRef));
      ref[strlen(ref)-1] = '\0';
      fprintf(stderr,
	      "Url: %s has size %lu, Last ref %s (%u), Load time %u\n",
	      centry->url, centry->size, ref, (unsigned int) centry->lastRef,
	      (unsigned int)centry->loadDelay);
    }

  qsort(CacheSortList, CacheSortListSize, sizeof(CacheEntry *), CacheComp);
  for (i = 0; (i < CacheSortListSize) && (CacheMemSize > target); i++) {
    centry = CacheSortList[i];
    if (!centry->data || !centry->size) continue;
    if (RoverDebug) fprintf(stderr,"Evict Url: %s w/ size %lu, Last ref %u\n",
			    centry->url, centry->size,
			    (unsigned int) centry->lastRef);
    if (centry->dirty) {
      rv = CacheSaveEntry(centry); /* Save object data to a file */
      if (rv != ROVER_OK) continue;
    }
    free(centry->data);
    centry->data = NULL;
    CacheMemSize -= centry->size;
  }
  return (0);
}

PUBLIC RoverErr RoverCacheSchedStat()
{
  return(ROVER_OK);
#if 0
  int count = 0;
  Tcl_HashSearch search;
  Tcl_HashEntry *hentry;
  CacheEntry *centry;
  RoverObjStat *list;
  char buf[32], *name;


  if (!(list = (RoverObjStat *)malloc(CacheSize*sizeof(RoverObjStat))))
    return(ROVER_NO_MEM);

  /* Add session number to the stat list !!! */

  if (RoverDebug)
    fprintf(stderr, "Ready to stat cache\n");
  for (hentry = Tcl_FirstHashEntry(&CacheHash, &search); hentry != NULL;
       hentry = Tcl_NextHashEntry(&search)) {
    centry = (CacheEntry *) Tcl_GetHashValue(hentry);
    if (centry && centry->size && (centry->cache != ROVER_NO_CACHE)) {
      sprintf(buf, "%lu ", centry->objTime);
      if (!(list[count].time = malloc(strlen(buf)+1))) return(ROVER_NO_MEM);
      strcpy(list[count].time, buf);
      name = (char *) Tcl_GetHashKey(&CacheHash, hentry);
      if (!(list[count].name = strdup(name))) return(ROVER_NO_MEM);
      count++;
    }
  }
  return(ROVER_OK);
#if 0
  i = RoverSessionStat(CacheSize, list);
  >>> Merge changes here
#endif
#endif
}

