 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTNet.c
 *
 * Description: Handles server-side object method invocations
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define ROVER_ACCESSMGR

/*
 * Parts of this code are based upon ifconfig.c
 *
 * Version:	@(#)ifconfig.c	2.30	10/07/93
 *
 * Author:	Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *		Copyright 1993 MicroWalt Corporation
 *
 *		This program is free software; you can redistribute it
 *		and/or  modify it under  the terms of  the GNU General
 *		Public  License as  published  by  the  Free  Software
 *		Foundation;  either  version 2 of the License, or  (at
 *		your option) any later version.
 *
 */

#include <sys/types.h>
#ifdef linux
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#endif /* linux */
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "RoverBase.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTNet.h"
#include "RoverTCP.h"

#ifdef linux
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#include <linux/ipx.h>
#endif /* linux */

#include <tcl.h>

PUBLIC HTNetState RoverNetState;
PUBLIC int NetChanged = 0;

PRIVATE RoverErr HTNetUpdate();

PUBLIC void HTNetStatus()
{
  HTNetTypes n = RoverNetState.NetAvail;
  RoverErr rv = HTNetUpdate();

  if (rv != ROVER_OK) return;
  NetChanged = 0;
  if (n != RoverNetState.NetAvail) NetChanged = 1;

  if (NetChanged) {
    int metric = 0;
    HTNetTypes type = NONE;
    char *name = HTNetFastInterface(&metric, &type);
    fprintf(stderr, "Fastest interface is %s (type: %s metric %d)\n",
	    name,
	    ((type == NONE) ? "NONE" :
	     ((type == LOCALHOST) ? "LOCALHOST" :
	      ((type == IP) ? "IP" : "<unknown>"))),
	    metric);
  }
}

PUBLIC char *HTNetFastInterface(int *metric, HTNetTypes  *type)
{
  int i, j = -1;

  *type = NONE;
  for (*metric = i = 0; i < RoverNetState.numInterfaces; i++) {
/*    fprintf(stderr, "name is %s\n", RoverNetState.interfaces[i].name); */
    /* Prefer higher type or metric */
    if ((RoverNetState.interfaces[i].type > *type) ||
	(RoverNetState.interfaces[i].metric > *metric)) {
      j = i;
      *metric = RoverNetState.interfaces[i].metric;
      *type = RoverNetState.interfaces[i].type;
    }
  }
  if (j == -1) return(NULL);
  return(RoverNetState.interfaces[j].name);
}


#ifndef linux
PRIVATE RoverErr HTNetUpdate()
{
  if (RoverDebug) fprintf(stderr, "Checking network status\n");
  if (RoverNetState.numInterfaces && RoverNetState.interfaces)
    return;
/*    free(RoverNetState.interfaces); */
    
  RoverNetState.numInterfaces = 2;
  RoverNetState.NetAvail = IP;
  if ((RoverNetState.interfaces =
       (HTNetIfs *)malloc(2*sizeof(HTNetIfs))) == NULL) {
    fprintf(stderr, "Out of memory! [HTNetUpdate:HTNet.c]\n");
    return(ROVER_NO_MEM);
  }

  strcpy(RoverNetState.interfaces[0].ifName, "lo0");
  RoverNetState.interfaces[0].name = "localhost";
  RoverNetState.interfaces[0].type = LOCALHOST;
  RoverNetState.interfaces[0].metric = 694;	/* 10Mb vs. 14.4Kb */

  strcpy(RoverNetState.interfaces[1].ifName, "le0");
  RoverNetState.interfaces[1].name = RoverHostName;
  RoverNetState.interfaces[1].type = IP;
  RoverNetState.NetAvail = IP;
  RoverNetState.interfaces[1].metric = 694;	/* 10Mb vs. 14.4Kb */
  return(ROVER_OK);
}

PUBLIC void HTNetShowInterfaces()
{
  return;
}
#else /* linux */

struct _HTNetInterface {
  char			ifName[IFNAMSIZ];	/* interface name	 */
  short			type;			/* if type		 */
  short			flags;			/* various flags	 */
  int			metric;			/* routing metric	 */
  int			mtu;			/* MTU value		 */
  struct ifmap		map;			/* hardware setup	 */
  struct sockaddr	addr;			/* IP address		 */
  struct sockaddr	dstaddr;		/* P-P IP address	 */
  struct sockaddr	broadaddr;		/* IP broadcast address	 */
  struct sockaddr	netmask;		/* IP network mask	 */
  struct sockaddr	ddpaddr;		/* Appletalk DDP address */
  int			has_ddp;
  char			hwaddr[32];		/* HW address		 */
  char			*name;			/* Host name 		 */
};
typedef struct _HTNetInterface HTNetInterface;

int skfd = -1;				/* generic raw socket desc.	*/
int ipx_sock = -1;			/* IPX socket			*/
int ax25_sock = -1;			/* AX.25 socket			*/
int inet_sock = -1;			/* INET socket			*/
int ddp_sock = -1;			/* Appletalk DDP socket		*/
int HostsInit = 0;
Tcl_HashTable HostsCache;

PRIVATE char *HTNetAddrToName(struct sockaddr *addr)
{
  FILE *file;
  int new;
  char line[1024], *ptr, *end, *centry = NULL, *key, *buf;
  Tcl_HashEntry *hentry;
  struct sockaddr_in *sin = (struct sockaddr_in *)addr;

  buf = inet_ntoa(sin->sin_addr);
  if (!HostsInit) {
    /* Scan /etc/hosts into a hash table */

    HostsInit = 1;
    Tcl_InitHashTable(&HostsCache, TCL_STRING_KEYS);
    if (RoverDebug) fprintf(stderr, "Scanning /etc/hosts for hostnames\n");

    file = fopen("/etc/hosts","r");
    if (file == NULL) {
      fprintf(stderr, "WARNING! Open of /etc/hosts failed\n");
      return(buf);
    }
    while (1) {
      line[0] = '\0';
      fgets(line, 1023, file);
      if (!line[0] && feof(file)) break;
      if ((line[0] == '\n') || (line[0] == '#')) /* Discard blanks/comments */
	continue;
      key = line;
      while (isspace(*key))		/* Skip leading whitespace */
	key++;
      if ((*key == '\0') || (*key == '#') || !isdigit(*key))
	continue;
      end = key;
      while (isdigit(*end) || (*end == '.'))	/* Find end of digits */
	end++;
      if (*end == '\0') continue;	/* Address-only entry (!?!) */
      *end = '\0';

      ptr = ++end;			/* Now grab the text */
      while (isspace(*ptr))		/* Skip leading whitespace */
	ptr++;
      if ((*ptr == '\0') || (*ptr == '#') || !isalpha(*ptr))
	continue;			/* Address-only entry (!?!) */
      end = ptr;
      while ((*end != '\0') && (*end != '#') && !isspace(*end))
	end++;				/* Find end of first entry */
      *end = '\0';

      if (RoverDebug) fprintf(stderr, "Addr %s is %s\n", key, ptr);
      hentry = Tcl_CreateHashEntry(&HostsCache, key, &new);
      if (!new) {			/* Check for duplicate keys */
	fprintf(stderr,
		"WARNING! Duplicate addr %s in /etc/hosts [HTNetAddrToName:HTNet.c]\n",
		key);
      } else {
	if ((centry = malloc(strlen(ptr)+1)) == NULL) {
	  fprintf(stderr, "Out of memory [HTNetAddrToName:HTNet.c]\n");
	  return(NULL);
	}
      }
      strcpy(centry, ptr);
      /* Save cache entry */
      Tcl_SetHashValue(hentry, centry);
    }
    fclose(file);
  }

  hentry = Tcl_FindHashEntry(&HostsCache, buf);
  if (hentry == NULL) {
    fprintf(stderr,
	    "WARNING: Addr %s not in /etc/hosts [HTNetAddrToName:HTNet.c]\n",
	    buf);
    hentry = Tcl_CreateHashEntry(&HostsCache, buf, &new);
    if (!new) {			/* Check for duplicate keys */
      fprintf(stderr,
	      "WARNING! Entry %s found? [HTNetAddrToName:HTNet.c]\n", buf);
    } else {
      if ((centry = malloc(strlen(buf)+1)) == NULL) {
	fprintf(stderr, "Out of memory [HTNetAddrToName:HTNet.c]\n");
	return(NULL);
      }
    }
    strcpy(centry, buf);
    /* Save cache entry */
    Tcl_SetHashValue(hentry, centry);
    return(buf);
  }
  centry = Tcl_GetHashValue(hentry);
  return(centry);
}

PRIVATE void HTNetIfePrint(HTNetInterface *ptr)
{
  int hf = ptr->type;
  char *dispname = "overruns";
  
  if (strncmp(ptr->ifName,"lo",2) == 0)
    hf = 255;
  	
  if (hf == ARPHRD_CSLIP || hf == ARPHRD_CSLIP6) {
    /* Overrun got reused: BAD - fix later */
    dispname =  "compressed";
  }
  	
  fprintf(stderr, "Interface name is %s\n", ptr->ifName);
  fprintf(stderr,"%-8.8s  ", ptr->ifName);

  fprintf(stderr, "Host name is %s\n", ptr->name);
  fprintf(stderr,"%-8.8s  ", ptr->name);

  fprintf(stderr,"          ");
  if (ptr->flags == 0) fprintf(stderr,"[NO FLAGS] ");
  if (ptr->flags & IFF_UP) fprintf(stderr,"UP ");
  if (ptr->flags & IFF_BROADCAST) fprintf(stderr,"BROADCAST ");
  if (ptr->flags & IFF_DEBUG) fprintf(stderr,"DEBUG ");
  if (ptr->flags & IFF_LOOPBACK) fprintf(stderr,"LOOPBACK ");
  if (ptr->flags & IFF_POINTOPOINT) fprintf(stderr,"POINTOPOINT ");
  if (ptr->flags & IFF_NOTRAILERS) fprintf(stderr,"NOTRAILERS ");
  if (ptr->flags & IFF_RUNNING) fprintf(stderr,"RUNNING ");
  if (ptr->flags & IFF_NOARP) fprintf(stderr,"NOARP ");
  if (ptr->flags & IFF_PROMISC) fprintf(stderr,"PROMISC ");
  if (ptr->flags & IFF_ALLMULTI) fprintf(stderr,"ALLMULTI ");
  if (ptr->flags & IFF_SLAVE) fprintf(stderr,"SLAVE ");
  if (ptr->flags & IFF_MASTER) fprintf(stderr,"MASTER ");
  if (ptr->flags & IFF_MULTICAST) fprintf(stderr,"MULTICAST ");
  fprintf(stderr," MTU:%d  Metric:%d\n", ptr->mtu, ptr->metric?ptr->metric:1);


  /* If needed, display the interface statistics. */
  if ((hf < 255) &&
      (ptr->map.irq||ptr->map.mem_start||ptr->map.dma||ptr->map.base_addr)) {
    fprintf(stderr,"          ");
    if (ptr->map.irq)
      fprintf(stderr,"Interrupt:%d ", ptr->map.irq);
    if (ptr->map.base_addr >= 0x100)/* Only print devs using it for I/O maps */
      fprintf(stderr,"Base address:0x%x ", ptr->map.base_addr);
    if (ptr->map.mem_start)	{
      fprintf(stderr,"Memory:%lx-%lx ", ptr->map.mem_start, ptr->map.mem_end);
    }
    if (ptr->map.dma)
      fprintf(stderr,"DMA chan:%x ", ptr->map.dma);
    fprintf(stderr,"\n");
  }
  fprintf(stderr,"\n");
}


/* Fetch the inteface configuration from the kernel. */
PRIVATE int HTNetIfFetch(char *ifname, HTNetInterface *ife)
{
  struct ifreq ifr;

  memset((char *) ife, 0, sizeof(HTNetInterface));
  strcpy(ife->ifName, ifname);

  strcpy(ifr.ifr_name, ifname);
  if (ioctl(skfd, SIOCGIFFLAGS, &ifr) < 0) return(-1);
  ife->flags = ifr.ifr_flags;

  strcpy(ifr.ifr_name, ifname);
  if (ioctl(inet_sock, SIOCGIFADDR, &ifr) < 0) {
    memset(&ife->addr, 0, sizeof(struct sockaddr));
  } else ife->addr = ifr.ifr_addr;
  ife->name = HTNetAddrToName(&(ifr.ifr_addr));

  strcpy(ifr.ifr_name, ifname);
    if (ioctl(skfd, SIOCGIFHWADDR, &ifr) < 0) {
    memset(ife->hwaddr, 0, 32);
  } else memcpy(ife->hwaddr,ifr.ifr_hwaddr.sa_data,8);
  ife->type=ifr.ifr_hwaddr.sa_family;

  strcpy(ifr.ifr_name, ifname);
  if (ioctl(skfd, SIOCGIFMETRIC, &ifr) < 0) {
    ife->metric = 0;
  } else ife->metric = ifr.ifr_metric;
  if (((ife->type == ARPHRD_CSLIP) || (ife->type == ARPHRD_CSLIP6)) &&
      (ife->metric == 0)) {
    /* Set metric for SLIP/CSLIP to 1 (14.4Kb/s) */
    ife->metric = 1;
  }


  strcpy(ifr.ifr_name, ifname);
  if (ioctl(skfd, SIOCGIFMTU, &ifr) < 0) {
    ife->mtu = 0;
  } else ife->mtu = ifr.ifr_mtu;

  strcpy(ifr.ifr_name, ifname);
  if (ioctl(skfd, SIOCGIFMAP, &ifr) < 0) {
    memset(&ife->map, 0, sizeof(struct ifmap));
  } else memcpy(&ife->map,&ifr.ifr_map,sizeof(struct ifmap));

  strcpy(ifr.ifr_name, ifname);
  if (ioctl(inet_sock, SIOCGIFDSTADDR, &ifr) < 0) {
    memset(&ife->dstaddr, 0, sizeof(struct sockaddr));
  } else ife->dstaddr = ifr.ifr_dstaddr;

  strcpy(ifr.ifr_name, ifname);
  if (ioctl(inet_sock, SIOCGIFBRDADDR, &ifr) < 0) {
    memset(&ife->broadaddr, 0, sizeof(struct sockaddr));
  } else ife->broadaddr = ifr.ifr_broadaddr;

  strcpy(ifr.ifr_name, ifname);
  if (ioctl(inet_sock, SIOCGIFNETMASK, &ifr) < 0) {
    memset(&ife->netmask, 0, sizeof(struct sockaddr));
  } else ife->netmask = ifr.ifr_netmask;
  
  strcpy(ifr.ifr_name, ifname);
  if (ioctl(skfd, SIOCGIFMAP, &ifr) < 0) {
    memset(&ife->map, 0, sizeof(struct ifmap));
  } else ife->map = ifr.ifr_map;
  
  /* DDP address maybe ? */
  strcpy(ifr.ifr_name, ifname);
  if (ioctl(ddp_sock, SIOCGIFADDR, &ifr) == 0) {
    ife->ddpaddr = ifr.ifr_addr;
    ife->has_ddp = 1;
  }
  
  return(0);
}


PRIVATE void HTNetIfPrint(char *ifname)
{
  char buff[1024];
  HTNetInterface ife;
  struct ifconf ifc;
  struct ifreq *ifr;
  int i;

  if (ifname == (char *)NULL) {
    ifc.ifc_len = sizeof(buff);
    ifc.ifc_buf = buff;
    if (ioctl(skfd, SIOCGIFCONF, &ifc) < 0) {
      fprintf(stderr, "SIOCGIFCONF: %s\n", strerror(errno));
      return;
    }

    ifr = ifc.ifc_req;
    for (i = ifc.ifc_len / sizeof(struct ifreq); --i >= 0; ifr++) {
      if (HTNetIfFetch(ifr->ifr_name, &ife) < 0) {
	fprintf(stderr, "%s: unknown interface.\n", ifr->ifr_name);
	continue;
      }

/*      if (((ife.flags & IFF_UP) == 0) && !opt_a) continue; */
      HTNetIfePrint(&ife);
    }
  } else {
    if (HTNetIfFetch(ifname, &ife) < 0)
      fprintf(stderr, "%s: unknown interface.\n", ifname);
    else HTNetIfePrint(&ife);
  }
}


PRIVATE void HTNetIfStatus(HTNetState *state)
{
  char buff[1024], *name, *ifname;
  HTNetInterface ife;
  struct ifconf ifc;
  struct ifreq *ifr;
  int i, count;

  ifc.ifc_len = sizeof(buff);
  ifc.ifc_buf = buff;
  if (ioctl(skfd, SIOCGIFCONF, &ifc) < 0) {
    fprintf(stderr, "SIOCGIFCONF: %s\n", strerror(errno));
    return;
  }

  count = ifc.ifc_len / sizeof(struct ifreq);
  if (state->numInterfaces && state->interfaces)
    free(state->interfaces);
    
  state->numInterfaces = count;
  if (count)
    state->NetAvail = LOCALHOST;
  else
    state->NetAvail = NONE;
  if ((state->interfaces =
       (HTNetIfs *)malloc(count*sizeof(HTNetIfs))) == NULL) {
    fprintf(stderr, "Out of memory! [HTNetIfStatus:HTNet.c]\n");
    state->numInterfaces = 0;
    return;
  }

  ifr = ifc.ifc_req;
  for (i = 0; i < count; i++, ifr++) {
    if (HTNetIfFetch(ifr->ifr_name, &ife) < 0) {
      fprintf(stderr, "%s: unknown interface.\n", ifr->ifr_name);
      continue;
    }
    ifname = ifr->ifr_name;
    name = ife.name;
    strcpy(state->interfaces[i].ifName, ifname);
    state->interfaces[i].name = name;
    if (strncmp(name,"lo",2) == 0) {
      state->interfaces[i].type = LOCALHOST;
      state->interfaces[i].metric = 694;	/* = 10Mb/s / 14.4Kb/s */
    } else {
      state->interfaces[i].type = IP;
      state->NetAvail = IP;
      state->interfaces[i].metric = ife.metric;
    }
  }
}


PRIVATE int HTNetSocketsOpen()
{
  inet_sock = socket(AF_INET, SOCK_DGRAM, 0);
  if (inet_sock != -1) {
    RoverSetCloseExec(inet_sock);
    return(inet_sock);
  }
  ipx_sock = socket(AF_IPX, SOCK_DGRAM, 0);
  if (ipx_sock != -1) {
    RoverSetCloseExec(ipx_sock);
    return(ipx_sock);
  }
  ax25_sock = socket(AF_AX25, SOCK_DGRAM, 0);
  if (ax25_sock != -1) {
    RoverSetCloseExec(ax25_sock);
    return(ax25_sock);
  }
  ddp_sock = socket(AF_APPLETALK, SOCK_DGRAM, 0);
  if (ddp_sock != -1) {
    RoverSetCloseExec(ddp_sock);
    return(ddp_sock);
  }
  /*
   *	If this is -1 we have no known network layers and its time to jump.
   */
  return ddp_sock;
}

PRIVATE RoverErr HTNetUpdate()
{
  if (RoverDebug) fprintf(stderr, "Checking network status\n");
  /* Create a channel to the NET kernel. */
  if ((skfd = HTNetSocketsOpen()) < 0) {
    fprintf(stderr,"Unable to open channel to kernel [HTNetUpdate:HTNet.c]\n");
    perror("socket");
    return(ROVER_ERR);
  }
  HTNetIfStatus(&RoverNetState);
  close(skfd);
  return(ROVER_OK);
}

PUBLIC void HTNetShowInterfaces()
{
  /* Create a channel to the NET kernel. */
  if ((skfd = HTNetSocketsOpen()) < 0) {
    fprintf(stderr,
	    "Kernel open channel failed [HTNetShowInterfaces:HTNet.c]\n");
    perror("socket");
    return;
  }

  /* Do we have to show the current setup? */
  HTNetIfPrint((char *)NULL);
  close(skfd);
}
#endif /* linux */

/*

Scheduler issues:
1. Using metric to estimate transmission time?
2. Queue of:
	a. User import/export traffic
	b. Cache stat traffic

Server daemon?



*/
