 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTObject.c
 *
 * Description: Handles reception/transmission of objects and execution
 *		of object methods (local and remote).
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/
#define ROVER_ACCESSMGR

#include "sys/types.h"

#include "RoverBase.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTSession.h"
#include "HTObject.h"
#include "HTCache.h"
#include "HTLog.h"
#include "HTRoverHTTP.h"
#include "RoverZlib.h"
#include "RoverFile.h"
#include "HTThread.h"
#include "RoverTest.h"

PUBLIC CONST HTStreamClass RoverObject;

PRIVATE int ObjDoAbort(HTStream *me);

/* Marshalled object format: ASCII digits
   2      6      7      9         1          1            1        7    4
[type] [msgID] [code] [data] [keepalive] [compress] [cache type] [log] [dv]\n
<data><dv>
 */


PRIVATE void KeepAlive(HTStream *me, char *c, int count, int compDone)
{
  char *tmp = NULL;

  if (RoverDebug)
    fprintf(stderr, "%s: Keepalive: alloc %d, sz %d, cnt %d, compDone %d, comp %d\n",
	    myAppID, me->alloc, me->size, count, compDone, me->compress);
  if ((me->alloc == me->size) && (compDone == 1)) {	/* Done with this request? */
    if (count > 0) {				/* Data for next request */
      if (RoverDebug) fprintf(stderr,"Got %d bytes of extra data\n",count);
      if ((tmp = malloc(count)) == NULL) {
	fprintf(stderr,"Out of memory! [KeepAliveCompress:HTObject.c]\n");
	ObjDoAbort(me);
	return;
      }
      memcpy((void *) tmp, (void *) c, (size_t) count);
    }
    (*me->isa->_free)(me);			/* Invoke free routine */
    memset((void *) me, 0, sizeof(*me));
    me->isa = &RoverObject;
    me->logID = -1;
    if (RoverDebug)
      fprintf(stderr,"%s: Initializing next batch request\n", myAppID);
    if (count) {
      if (RoverDebug)
	fprintf(stderr, "%s: Processing %d leftover bytes\n", myAppID,
		count);
      (*me->isa->put_block)(me, tmp, count);	/* Send first block */
      free(tmp);
    }
  }
}

PRIVATE char *ObjectType(RoverObjType type)
{
  return((type == ROVER_BINARY) ? "ROVER_BINARY" :
	 ((type == ROVER_BINARY_UPDATE) ? "ROVER_BINARY_UPDATE" :
	  ((type == ROVER_TCL) ? "ROVER_TCL" :
	   ((type == ROVER_TCL_UPDATE) ? "ROVER_TCL_UPDATE" :
	    ((type == ROVER_TCL_TK) ? "ROVER_TCL_TK" :
	     ((type == ROVER_TCL_TK_UPDATE) ? "ROVER_TCL_TK_UPDATE" :
	      ((type == ROVER_ERROR_OBJ) ? "ROVER_ERROR_OBJ" :
	       ((type == ROVER_HTTP) ? "ROVER_HTTP" :
		((type == ROVER_HTTP_UPDATE) ? "ROVER_HTTP_UPDATE" :
		 ((type == ROVER_LOG_SUFFIX) ? "ROVER_LOG_SUFFIX" :
		  ((type == ROVER_CMD_COMPLETE) ? "ROVER_CMD_COMPLETE" :
		   "unknown")))))))))));
}

PRIVATE void Obj_allocate(HTStream *me, int size, char *c)
{
  int rc, type;
  QRPCState *context;
  HTNetQ *q;

  me->tmpbuf[OBJHDR] = '\0';
  if (RoverDebug)
    fprintf(stderr, "Obj tmp buf: |%s|\n", me->tmpbuf);
  rc = sscanf(me->tmpbuf, "%d %d %d %d %d %d %d %d %d", &type, &(me->logID),
	      &(me->csize), &(me->dsize), &(me->keepalive), &(me->compress),
	      &(me->cache), &(me->lsize), &(me->dvsize));
  if (rc != 9) {
    fprintf(stderr,
	    "Scanned %d items; not 9 |%s| [Obj_allocate:HTObject.c]\n",
	    rc, me->tmpbuf);
    ObjDoAbort(me);
    return;
  }
  me->type = (RoverObjType) type;

  if (!me->request) {
    /* Must be a batched request */
    if (!(q = LogFindLogID(me->logID))) {
      /* Treat as an unexpected object */
      fprintf(stderr,
	      "Request for LogID %d not found [Obj_allocate:HTObject.c].\n",
	      me->logID);
      ObjDoAbort(me);
      return;
    }
    me->request = q->req;
    me->anchor = q->req->anchor;
    me->url = HTAnchor_physical(q->req->anchor);
  }
  if (RoverDebug) fprintf(stderr,"Object type is %s [%s]\n",
			  ObjectType(me->type), RoverCacheType(me->cache));
  switch (me->type) {
  case ROVER_BINARY:
  case ROVER_BINARY_UPDATE:
    context = (QRPCState *)(me->request->context);
    me->isa = &HTApplication;       /* Change stream processor */
    me->sourceFile = NULL;
    if (me->cache == ROVER_NO_CACHE) {
      fprintf(stderr,
	      "Got uncacheable binary for %s [Obj_allocate:HTObject.c]\n",
	      me->url);
      ObjDoAbort(me);
      return;
    }
    me->session = context->session;
    me->url = context->url;
    if (me->cache == ROVER_VERIFIED) {
      me->sourceFileName[0] = '\0';
      if (size &&RoverDebug)
	fprintf(stderr,
		"Data after verify hdr (%d) [Obj_allocate:HTObject.c]\n",
		size);
    } else {
      if (me->type != ROVER_BINARY_UPDATE)
	RoverCacheRequest(me->url, me->session, me->type);
      sprintf(me->sourceFileName, "%s/RoverAppXXXXXXXX", RoverCacheDir);
      if ((me->sourceFile = RoverGetUniqFile(me->sourceFileName)) == NULL) {
	fprintf(stderr,
		"Failed creating name: %s [Obj_allocate:HTObject.c]\n",
		me->sourceFileName);
	ObjDoAbort(me);
	return;
      }

      if (RoverDebug)
	fprintf(stderr, "Source file name is %s, expected size %d\n",
		me->sourceFileName, me->csize);
    }
    me->alloc = me->dvsize;
    me->size = me->csize = 0;
    if (RoverDebug) fprintf(stderr, "Depend. vector is %d bytes\n", me->alloc);
    if ((me->tdv = malloc(me->alloc+10)) == NULL) { 	/* Null terminated */
      fprintf(stderr,"Out of memory! [Obj_allocate:HTObject.c]\n");
      ObjDoAbort(me);
      return;
    }
    if (me->compress) {
      if ((me->cbuf = malloc(65536)) == NULL) {
	fprintf(stderr, "Out of memory! [Obj_allocate:HTObject.c]\n");
	ObjDoAbort(me);
	return;
      }
      if (!(me->inflate = InitUncompress())) {
	ObjDoAbort(me);
	return;
      }
    }
    /* Now flush any leftover info to the file */
    if (size) (me->isa->put_block)(me, c, size);
    break;

  case ROVER_TCL_UPDATE:
  case ROVER_TCL_TK_UPDATE:
    if (RoverDebug) {
      context = (QRPCState *)(me->request->context);
      fprintf(stderr,"Got update for: %d::%s [Obj_allocate:HTObject.c]\n",
	      context->session, context->url);
    }
    /* Fall-through to Tcl/Tk path */

  case ROVER_TCL:
  case ROVER_TCL_TK:
  case ROVER_ERROR_OBJ:
  case ROVER_LOG_SUFFIX:
    if ((me->type == ROVER_ERROR_OBJ) && (me->cache != ROVER_BEST_EFFORT)) {
      fprintf(stderr,
	      "Warning! Rover error obj is not ROVER_BEST_EFFORT (%d) [Obj_allocate:HTObject.c]\n",
	      me->cache);
    } else if (me->type == ROVER_LOG_SUFFIX) {
      if (me->cache != ROVER_NO_CACHE) {
	fprintf(stderr,"Log suffix has cache != ROVER_NO_CACHE [Obj_allocate:HTObject.c]\n");
	ObjDoAbort(me);
	return;
      }
    } else if (me->cache == ROVER_VERIFIED) {
      if (size && RoverDebug)
	fprintf(stderr,
		"Data after verify hdr (%d) [Obj_allocate:HTObject.c]\n",
		size);
    }
    /* Allocate code, data, and dv areas. We allocate the code area first
       so that we can realloc the area if the data changes */
    me->alloc = me->csize + me->dsize + me->dvsize + me->lsize;
    if ((me->code = malloc(me->alloc+10)) == NULL) { 	/* Null terminated */
      fprintf(stderr,"Out of memory! [Obj_allocate:HTObject.c]\n");
      ObjDoAbort(me);
      return;
    }
    me->data = me->code + me->csize;
    me->log = me->data + me->dsize;
    me->tdv = me->log + me->lsize;
    me->size = 0;
    if (me->compress) 
      if (!(me->inflate = InitUncompress())) {
	ObjDoAbort(me);
	return;
      }
    /* Now flush any leftover info to the file */
    if (size) (me->isa->put_block)(me, c, size);
    break;

  case ROVER_HTTP:
  case ROVER_HTTP_UPDATE:
    rc = me->request->net_info->sockfd;
    if (RoverDebug) fprintf(stderr, "%s: Got HTTP-in-HTTP response for socket %d [Obj_allocate:HTObject.c]\n", myAppID, rc);

    /* Steal the socket from the library */
    HTThreadState(rc, THD_CLR_READ);
    HTThreadState(rc, THD_CLR_WRITE);
    if (RoverUseStolenSocket(rc, c, size) != EVENT_OK) ObjDoAbort(me);
    return;
    break;

  case ROVER_CMD_COMPLETE:
    if (RoverDebug)
      fprintf(stderr, "%s: Got command complete response [Obj_allocate:HTObject.c]\n", myAppID);
    if (LogDelete(me->logID, 0) != ROVER_OK) {
      fprintf(stderr,
	      "%s: Log ID delete failed [Obj_allocate:HTObject.c]\n", myAppID);
      ObjDoAbort(me);
      return;
    }
    if (me->request->callback) {
      context = (QRPCState *)(me->request->context);
      rc = (HTCallbackTable[(RoverCBList)me->request->callback])(me->request,
								 context);
      if (rc < 0)
	fprintf(stderr, "%s: Callback error %d [Obj_allocate:HTObject.c]\n",
		myAppID, rc);
    }
#if 0
    if (RoverQRPCTcpTest < 0) RoverExit(0);
    else if (--ResetWaitCount <= 0) {
      fprintf(stderr, "Reset wait count is %d\n", ResetWaitCount);
      RoverExit(0);
    }      
#endif
    break;

  default:
    fprintf(stderr,"Unknown obj hdr type: %d [Obj_allocate:HTObject.c]\n",
	    me->type);
    ObjDoAbort(me);
    return;
  }
}

PRIVATE void Obj_put_block(HTStream *me, CONST char *c, int size)
{
  int rc = 0, count = 0, done = 0;
  if (RoverDebug)
    fprintf(stderr, "Block size %d of %d (%d so far, %d left)\n", size,
	    me->alloc, me->size, me->alloc - me->size);
  if (me->alloc) {	/* Check if we're already allocated the array */
    if (me->compress) {
      rc = UncompressBuffer(me->inflate, (char *) c, &size, me->code,
			    &(me->size), me->alloc, &count);
      if (rc == -1) {
	fprintf(stderr, "Uncompress error [Obj_put_block:HTObject.c]\n");
	me->inflate = NULL;
	ObjDoAbort(me);
	return;
      } else if ((rc == 1) || (rc == 2))
	done = 1;
    } else {
      done = 1;
      if ((me->size + size) <= me->alloc) {
	memcpy((void *) &(me->code[me->size]), (void *) c, (size_t) size);
	me->size += size;
	size = 0;
      } else {
	rc = me->alloc - me->size;
	memcpy((void *) &(me->code[me->size]), (void *) c, (size_t) rc);
	me->size += rc;
	size -= rc;
	if (RoverDebug)
	  fprintf(stderr, "Saving %d bytes after block end [Obj_put_block:HTObject.c]\n",size);
	memmove((void *) c, (void *)(c + rc), size);
      }
    }
    if (me->keepalive) KeepAlive(me, (char *) c, size, done);
  } else {		/* Still waiting for header */
    if ((me->size + size) >= OBJHDR) {	/* Have we got enough? */
      while (me->size < OBJHDR) {	/* Grab enough for the hdr */
	me->tmpbuf[me->size++] = *(c++);
	size--;
      }
      Obj_allocate(me, size, (char *)c);
    } else {
      memcpy((void *) &(me->tmpbuf[me->size]), (void *) c, (size_t) size);
      me->size += size;
    }
  }
}

PRIVATE void Obj_put_character(HTStream *me, char c)
{
  Obj_put_block(me, (CONST char *)&c, 1);
}

PRIVATE void Obj_put_string(HTStream *me, CONST char *c)
{
  Obj_put_block(me, c, strlen(c));
}

PRIVATE int Obj_abort(HTStream *me, HTError e) {
  /* Abort for some reason. Clean up by releasing space */
  if (RoverDebug)
    fprintf(stderr, "Aborting object load of %s [Obj_abort:HTObject.c]\n",
	    me->url);
  if (me->inflate) free(me->inflate);
  if (me->cbuf) free(me->cbuf);
  if (me->alloc) free(me->code); /* Free any code or data */

  /* Ignore HTTP, HTTP_UPDATE objects */
  if ((me->type == ROVER_HTTP) || (me->type == ROVER_HTTP_UPDATE)) { 
    if (LogDelete(me->logID, 0) != ROVER_OK)
      fprintf(stderr, "Log ID delete failed [Obj_abort:HTObject.c]\n");
    if (RoverHTTPMax) {		/* HTTP timing tests */
      if (RoverDebug) fprintf(stderr, "Calling CB from Obj_abort\n");
      RoverHTTPDoTest();
    }
  }

  free(me);
  return EOF;
}

PRIVATE void Obj_invokeCallbacks(HTStream *me, RoverObj* obj, char *mObj,
				      int size)
{
  QRPCState *next, *context = (QRPCState *)(me->request->context);
  int i, logID;

  if (context) {
    /* Shared contexts share the same log ID; so delete it only once */
    logID = context->logID;
    if (LogDelete(logID, me->keepalive) != ROVER_OK) {
      fprintf(stderr,
	      "%s: Log ID delete failed [Obj_invokeCallbacks:HTObject.c]\n", myAppID);
    }
  }

  while (context != NULL) {
    next = context->next;
    if (logID != context->logID) {	/* Shared context = shared log ID */
      fprintf(stderr,
	      "%s: WARNING! Got different log ID [Obj_invokeCallbacks:HTObject.c]\n", myAppID);
    }

    /* Put the object location in the context */
    if (obj) context->obj = obj;
    if (mObj) {
      context->marshallObj = mObj;
      context->marshallSize = size;
    } else {
      context->marshallObj = NULL;
      context->marshallSize = 0;
    }

    /* Set the type of the object based upon the type of the message */
    context->type = Object2RoverTypeString(me->type);

    /* Check the me->cache variable to see if the object was verified */

    if (!me->request->callback) {
      /* Ignore this, since the request may have already completed (i.e.,
	 returned HT_LOADED), in which case, the event loop will invoke
	 the callback for us.
	 */
      context = next;
      continue;
    }
    i = (HTCallbackTable[(RoverCBList)me->request->callback])(me->request,
							      context);
    if (i < 0)
      fprintf(stderr, "%s: Callback error %d [Obj_invokeCallbacks:HTObject.c]\n", myAppID, i);
    context = next;
  }
}

PRIVATE void Obj_freeComponents(RoverObj *obj)
{
  if (obj->tdsize) {
#if 0
    fprintf(stderr,
	    "%s: WARNING! Flushing changes to %d::%s!!!\n", myAppID,
	    obj->session, obj->url);
#endif
    if (obj->tdata) free(obj->tdata);
  }
  if (obj->url) free(obj->url);
  if (obj->code) free(obj->code);
  if (obj->inuse == AllFields) {	/* All fields dynamically allocated */
    if (obj->dsize) free(obj->data);
    if (obj->lsize) free(obj->log);
  }
  if (obj->dv) free(obj->dv);
  if (obj->unmarshall) free(obj->unmarshall);
  if (obj->name) free(obj->name);
  memset((void *) obj, 0, sizeof(RoverObj)); /* Set all pointers/fields to NULL */
  obj->inuse = Unused;
}

PUBLIC void RoverObjFree(RoverObj *obj)
{
  Obj_freeComponents(obj);
  free(obj);
}

PUBLIC RoverType Object2RoverTypeString(RoverObjType type)
{
  /* Check if object is an error response */
  if (type == ROVER_ERROR_OBJ) return(ROVER_ERROR_REPLY);
  else if (type == ROVER_LOG_SUFFIX) return(ROVER_LOGSUFFIX);
  else return(ROVER_OBJECT);
}

PRIVATE RoverObj *RoverObjectInit(HTStream *me, char **marshall, int *msize)
{
  QRPCState *context = (QRPCState *)(me->request->context);
  RoverObj *obj;
  unsigned long server, clock;
  char *data1 = NULL;
  RoverErr rc;

  *msize = 0;
  *marshall = NULL;
  if ((obj = (RoverObj *)calloc(1, sizeof(RoverObj))) == NULL) {
    fprintf(stderr, "Out of memory [RoverObjectInit:HTObject.c]\n");
    return(NULL);
  }

  /* Initialize the object */
  obj->tdata = NULL;
  obj->tdsize = 0;
  if (me->csize) me->code[me->csize-1] = '\0';
  if (me->dsize) me->data[me->dsize-1] = '\0';
  obj->cache = me->cache;
  obj->inuse = CodeOnly;
  obj->code = me->code;
  obj->csize = me->csize;
  obj->data = me->data;
  obj->dsize = me->dsize;
  obj->oper = NULL;
  obj->operCount = 0;
  obj->unmarshall = malloc(1);
  obj->unmarshall[0] = '\0';
  obj->name = malloc(1);
  obj->name[0] = '\0';

  obj->anchor = me->anchor;

  if (me->dvsize) sscanf((char *) me->tdv, "%lu_%lu", &server, &clock);
  else server = clock = 0;
  obj->dvsize = sizeof(dependV);
  if ((obj->dv = (dependV *)malloc(sizeof(dependV))) == NULL) {
    fprintf(stderr, "Out of memory! [RoverObjectInit:HTObject.c]\n");
    Obj_freeComponents(obj);
    return(NULL);
  }
  obj->dv[0].ServerID = server;
  obj->dv[0].ServerClock = clock;
  
  obj->type = Object2RoverTypeString(me->type);
  obj->interp = NULL;	/* For now, objects only execute in
				   application space */

  if (me->cache == ROVER_NO_CACHE) {
    obj->url = malloc(1);
    obj->url[0] = '\0';
    obj->session = -1;
  } else {		/* Only set these fields if the obj is being cached */
    obj->lsize = 0;
    obj->log = "";
    obj->session = context->session;
    if ((obj->url = strdup(context->url)) == NULL) {
      fprintf(stderr, "Out of memory! [RoverObjectInit:HTObject.c]\n");
      Obj_freeComponents(obj);
      return(NULL);
    }
    time(&(obj->lastRef));
    obj->loadDelay = obj->lastRef - context->loadTime;
    if (RoverDebug)
      fprintf(stderr, "Object `%s' load time: %lu seconds\n", obj->url,
	      obj->loadDelay);
    obj->refCount = 1;
    *msize = ObjectMarshallSize(obj);
    if ((data1 = *marshall = malloc(*msize)) == NULL) {
      fprintf(stderr, "Out of memory! [RoverObjectInit:HTObject.c]\n");
      Obj_freeComponents(obj);
      return(NULL);
    }
    MarshallObj(&data1, obj);
    rc = RoverCacheInsertData(obj->url, obj->session,me->type,me->cache,
			      *marshall, *msize,
			      (obj->dvsize ? ((time_t) obj->dv[0].ServerClock)
			       : context->loadTime), obj->loadDelay,
			      obj->lastRef, me->tdv, me->dvsize);
    if (rc != ROVER_OK)
      fprintf(stderr,"Cache insert fail %d::%s [RoverObjectInit:HTObject.c]\n",
	      obj->session, obj->url);
  }
  obj->lsize = me->lsize;
  obj->log = me->log;
  return(obj);
}

PRIVATE int Obj_free(HTStream *me)
{
  QRPCState *context = (QRPCState *)(me->request->context);
  RoverObj *obj;
  time_t now;
  HTRequest *req; 
  HTNetQ *q;
  char *data2;
  int rc, size;
  RoverErr rv;
  extern int RoverTestFinish;
  extern void TestSetTime();
#ifdef ROVER_RTT
  struct timeval tv;
  struct timezone tz;

  gettimeofday(&tv, &tz);
  fprintf(stderr,"%lu.%lu [Obj_free:HTOject.c]\n", tv.tv_sec, tv.tv_usec);
#endif /* ROVER_RTT */  
  
  if (RoverTestFinish) TestSetTime();

  /* Ignore COMMAND_COMPLETE objects */
  if (me->type == ROVER_CMD_COMPLETE) return(EOF);

  /* Ignore HTTP, HTTP_UPDATE objects */
  if ((me->type == ROVER_HTTP) || (me->type == ROVER_HTTP_UPDATE)) { 
    if (LogDelete(me->logID, 0) != ROVER_OK)
      fprintf(stderr, "Log ID delete failed [Obj_free:HTObject.c]\n");
    if (RoverHTTPMax) {		/* HTTP timing tests */
      if (RoverDebug) fprintf(stderr, "Calling CB from Obj_free\n");
      RoverHTTPDoTest();
    }
    return(EOF);
  }

  /* Done with receiving the object? */
  if (!me->alloc || (me->alloc != me->size)) {	/* Incomplete / nothing sent */
    if (!context) return(EOF);
    if (!(q = LogFindLogID(context->logID))) {
      /* Treat as an unexpected object */
      fprintf(stderr,
	      "Request for LogID %d not found [Obj_free:HTObject.c].\n",
	      context->logID);
      return(EOF);
    }
    /* 
     * Duplicate the request structure so that the httpd request can be
     *   handled with a copy of the original request. The original request
     *   is deleted when we return EOF.
     */
    req = Thread_new();
    *req = *(q->req);
    q->req = req;
    q->req->error_stack = HTList_new();	/* Allocate a new error stack */
#if 0
    LogRetry(me);
    Obj_noData(me);
#endif
    return(EOF);
  }

  if (me->dvsize == 0) *(me->tdv) = '\0';
  if (me->cache == ROVER_VERIFIED) {
    /* Get the object */
    if (RoverDebug)
      fprintf(stderr, "Cache verification message for %d::%s dv: %s\n",
	      context->session, context->url, me->tdv);
    time(&now);
    rv = RoverCacheRef(context->url, context->session, now);
    if (rv != ROVER_OK) {
      fprintf(stderr, "Cache ref update failed: %s\n", RoverErrType(rv));
      return(EOF);
    }
    rv = RoverCacheUpdateDependV(context->url,context->session, me->tdv,
				 me->dvsize);
    if (rv != ROVER_OK) {
      fprintf(stderr, "Cache update dv %d::%s failed\n", context->session,
	      context->url);
      return(EOF);
    }
    data2 = RoverCacheSearchData(context->url, context->session, &size);
    Obj_invokeCallbacks(me, NULL, data2, size);
    if (!me->keepalive) free(me);
    if (me->cbuf) free(me->cbuf);
    return(EOF);
  }

#ifdef ROVER_RTT
  fprintf(stderr, "RTT Recv Size is %d\n", me->size);
#endif /* ROVER_RTT */

  /* First, determine if we should cache the object */
  if ((me->type != ROVER_LOG_SUFFIX) && (me->cache == ROVER_NO_CACHE)) {
    /* Flush the cache, in case the object changed cacheability */
    if ((rc=Cache_ServerFlush(context->url, context->session)) != ROVER_OK) {
      fprintf(stderr,
	      "Server flush of %d::%s failed [Obj_free:HTObject.c]\n",
	      context->session, context->url);
      return(EOF);
    }
  }

  /* Next, initialize the object */
  if (!(obj = RoverObjectInit(me, &data2, &size))) {
    fprintf(stderr, "Object init failed [Obj_free:HTObject.c]\n");
    return(EOF);
  }

  /* If there's no log and the object is cacheable, send marshalled copy */
  if (!obj->lsize && (me->cache != ROVER_NO_CACHE))
    Obj_invokeCallbacks(me, obj, data2, size);
  else
    Obj_invokeCallbacks(me, obj, NULL, 0);

  if (me->cache == ROVER_NO_CACHE) {
    /* Free the object */
    Obj_freeComponents(obj);
    free(obj);
  }
  if (!me->keepalive) free(me);
  if (me->cbuf) free(me->cbuf);
  return(EOF);
}



/*		Structured Object Class
**		-----------------------
*/
PUBLIC CONST HTStreamClass RoverObject =
{		
	"Object Receiver",
	Obj_free,
	Obj_abort,
	Obj_put_character,
	Obj_put_string,
	Obj_put_block
}; 


/*		New object
**		----------
*/
PUBLIC HTStream *ObjectReceive ARGS5(
	HTRequest *,		request,
	void *,			param,
        HTFormat,               input_format,
        HTFormat,               output_format,
	HTStream *,		output_stream)
{
  HTStream* me = (HTStream*)calloc(sizeof(*me),1);
  me->isa = &RoverObject;
  me->anchor = request->anchor;
  me->url = HTAnchor_physical(request->anchor);
  me->request = request;
  me->logID = -1;
  return (HTStream*) me;
}

PUBLIC HTStream *ObjectCreate(HTRequest *request)
{
  return(ObjectReceive(request, NULL, NULL, NULL, NULL));
}


PRIVATE void Obj_putcAbort(HTStream *me, CONST char c) {
  return;
}

PRIVATE void Obj_putsAbort(HTStream *me, CONST char *c) {
  return;
}

PRIVATE void Obj_putbAbort(HTStream *me, CONST char *c, int length) {
  return;
}

PRIVATE int Obj_freeAbort(HTStream *me)
{
  /* Abort for some reason. Clean up by closing file and unlinking it */
  fprintf(stderr, "%s: Finishing aborted object load\n", myAppID);
  return(Obj_abort(me, (HTError) NULL));
}

/*		Structured Object Class
**		-----------------------
*/
PRIVATE CONST HTStreamClass ObjectAbort =
{		
	"Object Abort",
	Obj_freeAbort,
	Obj_abort,
	Obj_putcAbort, Obj_putsAbort, Obj_putbAbort
}; 

PRIVATE int ObjDoAbort(HTStream *me)
{
  /* Abort for some reason */
  fprintf(stderr, "%s: Aborting object load\n", myAppID);
  me->isa = &ObjectAbort;
  return EOF;
}

PUBLIC RoverObj *Cache_ServerCheck(char *url, int session)
{
  int size;
  char *data;

  if (RoverDebug) fprintf(stderr,"%s: Cache check for %d:%s\n", myAppID,
			  session, url);

  data = RoverCacheSearchData(url, session, &size);
  if (!data) return((RoverObj *)NULL);
  return(UnmarshallObj(&data));
}

PUBLIC RoverErr Cache_ServerFlush(char *url, int session)
{
  char *urn = RoverSessionGetURN(session);
  
  if (!urn) return(ROVER_INVLD_SES);
  if (RoverDebug)
    fprintf(stderr, "%s: Checking cache for %s:%s\n", myAppID, urn, url);

  return(RoverCacheFlush(url, session));
}

PUBLIC RoverErr Cache_ServerUpdate(char *url, int session, char *data,
				   unsigned int server, unsigned int clock,
				   int size, RoverCacheable cache)
{
  unsigned long dvsize;
  char *buf, dv[256];
  char *urn;
  RoverErr rc = ROVER_OK;

  if (cache == ROVER_NO_CACHE) return(RoverCacheFlush(url, session));
  
  
  if (!(urn = RoverSessionGetURN(session))) return(ROVER_INVLD_SES);
  if (RoverDebug) fprintf(stderr,"%s: Check'g cache for %s:%s\n", myAppID,
			  urn, url);

  if ((buf = (char *)malloc(size)) == NULL) {
    fprintf(stderr, "Out of memory! [Cache_ServerUpdate:HTObject.c]\n");
    return(ROVER_NO_MEM);
  }
  memcpy(buf, data, size);
  sprintf(dv, "%u_%u", server, clock);
  dvsize = strlen(dv)+1;
  rc = RoverCacheUpdateData(url, session, buf, size, dv, dvsize);
  if (rc != ROVER_OK) 
    fprintf(stderr, "Cache update %d::%s failed\n", session, url);
  else if (RoverDebug) 
    fprintf(stderr, "%s: Updated %s:%s\n", myAppID, urn, url);

  return(rc);
}

PUBLIC RoverErr Cache_ServerInsert(char *url, int session, char *data,
				   unsigned int server, unsigned int clock,
				   int size, RoverCacheable cache)
{
  unsigned long dvsize;
  char dv[256], *buf, *urn;
  time_t now;
  RoverErr rc = ROVER_OK;
  RoverCacheable cache1;
  char *tdv;
  RoverObjType type;

  if (cache == ROVER_NO_CACHE) return(ROVER_OK);
  if (!(urn = RoverSessionGetURN(session))) return(ROVER_INVLD_SES);

  if (RoverDebug)
    fprintf(stderr, "%s: Check'g cache for %s:%s\n", myAppID, urn, url);

  /* Check if object is already here already */
  if (RoverCacheSearchName(url, session, &cache1, &type, &tdv, &dvsize)) {
    if (RoverDebug)
      fprintf(stderr,
	      "%s: Cache hit %s:%s (%s) [Cache_ServerInsert:HTObject.c]\n", myAppID,
	      urn, url, RoverCacheType(cache1));
    return(ROVER_OBJ_EXISTS);
  } else {
    if ((buf = (char *)malloc(size)) == NULL) {
      fprintf(stderr, "Out of memory! [Cache_ServerInsert:HTObject.c]\n");
      return(ROVER_NO_MEM);
    }
    memcpy(buf, data, size);
    time(&now);
    sprintf(dv, "%u_%u", server, clock);
    dvsize = strlen(dv)+1;
    rc = RoverCacheInsertData(url, session, ROVER_TCL_TK, cache, buf, size,
			      ((dvsize != 0) ? ((time_t) clock) : now), 0, now,
			      dv, dvsize);
    if (rc != ROVER_OK)
      fprintf(stderr,
	      "Cache insert fail %d::%s [Cache_ServerInsert:HTObject.c]\n",
	      session, url);
  }
  return(rc);
}

#if 0
#include "RoverBuffer.h"

PRIVATE void Obj_noData(HTStream *me)
{
  HTStream *stream;
  RoverBuffer *err;
  HTErrSeverity severity;
  char buffer[128];
  QRPCState *context = (QRPCState *)(me->request->context);
  
  fprintf(stderr, "Opps! No data received! [Obj_noData:HTObject.c]\n");
  /* The error stack might contain general information to the client
     about what has been going on in the library (not only errors) */
  if (me->request->error_stack && 
      (err = RoverErrorMsg(me->request, &severity))) {
    /* Return a ROVER_ERROR_OBJECT */
    sprintf(buffer, "%2d %6d %7d %9d %1d %1d %1d %7d %4d\n", ROVER_ERROR_OBJ,
	    context->logID, 0, strlen((char *)RoverBufferBuf(err)), 0, 0,
	    ROVER_BEST_EFFORT, 0, 0);
    stream = ObjectCreate(me->request);
    (*stream->isa->put_string)(stream, buffer);
    (*stream->isa->put_string)(stream, err->data);
    (*stream->isa->_free)(stream);
    RoverBufferFree(err);
  }
}
#endif

