 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTObject.h
 *
 * Description: Handles reception/transmission of objects and execution
 *		of object methods (local and remote).
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#ifndef OBJ_H
#define OBJ_H

extern HTStream *ObjectReceive PARAMS((HTRequest *     request,
				       void *          param,
				       HTFormat        input_format,
				       HTFormat        output_format,
				       HTStream *      output_stream));

#if defined(ROVER_CLIENT) || defined(ROVER_ACCESSMGR)

#include "zlib.h"

/*		Rover Object
**		------------
*/

/* Marshalled object header format: ASCII digits
   2      6      7      9         1          1            1        7    4
[type] [msgID] [code] [data] [keepalive] [compress] [cache type] [log] [dv]\n

    ==> 47 characters
 */

#define OBJHDR 47

struct _HTStream {
  CONST HTStreamClass *	isa;

  RoverObjType type;		/* Rover Object type */
  char *url;			/* Object's url */
  int session;			/* Session ID of request */
  char tmpbuf[OBJHDR+1];	/* Buffer for storing transmission header */
  HTParentAnchor *anchor;	/* Object's parent anchor */
  HTRequest *request;		/* libwww request data structure */
  int cache;			/* Cacheable? flag */
  char *log;			/* Object's log suffix */
  int lsize;			/* Log suffix size */
  int logID;			/* Log ID of client's request */
  int alloc;			/* Allocated space for object */
  int size;			/* Number of bytes received so far */
  char *code;			/* Object's code */
  int csize;			/* Code size */
  char *data;			/* Object's data */
  int dsize;			/* Data size */
  char *tdv;			/* Marshalled dependency vector */
  dependV *dv;			/* Dependency vector */
  int dvsize;			/* Size of dependency vector */
  int compress;			/* Compression level */
  z_stream *inflate;		/* Decompression structure */
  char *cbuf;			/* Decompression buffer */
  int keepalive;		/* Keep connection alive for more responses */

  /* For Application binaries */
  FILE *sourceFile;		/* File pointer for application binary */
  char sourceFileName[256];	/* Name of application's file */
};

#define MAXOPENFILES	1024
extern FILE *OpenFiles[MAXOPENFILES];

extern HTStream *ObjectCreate(HTRequest *request);
extern RoverType Object2RoverTypeString(RoverObjType type);
extern void RoverObjFree(RoverObj *obj);
extern RoverErr ObjectFree(RoverObj *obj);
extern RoverErr Cache_Flush(char *url, int session);
extern RoverErr Cache_ClientFlush(char *url, int session);
extern RoverErr Cache_ServerFlush(char *url, int session);
extern RoverObj *Cache_check(char *url, int session);
extern RoverErr Cache_ClientInsert(RoverObj *obj);
extern RoverErr Cache_insert(RoverObj *obj);
extern void RoverObj_print(RoverObj *obj);
extern RoverObj *Cache_ServerCheck(char *url, int session);
extern RoverErr Cache_ServerUpdate(char *url, int session, char *data,
				   unsigned int server, unsigned int clock,
				   int size, RoverCacheable cache);
extern RoverErr Cache_ServerInsert(char *url, int session, char *data,
				   unsigned int server, unsigned int clock,
				   int size, RoverCacheable cache);
extern int Cache_ClientUpdate(RoverObj *obj);
extern RoverObj *Cache_fetch(char *url, int session, char *method,
			  RoverCallback callback, void *data, int *msgID);
extern char *RoverCacheType(RoverCacheable type);

#define ObjectMarshallSize(E)	((sizeof(int)*11)+(E)->dsize+(E)->csize+\
				 (E)->tdsize+(E)->dvsize+(E)->lsize+\
				 strlen((E)->url)+strlen((E)->unmarshall)+\
				 strlen((E)->name))

PRIVATE inline void MarshallObj(char **buffer, RoverObj *obj)
{
  /*
   * modifies: buffer (Must contain enough space!)
   * effects:  Encodes the object in the buffer.
   *           Advances buffer pointer by size of encoded object.
   *
   */
  MarshallBin(buffer, obj->csize, obj->code);
  MarshallBin(buffer, obj->dsize, obj->data);
  MarshallBin(buffer, obj->tdsize, obj->tdata);
  MarshallBin(buffer, obj->lsize, obj->log);
  MarshallBin(buffer, obj->dvsize, (char *) obj->dv);
  MarshallInt(buffer, (int) obj->type);
  MarshallInt(buffer, (int) obj->cache);
  MarshallInt(buffer, obj->session);
  MarshallString(buffer, obj->url);
  MarshallString(buffer, obj->unmarshall);
  MarshallString(buffer, obj->name);
}

PRIVATE inline RoverObj *UnmarshallObj(char **buffer)
{
  /*
   * modifies: buffer
   * effects:  Returns the object encoded by buffer.
   *           Advances buffer pointer by size of encoded object.
   *
   */
  RoverObj *rv = (RoverObj *) malloc(sizeof(RoverObj));

  if (rv == NULL) {
    fprintf(stderr,"Out of memory! [UnmarshallObj:<inline>]\n");
    RoverExit(-1);
  }

  rv->code = UnmarshallBin(buffer, &(rv->csize));
  rv->data = UnmarshallBin(buffer, &(rv->dsize));
  rv->tdata = UnmarshallBin(buffer, &(rv->tdsize));
  rv->log =  UnmarshallBin(buffer, &(rv->lsize));
  rv->dv = (dependV *) UnmarshallBin(buffer, &(rv->dvsize));
  rv->type = (RoverType) UnmarshallInt(buffer);
  rv->cache = (RoverCacheable) UnmarshallInt(buffer);
  rv->session = UnmarshallInt(buffer);
  rv->url = UnmarshallString(buffer);
  rv->unmarshall = UnmarshallString(buffer);
  rv->name = UnmarshallString(buffer);
  rv->inuse = AllFields;
  rv->deleted = rv->operCount = 0;
  rv->oper = NULL;
  rv->anchor = NULL;
  rv->interp = NULL;	/* Objects only execute in application space */
  return(rv);
}

PRIVATE inline void UnmarshallObjInPlace(char **buffer, RoverObj *rv)
{
  /*
   * modifies: buffer (Must contain enough space!)
   * effects:  Returns the object encoded by buffer.
   *           Advances buffer pointer by size of encoded object.
   *
   */
  rv->code = UnmarshallBin(buffer, &(rv->csize));
  rv->data = UnmarshallBin(buffer, &(rv->dsize));
  rv->tdata = UnmarshallBin(buffer, &(rv->tdsize));
  rv->log =  UnmarshallBin(buffer, &(rv->lsize));
  rv->dv = (dependV *) UnmarshallBin(buffer, &(rv->dvsize));
  rv->type = (RoverType) UnmarshallInt(buffer);
  rv->cache = (RoverCacheable) UnmarshallInt(buffer);
  rv->session = UnmarshallInt(buffer);
  rv->url = UnmarshallString(buffer);
  rv->unmarshall = UnmarshallString(buffer);
  rv->name = UnmarshallString(buffer);
  rv->inuse = AllFields;
  rv->deleted = rv->operCount = 0;
  rv->oper = NULL;
  rv->anchor = NULL;
  rv->interp = NULL;	/* Objects only execute in application space */
}

#endif /* defined(ROVER_CLIENT) || defined(ROVER_ACCESSMGR) */
#endif /* OBJ */
