 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTObjectClient.c
 *
 * Description: Handles local caching and executioon of objects
 *
 * Last Modified:  7-Feb-95 (adj)
 *
 * Global Functions:
 *    void set_prof_(pc, profileInfo *, const char *)
 *        Record address of profileInfo for procedure at pc.
 *        Called only from InitProfile_
 *
 * Global Variables: none
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/
#define ROVER_CLIENT

#include "sys/types.h"

#include "RoverBase.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTSession.h"
#include "HTObject.h"

PRIVATE Tcl_HashTable ObjClientCache;
PRIVATE int ClientCacheInit = 0;

PRIVATE void Obj_Callback(int appID, int toFD, int cbID, RoverType type,
			  int size, char *buf, char *data)
{
  AMRetVal *info;
  if (data) {
    info = (AMRetVal *) data;
    info->type = type;
    info->size = size;
    info->buf = buf;
    info->logID = 0;
    info->status = ROVER_NOTQUEUED;
  }
}

PRIVATE void ObjClientInit()
{
  Tcl_InitHashTable(&ObjClientCache, TCL_STRING_KEYS);
  ClientCacheInit = 1;
}

PUBLIC RoverErr ObjectFree(RoverObj *obj)
{
  if (obj->operCount > 0 || obj->tdsize) {
    fprintf(stderr,
	    "WARNING! Freeing obj (%d/%s:%s) with %d pending operations %d!\n",
	    obj->session, SessionTable[obj->session].ServiceURN, obj->url,
	    obj->operCount, obj->tdsize);
    return(ROVER_ERR);
  }
  if (obj->code) free(obj->code);
  if (obj->data) free(obj->data);
  if (obj->tdata) free(obj->tdata);
  if (obj->log) free(obj->log);
  if (obj->dv) free(obj->dv);
  if (obj->url) free(obj->url);
/*  if (obj->interp) free(obj->interp); */
  if (obj->unmarshall) free(obj->unmarshall);
  if (obj->name) free(obj->name);
  free(obj);
  return(ROVER_OK);
}

PRIVATE RoverErr SendObject(RoverObj *obj, RoverType type, int session, char *url)
{
  int size, objSize, msgID, rc;
  char *sbuf, *sbuf2;
  AMRetVal info;

  /* Send an object to the Access Manager */
  objSize = ObjectMarshallSize(obj);
  size = sizeof(int)*6+strlen(url)+objSize;
  if ((sbuf = sbuf2 = malloc(size)) == NULL) {
    fprintf(stderr, "Out of memory [SendObject:HTObjectClient.c]\n");
    return(ROVER_NO_MEM);
  }
  MarshallInt(&sbuf2, session);
  MarshallString(&sbuf2, url);
  MarshallInt(&sbuf2, obj->dv[0].ServerID);
  MarshallInt(&sbuf2, obj->dv[0].ServerClock);
  MarshallInt(&sbuf2, (int) obj->cache);
  MarshallInt(&sbuf2, objSize);
  MarshallObj(&sbuf2, obj);
  msgID = RoverClientSend(type, session, Obj_Callback, (char *)&info,
			  ROVER_NO_DELETE, size, sbuf, 1);
  if (msgID < 0) {
    fprintf(stderr,
	    "Client send err %d [SendObject:HTObjectClient.c]\n",
	    msgID);
    return(ROVER_SEND_ERR);
  }

  /* Should we wait here? */
  RoverClientWaitMsgID(msgID, 1);	/* Block until we get a response */
  RoverDeleteCallback(msgID);
  if (info.type != ROVER_REPLY) {
    fprintf(stderr,
	    "Reply %s (%d) != ROVER_REPLY [SendObject:HTObjectClient.c]\n",
	    RoverTypeString(info.type), info.type);
    if (info.buf) free(info.buf);
  }
  if (info.size != sizeof(int)) {
    fprintf(stderr,
	    "Reply size %d != sizeof(int) [SendObject:HTObjectClient.c]\n",
	    info.size);
    if (info.buf) free(info.buf);
  }
  if ((sbuf = info.buf) == NULL) {
    fprintf(stderr,
	    "NULL reply buffer [SendObject:HTObjectClient.c]\n");
  }
  rc = UnmarshallInt(&sbuf);
  free(info.buf);
  return(rc);
}

PUBLIC RoverErr Cache_Flush(char *url, int session)
{
  AMRetVal info;
  int rc, size, msgID;
  char *ses, *sbuf, *sbuf2;

  ses = SessionTable[session].ServiceURN;
  if (!SessionTable[session].inuse) {
    fprintf(stderr, "Invalid session ID [Cache_Flush:HTObjectClient.c]\n");
    return(ROVER_INVLD_SES);
  }

  if (RoverDebug) fprintf(stderr, "%s: Flushing %s:%s\n", myAppID, ses, url);
  if ((rc = Cache_ClientFlush(url, session)) != ROVER_OK) {
    if (RoverDebug) fprintf(stderr, "%s: Object not in cache (%s:%s)\n",
			    myAppID, ses, url);
  }

  /* Perform access manager cache flush */
  size = sizeof(int)*2+strlen(url);
  if ((sbuf = sbuf2 = malloc(size)) == NULL) return(0);
  MarshallInt(&sbuf2, session);
  MarshallString(&sbuf2, url);

  msgID = RoverClientSend(ROVER_FLUSH_CACHE, session, Obj_Callback,
			  (char *)&info, ROVER_NO_DELETE, size, sbuf, 1);
  if (msgID < 0) {
    fprintf(stderr,
	    "ROVER access mgr send FAIL %d [Cache_Flush:HTObjectClient.c]\n",
	    msgID);
    return(ROVER_SEND_ERR);
  }
  RoverClientWaitMsgID(msgID, 1);	/* Block until we get a response */
  RoverDeleteCallback(msgID);
  if (info.type != ROVER_REPLY) {
    fprintf(stderr,
	    "Reply %s (%d) != ROVER_REPLY [Cache_Flush:HTObjectClient.c]\n",
	    RoverTypeString(info.type), info.type);
    if (info.buf) free(info.buf);
  }
  if (info.size != sizeof(int)) {
    fprintf(stderr,
	    "Reply size %d != sizeof(int) [Cache_Flush:HTObjectClient.c]\n",
	    info.size);
    if (info.buf) free(info.buf);
  }
  if ((sbuf = info.buf) == NULL) {
    fprintf(stderr,
	    "NULL reply buffer [Cache_Flush:HTObjectClient.c]\n");
  }
  sbuf = info.buf;
  rc = UnmarshallInt(&sbuf);
  free(info.buf);
  return(rc);
}

PUBLIC RoverErr Cache_ClientFlush(char *url, int session)
{
  Tcl_HashEntry *hentry;
  RoverObj *obj;
  char ptr[1024];
  RoverErr rc = ROVER_OK;

  if (!SessionTable[session].inuse) {
    fprintf(stderr,
	    "Invalid session ID [Cache_ClientFlush:HTObjectClient.c]\n");
    return(ROVER_INVLD_SES);
  }
  if (!ClientCacheInit) ObjClientInit();

  sprintf(ptr, "%d::%s", session, url);
  ASSERT((strlen(ptr)+1)<=1024);
  if ((hentry = Tcl_FindHashEntry(&ObjClientCache, ptr)) == NULL) return(0);
  obj = (RoverObj *) Tcl_GetHashValue(hentry);
  rc = ObjectFree(obj);
  if (RoverDebug) 
    fprintf(stderr, "Flushed %s:%s\n", SessionTable[session].ServiceURN, url);
  Tcl_DeleteHashEntry(hentry);
  return(rc);
}

PUBLIC int Cache_ClientUpdate(RoverObj *obj)
{
  int ses = obj->session;

  if (RoverDebug) {
    fprintf(stderr, "In Cache_ClientUpdate got object:\n");
    RoverObj_print(obj); 
  }

  if (!SessionTable[ses].inuse) {
    fprintf(stderr,
	    "Invalid session ID [Cache_ClientUpdate:HTObjectClient.c]\n");
    return(TCL_ERROR);
  }
  if (RoverDebug)
    fprintf(stderr, "CLIENT: Updating %s:%s\n", SessionTable[ses].ServiceURN,
	    obj->url);

  Cache_ClientInsert(obj);

  /* Perform server cache update */
  return(SendObject(obj, ROVER_UPDATE_CACHE, ses, obj->url));
}

PUBLIC RoverObj *Cache_check(char *url, int session)
{
  Tcl_HashEntry *hentry;
  char ptr[1024];

  if (!ClientCacheInit) ObjClientInit();
  if (!SessionTable[session].inuse) {
    fprintf(stderr, "Invalid session ID [Cache_check:HTObjectClient.c]\n");
    return((RoverObj *)NULL);
  }

  sprintf(ptr, "%d::%s", session, url);
  ASSERT((strlen(ptr)+1)<=1024);
  hentry = Tcl_FindHashEntry(&ObjClientCache, ptr);
  if (hentry != NULL) return((RoverObj *)Tcl_GetHashValue(hentry));	
  return((RoverObj *)NULL);
}

PUBLIC RoverErr Cache_ClientInsert(RoverObj *obj)
{
  Tcl_HashEntry *hentry;
  RoverObj *centry = NULL;
  int new, session = obj->session;
  char *ptr, *url = obj->url;

  if (!ClientCacheInit) ObjClientInit();

  if (!SessionTable[session].inuse) {
    fprintf(stderr,
	    "Invalid session ID [Cache_ClientInsert:HTObjectClient.c]\n");
    return(ROVER_INVLD_SES);
  }
  if ((ptr = malloc(strlen(url)+32)) == NULL) {
    fprintf(stderr, "Out of memory [Cache_ClientInsert:HTObjectClient.c]\n");
    return(ROVER_NO_MEM);
  }
  sprintf(ptr, "%d::%s", session, url);
  hentry = Tcl_CreateHashEntry(&ObjClientCache, ptr, &new);
  if (!new) {				/* Check for duplicate urls */
    centry = (RoverObj *) Tcl_GetHashValue(hentry);
    if (obj != centry) {
      if (RoverDebug)
	fprintf(stderr,
		"Updating %s::%s [Cache_ClientInsert:HTObjectClient.c]\n",
		SessionTable[session].ServiceURN, url);
      
      ObjectFree(centry);   	/* Delete existing entry */
    } else
      return(ROVER_OK);
  }
  centry = obj;

  /* Save cache entry */
  Tcl_SetHashValue(hentry, (char *) centry);
  return(ROVER_OK);
}

PUBLIC RoverErr Cache_insert(RoverObj *obj)
{
  int session = obj->session;
  char *url = obj->url;

  if (!SessionTable[session].inuse) {
    fprintf(stderr, "Invalid session ID [Cache_insert:HTObjectClient.c]\n");
    return(ROVER_INVLD_SES);
  }

  Cache_ClientInsert(obj);

  /* Insert the object into the server's cache */
  return(SendObject(obj, ROVER_INSERT_CACHE, session, url));
}

PUBLIC RoverObj *Cache_fetch(char *url, int session, char *method,
			    RoverCallback callback, void *data, int *msgID)
{
  RoverObj *obj;
  char *sbuf, *sbuf2, *dv = "";
  int size, dvsize = 0;
  RoverType cmd;

  if (!ClientCacheInit) ObjClientInit();
  if (!strcmp(method, ROVER_IMPORT_CMD)) cmd = ROVER_IMPORT_OBJ;
  else if (!strcmp(method, ROVER_IMPORTDATA_CMD)) cmd = ROVER_IMPORT_DATA;
  else {
    fprintf(stderr, "Invalid method %s [Cache_fetch:HTObjectClient.c]\n", 
	    method);
    *msgID = ROVER_INVLD_METHOD;
    return NULL;	/* Make the compiler happy */
  }
  if ((obj = Cache_check(url, session)) != NULL) {
    if (RoverDebug)
      fprintf(stderr,
	      "Client cache hit %s::%s (%s)[Cache_fetch:HTObjectClient.c]\n",
	      SessionTable[session].ServiceURN, url,
	      RoverCacheType(obj->cache));
    switch (obj->cache) {
    case ROVER_NO_CACHE:	/* Do not cache */
      fprintf(stderr,
	      "Found NO_CACHE obj in cache! (%s::%s) [Cache_fetch:HTObjectClient.c]\n", 
	      SessionTable[session].ServiceURN, url);
      *msgID = ROVER_ERR;
      return NULL;	/* Make the compiler happy */
      break;

    case ROVER_CACHE_FOREVER:	/* Can be cached forever (immutable) */
      /* Away we go */
      *msgID = RoverCreateCB(session, callback, data, 1, 1);
      RoverGetCB(*msgID)->count++;
      return(obj);
      break;

    case ROVER_VERIFY:		/* Verify object before returning from cache */
    case ROVER_BEST_EFFORT:	/* Check w/ acc mgr before returning obj */
      /* Handle below */
      break;
    
    case ROVER_BE_VERIFIED:	/* Verification of client's copy of object */
    case ROVER_VERIFIED:	/* Verification of client's copy of object */
    case ROVER_CALLBACK:	/* Object registered for server callback */
    case ROVER_LEASE:		/* Object valid for specific time period */
    default:
      fprintf(stderr,
	      "Found invld/unimpl'd cache type %d (%s::%s) [Cache_fetch:HTObjectClient.c]\n",
	      (int) obj->cache, SessionTable[session].ServiceURN, url);
      *msgID = ROVER_ERR;
      return NULL;	/* Make the compiler happy */
      break;
    }
    dvsize = obj->dvsize;
    dv = (char *) obj->dv;
  }

  /* Go fetch the object */
  size = (3*sizeof(int))+strlen(url)+dvsize;
  if ((sbuf = sbuf2 = malloc(size)) == NULL) {
    fprintf(stderr, "Out of memory [Cache_fetch:HTObjectClient.c]\n");
    *msgID = ROVER_NO_MEM;
    return NULL;	/* Make the compiler happy */
  }
  
  /* Marshall up url and session number */
  MarshallString(&sbuf2, url);
  MarshallInt(&sbuf2, session);
  MarshallBin(&sbuf2, dvsize, dv);

  *msgID = RoverClientSend(cmd, session, callback, data, ROVER_NO_DELETE,
			   size, sbuf, 0);
  return((RoverObj *) NULL);
}

