 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTRover.c
 *
 * Description: Handles server-side object method invocations
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define ROVER_ACCESSMGR
#define DEFINE_CBTABLE
#define CATCH_SEGV

#include "pwd.h"
#include "sys/types.h"
#include "sys/stat.h"
#include "unistd.h"
#ifndef linux
#include "sys/uio.h"
#endif /* linux */
#include "fcntl.h"
#include "sys/time.h"
#include "signal.h"

#include "RoverBase.h"
#include "HTUtils.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTSession.h"
#include "HTObject.h"
#include "HTCache.h"
#include "HTLog.h"
#include "HTNet.h"
#include "HTRoverHTTP.h"
#include "RoverHost.h"
#include "RoverEvent.h"
#include "RoverTest.h"

/* Definitions for procedures not in the include files */
#ifndef linux
extern int putenv(char *string);
extern void bzero(char *b1, int length);
#endif /* !linux */

extern int select (int nfds, fd_set *readfds, fd_set *writefds,
		   fd_set *exceptfds, struct timeval *timeout);


PUBLIC char myAppID[32];
PUBLIC int RoverAMPid;
PUBLIC int ROVER_TraceFlag = 0;
PUBLIC int RoverArgc;
PUBLIC char **RoverArgv;
PUBLIC int AbortInProgress = 0;
PUBLIC int RoverDebug = 0;
PUBLIC char *RoverShutdownName;
PUBLIC int RoverCleanShutdown = 0;
PUBLIC int RoverCheckNet = 1;
PUBLIC int RoverCompress = 6;
PUBLIC int RoverBatch = 1;
PUBLIC char *RoverNobodyUser;
PUBLIC time_t RoverIncarnation = 0;

PRIVATE char *RoverShutdownDir;

PRIVATE void RoverDispatch(int appID, int toFD, int msgID, RoverType type,
			   int size, char *buf, char *data);
PRIVATE void RoverSetSignals();
PRIVATE void RoverUpdate(int appID, int toPipe, int watch);

#ifdef linux
#define SIG_HANDLER(NAME)	PRIVATE void NAME(int sig)
#else
#define SIG_HANDLER(NAME)	\
PRIVATE void NAME(int sig,int code,struct sigcontext *scp,char *addr)
#endif

PRIVATE char *signals[] = {
  NULL,
  "SIGHUP",
  "SIGINT",
  "SIGQUIT",
  "SIGILL",
  "SIGTRAP",
  "SIGSEGV",
  "SIGBUS",
  "SIGFPE",
  "SIGKILL",
  "SIGUSR1",
  "SIGSEGV",
  "SIGUSR2",
  "SIGPIPE",
  "SIGALRM",
  "SIGTERM",
  "SIGSTKFLT",
  "SIGCHLD",
  "SIGCONT",
  "SIGSTOP",
  "SIGTSTP",
  "SIGTTIN",
  "SIGTTOU",
  "SIGURG",
  "SIGXCPU",
  "SIGXFSZ",
  "SIGVTALRM",
  "SIGPROF",
  "SIGWINCH",
  "SIGIO",
  "SIGPWR",
  "SIGUNUSED"
};


/* Send to a ROVER client */
PUBLIC RoverErr RoverServerSend(int fd, int msgID, RoverType type, int count,
				char *buf, int dofree)
{
  RoverHdr *hdr;
  RoverErr rc;

  /* First, write the header */
  if ((hdr = (RoverHdr *) malloc(sizeof(RoverHdr))) == NULL) {
    fprintf(stderr, "Out of memory [RoverServerSend:HTRover.c]\n");
    return(ROVER_NO_MEM);
  }
  hdr->type = type;
  hdr->appID = 0;
  hdr->msgID = msgID;
  hdr->size = count;

  if (RoverDebug)
    fprintf(stderr,
	    "%s: Writing %d (%d/%d) bytes to pipe %d [App %d] (msgID %d type %s/%d) dofree: %d\n", myAppID,
	    (int)(sizeof(RoverHdr) + count), (int)sizeof(RoverHdr), count,
	    fd, Pipe2App[fd], msgID, RoverTypeString(type), type, dofree);

  /* Enqueue and flush queue */
  /* This approach is used to atomically send the header and data */
  rc = RoverEnqueue(&(ServerWriteQ[fd]), RoverServerRcvCheck, (char *)hdr,
		    sizeof(RoverHdr), (char *)hdr, sizeof(RoverHdr), fd,
		    "AM.0: Hdr Write", 1);
  if (rc != ROVER_OK) return(rc);
  if (!count) return(rc);
  rc = RoverEnqueue(&(ServerWriteQ[fd]), RoverServerRcvCheck, buf, count,
		    buf, count, fd, "AM.0: Data Write", dofree);
  if (rc != ROVER_OK) return(rc);
  RoverFlushQ(&(ServerWriteQ[fd]));
  RoverServerRcvCheck();
  
  if (ServerWriteQ[fd] != NULL) {
    if (RoverDebug)
      fprintf(stderr, "%s: Registering for write event on %d\n", myAppID, fd);
    HTEventRegister(Qhandlers[fd], EVENT_WRITE);	  /* Register client */
  }
  return(rc);
}

/* Send an integer response back to a client application */
PUBLIC RoverErr RoverServerSendInt(int toFD, int msgID, RoverType type,
				   int val)
{
  char *rbuf, *rbuf2;

  if ((rbuf2 = rbuf = malloc(sizeof(int))) == NULL) {
    fprintf(stderr,"Out of memory! [RoverServerSendInt:HTRover.c]\n");
    return(ROVER_NO_MEM);
  }
  MarshallInt(&rbuf2, val);
  if (RoverDebug)
    fprintf(stderr,
	    "Sending %s msg with value %d to app %d (fd %d) & msgID %d\n",
	    RoverTypeString(type), val, Pipe2App[toFD], toFD, msgID);
  return(RoverServerSend(toFD, msgID, type, sizeof(int), rbuf, 1));
}


/* This procedure is called to pickup pending data from the ROVER
   client. It assumes that the server has done a select on the file
   descriptor and data is pending (it won't block or fail, but it won't
   return data either).
 */
PUBLIC void RoverServerRcv(int fromFD, int toFD, RoverRcv *Rstate)
{
  RoverRcvFSM(fromFD, toFD, RoverDispatch, Rstate);
}

PRIVATE void RoverImport(int appID, int toFD, int msgID, RoverType type,
			 char *buf)
{
  int rv, rc, session, size, dvsize;
  dependV *depV;
  char *ptr = buf, *url, *dv = NULL;
  QRPCState *state, *state2;
  RoverCacheable cache;
  char *tdv;
  unsigned long tdvsize;
  RoverSession *sentry;
  RoverObjType ctype;
  HTRequest *req;
  char *cmd = ((type == ROVER_IMPORT_DATA) ? ROVER_IMPORTDATA_CMD :
	       ROVER_IMPORT_CMD);
  
  /* First, we get the information we need */
  url = UnmarshallString(&buf);
  session = UnmarshallInt(&buf);
  depV = (dependV *) UnmarshallBin(&buf, &dvsize);
  if (dvsize) {
    if ((dv = malloc(sizeof(char)*32)) == NULL) {
      fprintf(stderr, "Out of memory! [RoverImport:HTRover.c]\n");
      rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, ROVER_NO_MEM);
      if (rv != ROVER_OK)
	fprintf(stderr, "%s: Send fail %s [RoverImport:HTRover.c]\n",
		myAppID, RoverErrType(rv));
      return;
    }
    sprintf(dv, "%u_%u", depV->ServerID, depV->ServerClock);
    dvsize = strlen(dv)+1;
    ASSERT(dvsize <= 32);
  }
  free(depV);
  free(ptr);
  if (RoverDebug) fprintf(stderr,">>>CDvsize %d CDv [%s] 0x%x\n",dvsize,dv,
			  (unsigned int) dv);
  
  if (!(sentry = RoverSessionLookup(session))) {
    fprintf(stderr,	"Got invalid ses. ID %d [RoverImport:HTRover.c]\n",
	    session);
    rc = ROVER_INVLD_SES;
    rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverImport:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
    return;
  }
  if (RoverDebug) fprintf(stderr, "Importing url |%s| on session %d (%s)\n",
			  url, session, sentry->ServiceURN);
  if ((state = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
    fprintf(stderr, "Out of memory! [RoverImport:HTRover.c]\n");
    rc = ROVER_NO_MEM;
    rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverImport:HTRover.c]\n", myAppID,
	      RoverErrType(rv));
    return;
  }
  state->appID = appID;
  state->toFD = toFD;
  state->msgID = msgID;
  state->type = type;
  state->session = session;
  state->url = url;
  state->obj = NULL;
  state->argc = 0;
  state->clientDvsize = dvsize;
  state->clientDv = dv;
  state->dvsize = 0;
  state->dv = NULL;
  state->argv = NULL;
  state->next = NULL;
  state->execute = 1;
  state->IDsent = 0;
  time(&(state->loadTime));
  state->marshallObj = NULL;
  state->marshallSize = 0;
  
  /* Check if it's here already */
  if (RoverCacheSearchName(url, session, &cache, &ctype, &tdv, &tdvsize)) {
    state->type = Object2RoverTypeString(ctype);
    /* It's here, can we use it? */
    if (RoverDebug) fprintf(stderr, "%s: Cache hit %s:%s (%s) [RoverImport:HTRover.c]\n",
			    myAppID, sentry->ServiceURN, url,
			    RoverCacheType(cache));
    
    if (cache == ROVER_BEST_EFFORT) {
      if ((RoverNetState.NetAvail == IP) ||
	  !strncasecmp(sentry->CurrentServer, "localhost", 9)) {
	/* If there's an IP network interface available OR the service
	   is locally available, verify the object before returning it.
	   Otherwise, enqueue a verify request and return the object.
	   */
	if (RoverDebug) {
	  fprintf(stderr,
		  "Queuing BE obj for immediate verify [RoverImport:HTRover.c]\n");
	}
      } else {
	/* Send off a verify request and return the object */
	if (RoverDebug) {
	  fprintf(stderr,
		  "Queuing obj for BE verify [RoverImport:HTRover.c]\n");
	}
	if ((state2 = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
	  fprintf(stderr,"Out of memory! [RoverImport:HTRover.c]\n");
	  rc = ROVER_NO_MEM;
	  rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
	  if (rv != ROVER_OK)
	    fprintf(stderr, "%s: Send fail %s [RoverImport:HTRover.c]\n",
		    myAppID, RoverErrType(rv));
	  return;
	}
	state2->appID = appID;
	state2->execute = 0;
	state2->toFD = toFD;
	state2->msgID = msgID;
	state2->type = Object2RoverTypeString(ctype);
	state2->session = session;
	state2->url = copyString(url);
	state2->obj = NULL;
	state2->argc = 0;
	state2->dv = NULL;
	state2->dvsize = 0;
	state2->clientDvsize = state->clientDvsize;
	if (!state->clientDv) 
	  state2->clientDv = NULL;
	else {
	  if ((state2->clientDv = (char *)malloc(state->clientDvsize))
	      == NULL) {
	    fprintf(stderr,"Out of memory! [RoverImport:HTRover.c]\n");
	    rc = ROVER_NO_MEM;
	    rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
	    if (rv != ROVER_OK)
	      fprintf(stderr,"%s: Send fail %s [RoverImport:HTRover.c]\n",
		      myAppID, RoverErrType(rv));
	    return;
	  }
	  memcpy(state2->clientDv, state->clientDv, state->clientDvsize);
	}
	state2->argv = NULL;
	state2->next = NULL;
	state2->IDsent = 1;
	state2->marshallObj = NULL;
	state2->marshallSize = 0;
	time(&(state2->loadTime));
	
	/* memory leak on state2 */
	rc = RoverSessionRead(session, url, cmd, NullCB, state2,
			   MIN_NET_PRIORITY);
	if (rc < 0) {
	  fprintf(stderr,
		  "Session read failed %d [RoverImport:HTRover.c]\n",
		  rc);
	  rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
	  if (rv != ROVER_OK)
	    fprintf(stderr, "%s: Send fail %s [RoverImport:HTRover.c]\n",
		    myAppID, RoverErrType(rv));
	  return;
	}
	
	state->IDsent = 1;	/* It's an import and we send it right away
				   so, no logid to send */
	req = Thread_new();
	req->context = (void *) state;	/* Save for later use */
	state->type = Object2RoverTypeString(ctype);
	state->marshallObj = RoverCacheSearchData(url, session, &size);
	state->marshallSize = size;
	(HTCallbackTable[ImportObjCB])(req, state);
	Thread_delete(req);
	return;
      }
    } else if (cache != ROVER_VERIFY) {
      /* Away, we go */
      state->IDsent = 1;	/* It's an import and we send it right away
				   so, no logid to send */
      req = Thread_new();
      req->context = (void *) state;	/* Save for later use */
      state->type = Object2RoverTypeString(ctype);
      state->marshallObj = RoverCacheSearchData(url, session, &size);
      state->marshallSize = size;
      (HTCallbackTable[ImportObjCB])(req, state);
      Thread_delete(req);
      return;
    }
  }
  /* Go fetch the object */
  time(&(state->loadTime));
  rc = RoverSessionRead(session, url, cmd, ImportObjCB, state,
		     MAX_NET_PRIORITY);
  if (rc < 0) {
    fprintf(stderr, "Session read failed: %s [RoverImport:HTRover.c]\n",
	    RoverErrType(rc));
    rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverImport:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
    return;
  }
  
  if (LogFindLogID(rc) && !state->IDsent) {
    /* Request hasn't already completed */
    state->IDsent = 1;
    rv = RoverServerSendInt(toFD, msgID, ROVER_LOGID, rc);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: LogID send fail %s [RoverImport:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
  }
}

/* Server-side dispatch procedure */
PRIVATE void RoverDispatch(int appID, int toFD, int msgID, RoverType type,
			   int size, char *buf, char *data)
{
  int i, loc, ses, rvsize, rc = 0, version, objSize, consist, authSize, appID2;
  char *url, *rbuf, *rbuf2, *ptr = buf, *session, *Svc, *auth, *name;
  char *operation;
  unsigned int server, clock;
  RoverCacheable cache;
  RoverLoc excl;
  QRPCState *state;
  RoverErr rv;

  switch(type) {
  case ROVER_HELLO:
    fprintf(stderr,
	    "%s: Old (v1) Rover client tried to contact Rover Access Manager...ignoring client\n", myAppID);
    break;

  case ROVER_HELLO_1:
    /* Determine client version number 
       Format:
       Version	Client sends	Client expects
       ===============================================================
          2	Nothing		Empty reply
	  3	Version #	RoverPort and RoverNetState.NetAvail
	  4	Version #	RoverPort, RoverNetState.NetAvail,
				  RoverCompress, and RoverBatch
	  5	Version #	RoverPort, RoverNetState.NetAvail,
				  RoverCompress, RoverBatch, client AppID,
				  and Rover AM's pid
     */
    if (!size) version = 2;
    else version  = UnmarshallInt(&buf);
    free(ptr);
    if (RoverDebug)
      fprintf(stderr, "%s: %sROVER_HELLO_1 msg from %s\n", myAppID,
	      (version < 5) ? "WARNING! - " : "",
	      ((version == 2) ? "old v2 client" :
	       ((version == 3) ? "old v3 client" : 
		((version == 4) ? "old v4 client" : "v5 client"))));
    if (version == 2) {
      /* Old client, don't send info */
      rv = RoverServerSend(toFD, msgID, ROVER_REPLY, 0, (char *)NULL, 0);
      if (rv != ROVER_OK)
	fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
		myAppID, RoverErrType(rv));

      free(ptr);
      return;
    }
    if (version == 3) {
      /* Return RoverPort and RoverNetState.NetAvail */
      rvsize = 2*sizeof(int);
    } else if (version == 4) {
      /* Return RoverPort, RoverNetState.NetAvail, RoverCompress,
	 and RoverBatch */
      rvsize = 4*sizeof(int);
    } else {
      /* Return RoverPort, RoverNetState.NetAvail, RoverCompress,
	 RoverBatch, client appID, and AM appID  */
      rvsize = 6*sizeof(int);
    }
    if ((rbuf2 = rbuf = malloc(rvsize)) == NULL) {
      fprintf(stderr,"Out of memory! [RoverDispatch:HTRover.c]\n");
      rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, ROVER_NO_MEM);
    } else {
      MarshallInt(&rbuf2, RoverPort);
      MarshallInt(&rbuf2, RoverNetState.NetAvail);
      if (version >= 4) {
	MarshallInt(&rbuf2, RoverCompress);
	MarshallInt(&rbuf2, RoverBatch);
      }
      if (version == 5) {
	if (appID == -1) appID = RoverAppPipe2AppID(toFD);
	MarshallInt(&rbuf2, appID);
	MarshallInt(&rbuf2, RoverAMPid);
      }
      rv = RoverServerSend(toFD, msgID, ROVER_REPLY, rvsize, rbuf, 1);
    }
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
	      myAppID, RoverErrType(rv));

    break;

  case ROVER_FLUSH_HTTP:
    /* buf should contain a URL */
    url = UnmarshallString(&buf);
    rc = cachedeletepage(url);
/*    cachegc(); */
    free(url);
    free(ptr);
    rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
    if (!rc) RoverUpdateHTTPAll();
    break;
    
  case ROVER_FLUSH_CACHE:
    /* buf should contain session ID and a URL */
    ses = UnmarshallInt(&buf);
      
    if (!(session = RoverSessionGetURN(ses))) {
      fprintf(stderr,"Invalid ses. ID %d [RoverDispatch:HTRover.c]\n",ses);
      rc = ROVER_INVLD_SES;
    } else {
      url = UnmarshallString(&buf);
      if (RoverDebug) 
	fprintf(stderr,"%s: Flush %d::%s %s\n", myAppID, ses, url, session);
      rc = Cache_ServerFlush(url,ses);
      free(url);
    }
    free(ptr);
    rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
    break;

  case ROVER_UPDATE_CACHE:
  case ROVER_INSERT_CACHE:
    /* buf should contain a session ID, url, server id, server clock, 
       cache type, object size, and marshalled object
     */
    ses = UnmarshallInt(&buf);
    if (!(session = RoverSessionGetURN(ses))) {
      fprintf(stderr,"Invalid ses. ID %d [RoverDispatch:HTRover.c]\n",ses);
      rc = ROVER_INVLD_SES;
    } else {
      url = UnmarshallString(&buf);
      server = (unsigned int) UnmarshallInt(&buf);
      clock = (unsigned int) UnmarshallInt(&buf);
      cache = (RoverCacheable) UnmarshallInt(&buf);
      objSize = UnmarshallInt(&buf);
      if (RoverDebug) fprintf(stderr,"%s: %s: Ses - %s/%d url - %s\n",myAppID, 
			      RoverTypeString(type), session, ses, url);
      if (type == ROVER_UPDATE_CACHE)
	rc = Cache_ServerUpdate(url, ses, buf, server, clock, objSize,cache);
      else
	rc = Cache_ServerInsert(url, ses, buf, server, clock, objSize,cache);
      free(url);
    }
    free(ptr);
    rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
    break;

  case ROVER_NEW_SESSION:
    /* buf should contain Service URN, consistency req., exclusivity req.,
       appID, username, and an authenticator
     */

    Svc = UnmarshallString(&buf);
    consist = UnmarshallInt(&buf);
    excl = (RoverLoc) UnmarshallInt(&buf);
    appID2 = UnmarshallInt(&buf);
    name = UnmarshallString(&buf);
    auth = (char *) UnmarshallBin(&buf, &authSize);
    rc = RoverSessionServerCreate(Svc, consist, excl, appID2,name, authSize,auth);
    free(Svc);
    free(ptr);
    rv = RoverServerSendInt(toFD, msgID, ROVER_NEW_SESSION, rc);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
    break;

  case ROVER_END_SESSION:
    /* buf should contain Session ID and appID */

    ses = UnmarshallInt(&buf);
    appID2 = UnmarshallInt(&buf);
    free(ptr);
    rc = RoverSessionServerRelease(ses, appID2);
    rv = RoverServerSendInt(toFD, msgID, ROVER_END_SESSION, rc);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
    break;

  case ROVER_RESET_COMMAND:
    /* buf should contain Session ID and url */

    ses = UnmarshallInt(&buf);

    if (!(session = RoverSessionGetURN(ses))) {
      fprintf(stderr, "Invalid ses. ID %d [RoverDispatch:HTRover.c]\n", ses);
      rc = ROVER_INVLD_SES;
      free(ptr);
    } else {
      url = UnmarshallString(&buf);
      free(ptr);
      if (RoverDebug)
	fprintf(stderr,"Resetting Rover server with %s for session %d:%s\n",
		  url, ses, session);

      if ((state = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
	fprintf(stderr,"Out of memory! [RoverDispatch:HTRover.c]\n");
	free(url);
	rc = ROVER_NO_MEM;
      } else {
	state->appID = appID;
	state->toFD = toFD;
	state->msgID = msgID;
	state->type = 0;
	state->session = ses;
	state->url = url;
	state->obj = NULL;
	state->clientDvsize = state->dvsize = state->argc = state->execute = 
	  state->IDsent = 0;
	state->clientDv = state->dv = NULL;
	state->argv = NULL;
	state->next = NULL;
	state->marshallObj = NULL;
	state->marshallSize = 0;
      
	rc = RoverSessionRead(ses, state->url, ROVER_RESET_CMD, NullCB, state,
			   MAX_NET_PRIORITY);
	if (rc < 0) 
	  fprintf(stderr,"Session read failed: %s [RoverDispatch:HTRover.c]\n",
		  RoverErrType(rc));
      }
    }
    rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
    break;

  case ROVER_GET_PID:
    /* buf should contain nothing */
    if ((rc = RoverAppGetPID(appID)) < 0)
      fprintf(stderr, "Got invalid appID! [RoverDispatch:HTRover.c]\n");

    rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
    free(ptr);
    break;

  case ROVER_IMPORT_DATA:
  case ROVER_IMPORT_OBJ:
    RoverImport(appID, toFD, msgID, type, buf);
    break;

  case ROVER_EXPORT_OBJ:
#ifdef ROVER_RTT
    struct timeval tv;
    struct timezone tz;
    gettimeofday(&tv, &tz);
    fprintf(stderr,"%lu.%lu [RoverDispatch:HTRover.c]\n", tv.tv_sec, tv.tv_usec);
#endif /* ROVER_RTT */

    /* First, we get the information we need */
    url = UnmarshallString(&buf);
    ses = UnmarshallInt(&buf);
    if (!(session = RoverSessionGetURN(ses))) {
      fprintf(stderr,"Got invalid ses. ID %d [RoverDispatch:HTRover.c]\n",
	      ses);
      rc = ROVER_INVLD_SES;
      free(url);
      rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
      if (rv != ROVER_OK)
	fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
		myAppID, RoverErrType(rv));
      return;
    }
    operation = UnmarshallString(&buf);
    free(ptr);
    if (RoverDebug)
      fprintf(stderr, "%s: Exporting url |%s| on session %d (%s) w/ |%s|\n",
	      myAppID, url, ses, session, operation);
    /* Export the object */
    if ((state = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
      fprintf(stderr,"Out of memory! [RoverDispatch:HTRover.c]\n");
      rc = ROVER_INVLD_SES;
      free(url);
      free(operation);
      rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
      if (rv != ROVER_OK)
	fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
		myAppID, RoverErrType(rv));
      return;
    }
    state->appID = appID;
    state->execute = 1;
    state->toFD = toFD;
    state->msgID = msgID;
    state->type = type;
    state->session = ses;
    state->url = url;
    state->obj = NULL;
    state->clientDvsize = state->dvsize = state->argc = 0;
    state->clientDv = state->dv = NULL;
    state->argv = NULL;
    state->next = NULL;
    time(&(state->loadTime));
    state->marshallObj = NULL;
    state->marshallSize = 0;
    state->IDsent = 0;

    rc = RoverSessionWrite(ses, url, operation, ROVER_EXPORT_CMD, ExportObjCB,
			state);
    if (rc < 0) {
      fprintf(stderr,"Session write fail %s [RoverDispatch:HTRover.c]\n",
	      RoverErrType(rc));
      rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
      if (rv != ROVER_OK)
	fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
		myAppID, RoverErrType(rv));
      return;
    }

    if (LogFindLogID(rc)) { 
      /* Request hasn't already completed */
      if (!state->IDsent) {
	state->IDsent = 1;
	rv = RoverServerSendInt(toFD, msgID, ROVER_LOGID, rc);
	if (rv != ROVER_OK)
	  fprintf(stderr,
		  "%s: LogID send fail %s [RoverDispatch:HTRover.c]\n",
		  myAppID, RoverErrType(rv));

      }
    }
      
#if 0
    if (LogChanged) RoverUpdateLogAll();
    LogFlush();	/* Flush log AFTER we've returned the log ID to
			   avoid a possible race condition */
#endif
    break;
  case ROVER_LOAD_APP:
    /* First, we get the information we need */
    url = UnmarshallString(&buf);
    if ((state = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
      fprintf(stderr,"Out of memory! [RoverDispatch:HTRover.c]\n");
      free(url);
      rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, ROVER_NO_MEM);
      if (rv != ROVER_OK)
	fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
		myAppID, RoverErrType(rv));
      return;
    }
    state->appID = appID;
    state->argc = UnmarshallInt(&buf);
    if (RoverDebug) fprintf(stderr, "Loading application URL %s w/ %d args\n",
			    url, state->argc);
    state->clientDvsize = state->dvsize = 0;
    state->clientDv = state->dv = NULL;
    state->toFD = toFD;
    state->msgID = msgID;
    state->type = type;
    state->session = 0;
    state->url = url;
    state->execute = 1;
    state->obj = NULL;
    state->next = NULL;
    state->IDsent = 0;

    /* Allocate space for argv[0] and NULL argv[n] */
    if (!(state->argv = (char **) malloc((state->argc+2)*sizeof(char *)))) {
      fprintf(stderr, "Out of memory [RoverDispatch:HTRover.c]\n");
      free(url);
      free(state);
      return;
    }
    state->argc++;	/* Account for image name */
    for (i = 1; i < state->argc; i++)
      state->argv[i] = UnmarshallString(&buf);
    state->argv[0] = "Rover_Application";
    state->argv[i] = NULL;
    time(&(state->loadTime));
    state->marshallObj = NULL;
    state->marshallSize = 0;
	
    free(ptr);
    rc = App_LocalExec(url, LoadAppCB, state);
    if (rc < 0) {
      fprintf(stderr,"Appl'n load fail %s [RoverDispatch:HTRover.c]\n",
	      RoverErrType(rc));
      rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
      if (rv != ROVER_OK)
	fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
		myAppID, RoverErrType(rv));
      return;
    }
    break;

  case ROVER_SEND_SMTP:
    /* First, we get the information we need: session, url, and data */
    ses = UnmarshallInt(&buf);
    url = UnmarshallString(&buf);
    rbuf = UnmarshallString(&buf);
    free(ptr);
    if (!(session = RoverSessionGetURN(ses))) {
      fprintf(stderr, "Bad session ID %d [RoverDispatch:HTRover.c]\n", ses);
      rc = ROVER_INVLD_SES;
    } else {
      if (RoverDebug)
	fprintf(stderr,	"%s: Fwd SMTP request on session %d/%s w/|%s|\n",
		myAppID, ses, session, rbuf);

      rc =  LogEnqueueSMTP(ses, url, rbuf, strlen(rbuf)+1, MAX_NET_PRIORITY);
      if (rc < 0)
	fprintf(stderr,"SMTP enq. fail: %s [RoverDispatch:HTRover.c]\n",
		RoverErrType(rc));
    }
    rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, rc);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
    break;

  case ROVER_INV:
    /* First, we get the information we need */
    ses = UnmarshallInt(&buf);
    loc = UnmarshallInt(&buf);
    url = UnmarshallString(&buf);
/*      method = UnmarshallString(&buf); */
    free(ptr);

    if ((state = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
      fprintf(stderr,"Out of memory! [RoverDispatch:HTRover.c]\n");
      free(url);
      rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, ROVER_NO_MEM);
      if (rv != ROVER_OK)
	fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
		myAppID, RoverErrType(rv));
      return;
    }
    state->appID = appID;
    state->toFD = toFD;
    state->msgID = msgID;
    state->type = type;
    state->session = ses;
    state->url = url;
    state->obj = NULL;
    state->clientDvsize = state->dvsize = state->argc = 0;
    state->clientDv = state->dv = NULL;
    state->argv = NULL;
    state->next = NULL;
    state->execute = 1;
    time(&(state->loadTime));
    state->marshallObj = NULL;
    state->marshallSize = 0;
    state->IDsent = 0;

    if (loc == ROVER_WORLD) {
      /* Send off everything to someone else */
    } else {
      /* Now we grab a copy of the object
	 (Need a locality check here or there?) */
      rc = RoverSessionRead(ses, url, ROVER_IMPORT_CMD, LocalInvCB, state,
			 MAX_NET_PRIORITY);
      /* ERROR check here */
    }
    break;

  case ROVER_SET_NET:
    /* Change net status and send callbacks to all list parties */
    if (size < sizeof(int)) {
      fprintf(stderr,
	      "Warning no net setting in msg! [RoverDispatch:HTRover.c]\n");
      return;
    }
    RoverNetState.NetAvail = UnmarshallInt(&buf);
    if (RoverDebug) fprintf(stderr, "%s: New net status is %d\n", myAppID,
			    RoverNetState.NetAvail);
    if (RoverNetState.NetAvail) {
      LogFlush();
      RoverUpdateLogAll();
    }
    RoverUpdateNetAll();
    free(ptr);
    break;

  case ROVER_SET_COMPRESS:
    /* Change compression level and send callbacks to all list parties */
    if (size < sizeof(int)) {
      fprintf(stderr,
	      "Warning no net setting in msg! [RoverDispatch:HTRover.c]\n");
      return;
    }
    RoverCompress = UnmarshallInt(&buf);
    if (RoverDebug) fprintf(stderr, "%s: New compression level is %d\n",
			    myAppID, RoverCompress);
    RoverUpdateCompressAll();
    free(ptr);
    break;

  case ROVER_SET_BATCH:
    /* Change batching status and send callbacks to all list parties */
    if (size < sizeof(int)) {
      fprintf(stderr,
	      "Warning no net setting in msg! [RoverDispatch:HTRover.c]\n");
      return;
    }
    RoverBatch = UnmarshallInt(&buf);
    if (RoverDebug) fprintf(stderr, "%s: New batching var: %d\n", myAppID,
			    RoverBatch);
    RoverUpdateAll();
    free(ptr);
    break;

  case ROVER_SET_WATCH:
    /* Change net status and send callbacks to all list parties */
    if (size < sizeof(int)) {
      fprintf(stderr, "Set watch msg too short! [RoverDispatch:HTRover.c]\n");
      return;
    }
    i =  UnmarshallInt(&buf);
    if ((rc = RoverAppSetWatch(appID, i)) != ROVER_OK)
      fprintf(stderr,"Error setting app watch info [RoverDispatch:HTRover.c]\n");
    else {
      if (RoverDebug) fprintf(stderr, "%s: New update status for %d is %d\n",
			      myAppID, appID, i);
      RoverUpdate(appID, toFD, i);
    }
    free(ptr);
    break;

  default:
    fprintf(stderr, "Unknown msg type: %d [RoverDispatch:HTRover.c]\n",
	    type);
    rv = RoverServerSendInt(toFD, msgID, ROVER_REPLY, ROVER_ERR);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverDispatch:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
    break;
  }
}

PRIVATE BOOL RoverImportObj(HTRequest *req, QRPCState *state)
{
  RoverObj *obj, *obj2;
  int toFD = state->toFD, msgID = state->msgID, size, del = 1, ls = 0;
  char *sbuf, *sbuf2;
  RoverCacheable cache;
  unsigned long dvsize;
  RoverObjType type;
  RoverErr rv;

  if (RoverDebug)
    fprintf(stderr, "%s: Reached RoverImportObj:HTRover.c\n", myAppID);
  /* Check consistency HERE */

  /* Add the client msgID or appID to the object for callback purposes? */

  if (toFD != -1) {
    if (!state->IDsent) {
      state->IDsent = 1;
      rv = RoverServerSendInt(toFD, state->msgID, ROVER_LOGID, state->logID);
      if (rv != ROVER_OK)
	fprintf(stderr,"%s: LogID send fail %s [RoverImportObj:HTRover.c]\n",
		myAppID, RoverErrType(rv));

    }

    if (state->marshallObj) {
      if (RoverDebug) {
	fprintf(stderr, ">>>HERE 5: CDvsize %d CDv [%s] 0x%x\n",
		state->clientDvsize, state->clientDv,
		(unsigned int) state->clientDv);
	if (RoverCacheSearchName(state->url, state->session, &cache, &type,
				 &sbuf2, &dvsize))
	  fprintf(stderr, ">>>HERE 5a: Dvsize %lu Dv [%s]\n", dvsize, sbuf2);
      }
      del = 0;
      /* Check dv here */
      if (state->clientDvsize &&
	  (RoverCacheSearchName(state->url, state->session, &cache, &type,
				&sbuf2, &dvsize)) &&
	  (dvsize == state->clientDvsize) &&
	  !memcmp(state->clientDv, sbuf2, dvsize)) {
	if (RoverDebug) fprintf(stderr, ">>>HERE 5b - VERIFIED\n");
	state->type = ROVER_CACHE_OBJECT;
	sbuf = "";
	size = 0;
      } else {
	if (RoverDebug) fprintf(stderr, ">>>HERE 5c - OBJECT\n");
	sbuf = state->marshallObj;
	size = state->marshallSize;
      }
      if (RoverDebug) fprintf(stderr, ">>>HERE 6\n");
    } else {
      if (RoverDebug) fprintf(stderr, ">>>HERE 7\n");
      /* Check dv here */
      obj = state->obj;
      if (state->clientDvsize && (obj->dvsize == state->clientDvsize) &&
	  !memcmp(state->clientDv, obj->dv, obj->dvsize)) {
	if (RoverDebug) fprintf(stderr, ">>>HERE 7a - VERIFIED\n");
	state->type = ROVER_CACHE_OBJECT;
	sbuf = "";
	size = 0;
	del = 0;
      } else {
	if (RoverDebug) fprintf(stderr, ">>>HERE 7 - OBJECT\n");
	if (state->type == ROVER_LOGSUFFIX) {
	  /* Client doesn't have the object, so send obj along with suffix */
	  if (RoverDebug) fprintf(stderr, "Adding object to log suffix\n");
	  state->type = ROVER_OBJECT;
	  sbuf2 = RoverCacheSearchData(state->url, state->session, &size);
	  obj2 = UnmarshallObj(&sbuf2);
	  obj2->dvsize = obj->dvsize;
	  obj2->dv = obj->dv;
	  obj2->dv[0].ServerID = obj->dv[0].ServerID;
	  obj2->dv[0].ServerClock = obj->dv[0].ServerClock;
  	  obj2->type = obj->type;
	  obj2->lsize = obj->lsize;
	  obj2->log = obj->log;
	  ls = 1;
	  obj = obj2;
	}
	size = ObjectMarshallSize(obj);
	if ((sbuf2 = sbuf = malloc(size)) == NULL) {
	  fprintf(stderr,"Out of memory! [RoverImportObj:HTRover.c]\n");
	  if (state->url) free(state->url);
	  if (state->dv) free(state->dv);
	  if (state->clientDv) free(state->clientDv);
	  free(state);
	  return(ROVER_NO_MEM);
	}
	MarshallObj(&sbuf2, obj);
	if (ls == 1) {
	  /* Free the object (delete shared fields first) */
	  obj->dv = NULL;
	  obj->log = NULL;
	  RoverObjFree(obj);
	}
      }
    }
    if (RoverDebug) fprintf(stderr, "Object %s size is %d, type %s/%d\n",
			    state->url, size, RoverTypeString(state->type),
			    state->type);
    rv = RoverServerSend(toFD, msgID, state->type, size, sbuf, del);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverImportObj:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
  }
  if (state->url) free(state->url);
  if (state->dv) free(state->dv);
  if (state->clientDv) free(state->clientDv);
  free(state);
  LogFlush();
  return(ROVER_OK);
}

PRIVATE BOOL RoverExportObj(HTRequest *req, QRPCState *state)
{
  RoverObj *obj;
  int toFD = state->toFD, msgID = state->msgID, size, del = 1;
  char *sbuf, *sbuf2;
  RoverErr rv;

  if (RoverDebug)
    fprintf(stderr, "%s: Reached RoverExportObj:HTRover.c\n", myAppID);
  if (toFD != -1) {
    if (!state->IDsent) {
      state->IDsent = 1;
      rv = RoverServerSendInt(toFD, state->msgID, ROVER_LOGID, state->logID);
      if (rv != ROVER_OK)
	fprintf(stderr, "%s: LogID send fail %s [RoverExportObj:HTRover.c]\n",
		myAppID, RoverErrType(rv));
    }
    if (state->marshallObj) {
      sbuf = state->marshallObj;
      size = state->marshallSize;
      del = 0;
    } else {
      obj = state->obj;
      size = ObjectMarshallSize(obj);
      if ((sbuf2 = sbuf = malloc(size)) == NULL) {
	fprintf(stderr,"Out of memory! [RoverExportObj:HTRover.c]\n");
	if (state->url) free(state->url);
	if (state->dv) free(state->dv);
	if (state->clientDv) free(state->clientDv);
	free(state);
	return(ROVER_NO_MEM);
      }
      MarshallObj(&sbuf2, obj);
    }
    if (RoverDebug) fprintf(stderr, "Object %s size is %d, type %s/%d\n",
			    state->url, size, RoverTypeString(state->type),
			    state->type);
    rv = RoverServerSend(toFD, msgID, state->type, size, sbuf, del);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverExportObj:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
  }
  if (state->url) free(state->url);
  if (state->dv) free(state->dv);
  if (state->clientDv) free(state->clientDv);
  free(state);
  LogFlush();
  return(ROVER_OK);
}


PRIVATE BOOL RoverLoadApp(HTRequest *req, QRPCState *state)
{
  int rc1, toFD = state->toFD, msgID = state->msgID;
  char *sbuf, *sbuf2;
  RoverErr rv;

  if (RoverDebug)
    fprintf(stderr, "%s: Reached RoverLoadApp:HTRover.c\n", myAppID);
  if (state->url) free(state->url);
  if (state->argc) {
    /* Free the arguments */
    for (rc1 = 1; rc1 < state->argc; rc1++)
      free(state->argv[rc1]);
    free(state->argv);
  }
  if (state->dv) free(state->dv);
  if (state->clientDv) free(state->clientDv);
  
  if (toFD != -1) {
    if ((sbuf2 = sbuf = malloc(sizeof(int))) == NULL) {
      fprintf(stderr,"Out of memory! [RoverLoadApp:HTRover.c]\n");
      free(state);
      return(ROVER_NO_MEM);
    }
    MarshallInt(&sbuf2, state->type);
    rv = RoverServerSend(toFD, msgID, ROVER_REPLY, sizeof(int), sbuf, 1);
    if (rv != ROVER_OK)
      fprintf(stderr, "%s: Send fail %s [RoverLoadApp:HTRover.c]\n",
	      myAppID, RoverErrType(rv));
  }
  free(state);
  return(ROVER_OK);
}

PRIVATE BOOL RoverVerify(HTRequest *req, QRPCState *state)
{
  int rc1;

  if (RoverDebug)
    fprintf(stderr, "%s: Reached RoverVerify:HTRover.c\n", myAppID);
  if (state->url) free(state->url);
  if (state->argc) {
    /* Free the arguments */
    for (rc1 = 1; rc1 < state->argc; rc1++)
      free(state->argv[rc1]);
    free(state->argv);
  }
  if (state->dv) free(state->dv);
  if (state->clientDv) free(state->clientDv);
  free(state);
  return(ROVER_OK);
}


PRIVATE BOOL RoverLocalInvocation(HTRequest *req, QRPCState *state)
{
  fprintf(stderr, "%s: WARNING! Local AM invocation not supported (%d:%s)\n",
	  myAppID, state->session, state->url);
#if 0
  int session = state->session;
  char *url = state->url;
  RoverObj *obj = state->obj;
  int toFD = state->toFD;
  int msgID = state->msgID;
  /* Check consistency HERE */

  /* Create a tcl interpreter and run it */
#endif

  if (state->url) free(state->url);
  if (state->dv) free(state->dv);
  if (state->clientDv) free(state->clientDv);
  free(state);
  return(ROVER_OK);
}

/* Cleanup and exit after an error */
PUBLIC void RoverExit(int ex)
{
  RoverErr rc;

  if (ex < 0) {
    if (AbortInProgress) {
      fprintf(stderr, "FATAL ERROR! Recursive error, access manager dying...\n");
      abort();
    }
    fprintf(stderr, "FATAL ERROR! Access manager cleaning up and dying...\n");
    AbortInProgress = 1;
    if ((rc = RoverSessionSave()) != ROVER_OK) 
      fprintf(stderr,"Session save failed during abort %s\n",RoverErrType(rc));
    if ((rc = RoverCacheSave()) != ROVER_OK) 
      fprintf(stderr, "Cache save failed during abort %s\n",RoverErrType(rc));
    LogClose();
    RoverAppCloseAll();
    abort();
  } else {
    fprintf(stderr, "Access Manager shutting down...\n");
    exit(ex);
  }
}

PUBLIC HTEventState RoverTimer(HTRequest ** request)
{
  RoverErr rc;

  if (RoverCheckNet) HTNetStatus();
  RoverSessionUpdateServers();
#if 0
  /* No need to let the slip link settle, the AM will retry any failed
   *   connection attempts
   */
  if (NetChanged && (RoverNetState.NetAvail == IP))
    sleep(5);	/* Let slip link settle */
#endif
  LogFlush();
  LogSave(1);
  if ((rc = RoverCacheSave()) != ROVER_OK) RoverExit(-1);
  if ((rc = RoverSessionSave()) != ROVER_OK) RoverExit(-1);
  RoverAppPipeFlush();
  if (NetChanged) RoverUpdateNetAll();
  NetChanged = 0;
  RoverUpdateLogAll();
#if 0
  if (HTAppCount <= 0) {	/* Everyone's gone */
    fprintf(stderr,"%s: No applications left...exiting\n", myAppID);
    if ((rc = RoverSessionSave()) != ROVER_OK) RoverExit(-1);
    if ((rc = RoverCacheSave()) != ROVER_OK) RoverExit(-1);
    LogClose();
    return(EVENT_QUIT);
  }
#endif
  return(EVENT_OK);
}

/* Called when a write to a child application fails. */
SIG_HANDLER(RoverPipeSigHandler)
{
  fprintf(stderr,"Ignoring SIGCLD/SIGPIPE. [RoverPipeSigHandler:HTRover.c]\n");
  signal(SIGPIPE, RoverPipeSigHandler);	/* Catch broken pipe signals */
}

/* Called when we have BIG problems. */
SIG_HANDLER(RoverBusSigHandler)
{
  fprintf(stderr, "Received SIGBUS. [RoverBusSigHandler:HTRover.c]\n");
  RoverExit(-1);
}


/* Called when we have BIG problems. */
SIG_HANDLER(RoverSegVSigHandler)
{
  fprintf(stderr, "Received SIGSEGV. [RoverSegvSigHandler:HTRover.c]\n");
  RoverExit(-1);
}


/* Called when we get a funny signal. */
SIG_HANDLER(RoverBaseSigHandler)
{
  fprintf(stderr,
	  "%s: WARNING: Ignoring received signal %s (%d). [RoverBaseSigHandler:HTRover.c]\n",
	  myAppID, signals[sig], sig);
  RoverSetSignals();
}


PRIVATE void RoverSetSignals()
{
  signal(SIGPIPE, RoverPipeSigHandler);	/* Catch broken pipe signals */

  /* Catch signals, so we can clean up */
#ifdef CATCH_SEGV
  signal(SIGBUS, RoverBusSigHandler);
  signal(SIGSEGV, RoverSegVSigHandler);
#endif /* CATCH_SEGV */

  /* Catch other bad signals that we ignore (except SIGABRT) */
  signal(SIGQUIT, RoverBaseSigHandler);
  signal(SIGILL, RoverBaseSigHandler);
  signal(SIGTRAP, RoverBaseSigHandler);
  signal(SIGFPE, RoverBaseSigHandler);
#ifdef sparc
  signal(SIGEMT, RoverBaseSigHandler);
  signal(SIGSYS, RoverBaseSigHandler);
  signal(SIGLOST, RoverBaseSigHandler);
#endif

}

/* 
 * Check if this was a clean shutdown
 */
PRIVATE void RoverCheckShutdown()
{
  int rc;
  struct stat sbuf;

  RoverShutdownDir =(char *)malloc(strlen(ROVER_SHUTDOWN_DIRECTORY)+
				   strlen(RoverHome)+2);
  if (RoverShutdownDir == NULL) {
    fprintf(stderr, "Out of memory [RoverCheckShutdown:HTRover.c]\n");
    RoverExit(-1);
  }
  sprintf(RoverShutdownDir, "%s/%s", RoverHome, ROVER_SHUTDOWN_DIRECTORY);
  if (stat(RoverShutdownDir, &sbuf)) {
    RoverCleanShutdown = 1;	/* Clean shutdown if no dir exists */
    fprintf(stderr, "Creating Rover shutdown directory (%s)\n",
	    RoverShutdownDir);
    rc = mkdir(RoverShutdownDir, 0777);
    if (rc) {
      fprintf(stderr,
	      "Error creating shutdown dir (%s) %d [RoverCheckShutdown:HTRover.c]\n",
	      RoverShutdownDir, rc);
      RoverExit(-1);
    }
  } else if (!S_ISDIR(sbuf.st_mode)) {
    fprintf(stderr,
	    "Shutdown dir exists, but not as dir [RoverCheckShutdown:HTRover.c]\n");
    RoverExit(-1);
  }
  RoverShutdownName = (char *) malloc(strlen(ROVER_GOOD_SHUTDOWN)+
				      strlen(RoverShutdownDir)+2);
  if (RoverShutdownName == NULL) {
    fprintf(stderr, "Out of memory [RoverCheckShutdown:HTRover.c]\n");
    RoverExit(-1);
  }
  sprintf(RoverShutdownName, "%s/%s", RoverShutdownDir, ROVER_GOOD_SHUTDOWN);
  if (!stat(RoverShutdownName, &sbuf)) {
    RoverCleanShutdown = 1;	/* Must have been a clean shutdown */
    unlink(RoverShutdownName);
  }
  if (RoverCleanShutdown) {
    char *t = ctime(&(sbuf.st_mtime));
    t[strlen(t)-1] = '\0';
    fprintf(stderr, "Clean shutdown at %s.\n", t);
    if (RoverResetMsgIDs) fprintf(stderr, "Resetting message series.\n");

  } else
    fprintf(stderr, "Bad shutdown, message series will be reset.\n");
}


PUBLIC void RoverInit(int argc, char **argv)
{
  int i;
  RoverErr rc;

  RoverAMPid = getpid();
  RoverArgc = argc;
  RoverArgv = argv;
  time(&RoverIncarnation);

  sprintf(myAppID, "AM");
  if (RoverGetHostName() == NULL) RoverExit(-1);
  if (RoverGetUserName() == NULL) RoverExit(-1);
  fprintf(stderr, "Rover Access Manager (pid %d) running for %s on host %s (incarnation: %u)\n",
	  RoverAMPid, RoverUserName, RoverHostName, (unsigned int) RoverIncarnation);
  RoverHome = getenv ("HOME");
  if (RoverHome == NULL) {
    fprintf(stderr, "Error! $HOME not set [RoverInit:HTRover.c]\n");
    RoverExit(-1);
  }

  if ((RoverNobodyUser = (char *) malloc(strlen("nobody_XXXXX")+1)) == NULL) {
    fprintf(stderr, "Out of memory [RoverInit:HTRover.c]\n");
    RoverExit(-1);
  }
  sprintf(RoverNobodyUser, "nobody_%05d", RoverPort);

  RoverCheckShutdown();
  HTNetStatus();
  fprintf(stderr,"Got %d Rover net interfaces\n", RoverNetState.numInterfaces);
  if (RoverDebug) 
    for (i = 0; i < RoverNetState.numInterfaces; i++) {
      fprintf(stderr, "Interface %d (%s): %s Type %s Metric %d\n", i,
	      RoverNetState.interfaces[i].ifName,
	      RoverNetState.interfaces[i].name,
	      ((RoverNetState.interfaces[i].type == NONE) ? "NONE" :
	       ((RoverNetState.interfaces[i].type == LOCALHOST) ? "LOCALHOST" :
		((RoverNetState.interfaces[i].type == IP) ? "IP" :
		 "<unknown>"))),
	      RoverNetState.interfaces[i].metric);
    }
  if (!RoverCheckNet) fprintf(stderr, "Automatic network updating disabled\n");
  if (RoverCompress) fprintf(stderr, "Compression enabled, level %d\n", 
			     RoverCompress);
  if (!RoverBatch) fprintf(stderr, "Automatic QRPC batching disabled\n");
  RoverSetSignals();
  if ((rc = RoverSessionInit()) != ROVER_OK) RoverExit(-1);
  if ((rc = RoverAMAppInit()) != ROVER_OK) RoverExit(-1);
  if ((rc = RoverCacheInit()) != ROVER_OK) RoverExit(-1);
  if ((rc = LogInit()) != ROVER_OK) RoverExit(-1);

  RoverConversionInit();
  if ((rc = RoverHTTPInit()) != ROVER_OK) RoverExit(-1);
}

PRIVATE void RoverUpdateCompress(int appID, int toPipe, int watch)
{
  RoverErr rv;

  rv = RoverServerSendInt(toPipe, 0, ROVER_COMPRESS_UPDATE, RoverCompress);
  if (rv != ROVER_OK)
    fprintf(stderr, "%s: Send fail %s [RoverUpdateCompress:HTRover.c]\n",
	    myAppID, RoverErrType(rv));
}

PUBLIC void RoverUpdateCompressAll()
{
  RoverAppDoUpdate(ROVER_WATCH_COMPRESS, RoverUpdateCompress);
}

PRIVATE void RoverUpdateBatch(int appID, int toPipe, int watch)
{
  RoverErr rv;

  rv = RoverServerSendInt(toPipe, 0, ROVER_BATCH_UPDATE, RoverBatch);
  if (rv != ROVER_OK)
    fprintf(stderr, "%s: Send fail %s [RoverUpdateBatch:HTRover.c]\n",
	    myAppID, RoverErrType(rv));
}

PUBLIC void RoverUpdateBatchAll()
{
  RoverAppDoUpdate(ROVER_WATCH_BATCH, RoverUpdateBatch);
}

PRIVATE void RoverUpdateNet(int appID, int toPipe, int watch)
{
  RoverErr rv;

  rv = RoverServerSendInt(toPipe, 0, ROVER_NET_UPDATE, RoverNetState.NetAvail);
  if (rv != ROVER_OK)
    fprintf(stderr, "%s: Send fail %s [RoverUpdateNet:HTRover.c]\n",
	    myAppID, RoverErrType(rv));

}

PRIVATE void RoverUpdateLog(int appID, int toPipe, int watch)
{
  char *urn, *rbuf, *rbuf2, buf[132];
  int size = 0, i, count[MAX_NET_PRIORITY+2], total = 0;
  HTNetQ *q;
  RoverErr rv;

  /* 
     The marshalled log has the following format:
        [# log records]
        [<Q size>
	 [<Log ID>
	  <Session Name>
	  <URL>
	  <Post data or blank>]
	     :
	     :
	]
	     :
	     :
   */

  size = ((MAX_NET_PRIORITY+2)*sizeof(int));	/* 1 count field for each Q */
  for (i = MAX_NET_PRIORITY+1; i >= 0; i--) {
    count[i] = 0;
    if (i == MAX_NET_PRIORITY+1)
      q = RoverSentQ;
    else
      q = RoverLogQ[i];
    while (q != NULL) {
      count[i]++;
      size += sizeof(int);	/* Log id: */
      if (!(urn = RoverSessionGetURN(q->session))) {
	fprintf(stderr,
		"Ignoring invalid ses. ID %d [RoverUpdateLog:HTRover.c]\n",
		q->session);
	continue;
      }
      size += (strlen(urn)+strlen(q->url)+(2*sizeof(int)));
      size += (6+65+sizeof(int));	/* 65 chars + 5 \b + int */
      q = q->next;
    }
    total += count[i];
  }

  size += sizeof(int);		/* One int for total */
  if ((rbuf2 = rbuf = malloc(size)) == NULL) {
    fprintf(stderr,"Out of memory! [RoverUpdate:HTRover.c]\n");
    return;
  }

  if (RoverDebug) fprintf(stderr,"%s: Sending %d log records\n",myAppID,total);
  MarshallInt(&rbuf2, total);
  for (i = MAX_NET_PRIORITY+1; i >= 0; i--) {
    MarshallInt(&rbuf2, count[i]);
    if (i == MAX_NET_PRIORITY+1) q = RoverSentQ;
    else q = RoverLogQ[i];
    while (q != NULL) {
      if (!(urn = RoverSessionGetURN(q->session))) {
	fprintf(stderr,
		"Ignoring invalid ses. ID %d [RoverUpdateLog:HTRover.c]\n",
		q->session);
	continue;
      }
      MarshallInt(&rbuf2, q->logID);
      MarshallString(&rbuf2, urn);
      MarshallString(&rbuf2, q->url);
      strcpy(buf, "     ");
#if 0
      if (q->req && q->req->post_content) strncat(buf,q->req->post_content,65);
#else
      if (q->command) strncat(buf, q->command, 65);
#endif
      MarshallString(&rbuf2, buf);
      q = q->next;
    }
  }
  ASSERT(size>=(rbuf2-rbuf));
  rv = RoverServerSend(toPipe, 0, ROVER_LOG_UPDATE, size, rbuf,1);
  if (rv != ROVER_OK)
    fprintf(stderr, "%s: Send fail %s [RoverUpdateLog:HTRover.c]\n",
	    myAppID, RoverErrType(rv));

}

PRIVATE void RoverUpdateCache(int appID, int toPipe, int watch)
{
}

PRIVATE void RoverUpdate(int appID, int toPipe, int watch)
{
  if (watch & ROVER_WATCH_NET) RoverUpdateNet(appID, toPipe, watch);
  if (watch & ROVER_WATCH_APPS) RoverUpdateApps(appID, toPipe, watch);
  if (watch & ROVER_WATCH_LOG) RoverUpdateLog(appID, toPipe, watch);
  if (watch & ROVER_WATCH_CACHE) RoverUpdateCache(appID, toPipe, watch);
  if (watch & ROVER_WATCH_HTTP) RoverUpdateHTTP(appID, toPipe, watch);
}

PUBLIC void RoverUpdateAll()
{
  int flag = (ROVER_WATCH_NET | ROVER_WATCH_APPS | ROVER_WATCH_LOG |
	      ROVER_WATCH_CACHE | ROVER_WATCH_HTTP);

  RoverAppDoUpdate(flag, RoverUpdate);
}

PUBLIC void RoverUpdateCacheAll()
{
  RoverAppDoUpdate(ROVER_WATCH_CACHE, RoverUpdateCache);
}

PUBLIC void RoverUpdateNetAll()
{
  RoverAppDoUpdate(ROVER_WATCH_NET, RoverUpdateNet);
}

PUBLIC void RoverUpdateLogAll()
{
  if (!LogChanged)
    return;
  else
    LogChanged = 0;		/* Set now in case it changes during update */

  RoverAppDoUpdate(ROVER_WATCH_LOG, RoverUpdateLog);
}
