/**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTRover.h
 *
 * Description: Handles client and server side object method invocations
 *
 * Last Modified:  8-Feb-95 (adj)
 *
 * Global Functions:
 *    PUBLIC HTStream *HTAppRun(HTRequest *request,
 *				void *param,
 *				HTFormat input_format,
 *			        HTFormat output_format,
 *				HTStream *output_stream)
 *        Converter for application/x-application MIME stream. Downloads
 *        application binary code, creates pipes for IPC, and forks off
 *        a separate process.
 *
 * Global Variables: PUBLIC CONST HTStreamClass HTApplication
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#ifndef HTROVER_H
#define HTROVER_H

#include "HTList.h"
#include "RoverErr.h"

/* Macros and other defines */
#undef VL
#define VL "0.2"

#define USE_POST_METHOD

/* Macros and other defines */
extern char *RoverLibVersion;            /* Application version */

#define TclGoodEval(E)	(((E) == TCL_OK) || ((E) == TCL_RETURN))

#define ROVER_CLIENT_VERSION	5
#define INIT_TCL_URL 		"init.tcl.obj"
#define INIT_TK_URL 		"lib/tk/tk.tcl"
#define ROVER_LOG_DIRECTORY 	ROVER_CACHE_DIRECTORY
#define ROVER_SESSION_DIRECTORY	ROVER_CACHE_DIRECTORY
#define ROVER_SHUTDOWN_DIRECTORY ROVER_CACHE_DIRECTORY
#define ROVER_SESSION_INFO 	"rover-session-info"
#define ROVER_LOG 		"rover-log"
#define ROVER_SESSION_INIT 	".rover-session-init"
#define ROVER_GOOD_SHUTDOWN 	"rover-good-shutdown"
#define ROVER_MAIL_CMD		"/usr/lib/sendmail"

#define ROVER_ACCESS_METHOD	"http://"
#define ROVER_POST_TYPE		"application/x-rover"
#define ROVER_QUERY_SEP		"?"
#define ROVER_SPEC_SEP		"="
#define ROVER_ELT_SEP		"&"


typedef enum _RoverFree {
  ROVER_DELETE = 0,		/* Release space after use */
  ROVER_NO_DELETE = 1		/* Manual release */
} RoverFree;

#ifndef linux
typedef unsigned long ulong;
#endif

typedef enum _RoverLoc {
  ROVER_ANY = 0,
  ROVER_APPL = 1,
  ROVER_MH = 128,
  ROVER_WORLD = 240,
  ROVER_MAX = 256
} RoverLoc;

#define Location(E)	((E < ROVER_ANY) ? ROVER_ERR : \
			 ((E < ROVER_APPL) ? ROVER_ANY : \
			  ((E < ROVER_MH) ? ROVER_APPL : \
			   ((E < ROVER_WORLD) ? ROVER_MH : \
			    ((E < ROVER_MAX) ? ROVER_WORLD : ROVER_ERR)))))

#define LocationStr(E)	((E == ROVER_ANY) ? "ROVER_ANY" : \
			 ((E == ROVER_APPL) ? "ROVER_APPL" : \
			  ((E == ROVER_MH) ? "ROVER_MH" : \
			   ((E == ROVER_WORLD) ? "ROVER_WORLD" :"ROVER_ERR"))))

typedef enum _TCLtype {
  TCL_ONLY = 0,			/* Only start Tcl (no Tk) */
  TCL_AND_TK = 1		/* Start Tcl and Tk */
} TCLtype;

extern int RoverHTTPPrefetch;
extern Tcl_HashTable CBTable;
extern int RoverMsgID;
extern int RoverInRcvCheck;
extern RoverState me;
extern int tcl_tkSession;
extern int isClient;
extern int ROVER_TraceFlag;
extern char *RoverCacheDir;
extern int ClientInit;
extern HTAppInitializer RoverAppInit;
extern int RoverPort;
extern HTCallback HTCallbackTable[];
extern int RoverCheckNet;

extern RoverErr RoverClientInit(TCLtype type, int argc, char **argv);

#if defined(ROVER_CLIENT) || defined(ROVER_ACCESSMGR)

struct _RoverCB {
  int session;			/* Session ID */
  RoverFree del;		/* No delete upon use bit */
  RoverCallback callback;	/* The callback */
  int count;			/* Number of times executed */
  void *data;			/* Callback data area (provided at call time)*/
  RoverStatus status;
  int logID;
};
typedef struct _RoverCB RoverCB;  

struct _RoverResult {
  RoverErr rc;
  int count;
  char *result;
  RoverCallback callback;	/* The callback */
};
typedef struct _RoverResult RoverResult;

struct _AMRetVal {
  RoverType type;
  int size;
  char *buf;
  int logID;
  RoverStatus status;
};
typedef struct _AMRetVal AMRetVal;

#define ROVER_NO_SET_ARGV	0
#define ROVER_SET_ARGV		1

extern Tk_Window mainWindow;
extern Tcl_Interp *globalTclInterp;
extern char *RoverNobodyUser;
extern char *RoverShutdownName;
extern int RoverCleanShutdown;
extern int RoverCompress;
extern int RoverBatch;
extern int RoverAMPid;

extern HTEventState RoverTimer(HTRequest **request);
extern char *copyString(char *string);
extern int RoverCreateCB(int session, RoverCallback callback, void *data,
			 RoverFree del, int count);
extern RoverCB *RoverGetCB(int msgID);
extern char *RoverTypeString(RoverType type);
extern char *RoverErrType(RoverErr err);
extern char *RoverStatusType(RoverStatus status);
extern int RoverIsStatusMsg(RoverType type);
extern RoverErr RoverEnqueue(RoverQ **queue, void (*rcv)(), char *buf,
			     int size, char *ptr, int left, int tp, char *err,
			     int dofree);
extern int RoverDoSend(char *buf, int size, int tp, char *err,
		       void (*rcv)(), RoverQ **queue, int dofree);
extern int RoverFlushQ(RoverQ **queue);
extern int RoverClientSend(RoverType type, int session, RoverCallback callback,
			   void *data, RoverFree del, int count, char *buf,
			   int dofree);
extern int RoverClientCheckMsgID(int msgID, int count);
extern void RoverClientWaitMsgID(int msgID, int count);
extern void RoverServerRcv(int fromFD, int toFD, RoverRcv *Rstate);
extern void RoverClientRcvWait();
extern void RoverClientRcvCheck();
extern void RoverRcvFSM(int fromFD, int toFD, RoverCallback dispatch,
			RoverRcv *Rstate);
extern RoverErr RoverDeleteCallback(int msgID);
extern RoverErr RoverQRPC(int session, char *url, char *method,
			    int location, RoverCallback callback,
			    char *cbdata, int count, char *buf, int setArgv);
extern void RoverClientMainLoop();
extern void NullCallback(int appID, int toFD, int cbID, RoverType type,
			 int size, char *buf, char *data);
extern void RoverInit();
extern void RoverConversionInit();

extern void RoverUpdateAll();
extern void RoverUpdateNetAll();
extern void RoverUpdateAppsAll();
extern void RoverUpdateLogAll();
extern void RoverUpdateCompressAll();
extern void RoverUpdateBatchAll();
extern void RoverUpdateCacheAll();

extern RoverErr RoverServerSend(int fd, int msgID, RoverType type, int count,
				char *buf, int dofree);

extern RoverErr RoverServerSendInt(int toFD, int msgID, RoverType type,
				   int val);

#ifndef NO_MARSHALLING
#define PRIVATE static		/* Accessible only within this module */

/* Marshalling and Unmarshalling routines "borrowed" from rt_marshall.h
   and rt_unmarshall.h
   */

PRIVATE inline void MarshallChar(char **buffer, char value)
{
  /*
   * modifies: buffer
   * effects:  Encodes the char value in the buffer.
   *           Advances buffer pointer by size of encoded char.
   *
   */
  
  *((char *) *buffer) = value;
  *buffer += 1;
}

PRIVATE inline void MarshallShort(char **buffer, short value)
{
  /*
   * modifies: buffer
   * effects:  Encodes the short value in the buffer.
   *           Advances buffer pointer by size of encoded short.
   *
   */
  
  *((short *) *buffer) = value;
  *buffer += sizeof(short);
}

PRIVATE inline void MarshallInt(char **buffer, unsigned int value)
{
  /*
   * modifies: buffer
   * effects:  Encodes the integer value in the buffer.
   *           Advances buffer pointer by size of encoded integer.
   *
   */
  
#if defined(decstation) || defined(linux)
  *((int *) *buffer) = value;
#else		/* Sparcs don't like unaligned accesses */
  memcpy((void *) *buffer, (void *) (&value), (size_t) sizeof(int));
#endif /* decstation || linux */
  *buffer += sizeof(int);
}

PRIVATE inline void MarshallUlong(char **buffer, ulong value)
{
  /*
   * modifies: buffer
   * effects:  Encodes the ulong value in the buffer.
   *           Advances buffer pointer by size of encoded ulong.
   *
   */

#if defined(decstation) || defined(linux)
  *((ulong *) *buffer) = value;
#else		/* Sparcs don't like unaligned accesses */
  memcpy((void *) *buffer, (void *) (&value), (size_t) sizeof(ulong));
#endif /* decstation || linux */
  *buffer += sizeof(ulong);
}

PRIVATE inline void MarshallFloat(char **buffer, float value)
{
  /*
   * modifies: buffer
   * effects:  Encodes the float value in the buffer.
   *           Advances buffer pointer by size of encoded float.
   *
   */

#if defined(decstation) || defined(linux)
  *((float *) *buffer) = value;
#else		/* Sparcs don't like unaligned accesses */
  memcpy((void *) *buffer, (void *) (&value), (size_t) sizeof(float));
#endif /* decstation || linux */
  *buffer += sizeof(float);
}

PRIVATE inline void MarshallDouble(char **buffer, double value)
{
  /*
   * modifies: buffer
   * effects:  Encodes the double value in the buffer.
   *           Advances buffer pointer by size of encoded float.
   *
   */
  
#if defined(decstation) || defined(linux)
  *((double *) *buffer) = value;
#else		/* Sparcs don't like unaligned accesses */
  memcpy((void *) *buffer, (void *) (&value), (size_t) sizeof(double));
#endif /* decstation || linux */
  *buffer += sizeof(double);
}

PRIVATE inline void MarshallString(char **buffer, char *string)
{
  /*
   * modifies: buffer
   * effects:  Encodes the string in the buffer.
   *           Advances buffer pointer by size of encoded string.
   *
   */
  int size = strlen(string);
  MarshallInt(buffer, size);
  memcpy((void *) *buffer, (void *) string, (size_t) size);
  *buffer += size;
}

PRIVATE inline void MarshallBin(char **buffer, int size, char *string)
{
  /*
   * modifies: buffer
   * effects:  Encodes the binary data in the buffer.
   *           Advances buffer pointer by size of encoded binary data.
   *
   */
  MarshallInt(buffer, size);
  memcpy((void *) *buffer, (void *) string, (size_t) size);
  *buffer += size;
}

PRIVATE inline char UnmarshallChar(char **buffer)
{
  /*
   * modifies: buffer
   * effects:  Returns the character encoded by buffer.
   *           Advances buffer pointer by size of encoded character.
   *
   */
  
  char value = (char) **buffer;
  *buffer += 1;
  return(value);
}

PRIVATE inline int UnmarshallInt(char **buffer)
{
  /*
   * modifies: buffer
   * effects:  Returns the integer encoded by buffer.
   *           Advances buffer pointer by size of encoded integer.
   *
   */
  
#ifdef decstation
  int value = (int) *((int *) *buffer);
#else		/* Sparcs don't like unaligned accesses */
  int value;
  memcpy((void *) (&value), (void *) *buffer, (size_t) sizeof(int));
#endif /* decstation */
  *buffer += sizeof(int);
  return(value);
}

PRIVATE inline ulong UnmarshallUlong(char **buffer)
{
  /*
   * modifies: buffer
   * effects:  Returns the ulong encoded by buffer.
   *           Advances buffer pointer by size of encoded ulong.
   *
   */

#ifdef decstation  
  ulong value = (ulong) *((ulong *) *buffer);
#else		/* Sparcs don't like unaligned accesses */
  ulong value;
  memcpy((void *) (&value), (void *) *buffer, (size_t) sizeof(ulong));
#endif /* decstation */
  *buffer += sizeof(ulong);
  return(value);
}

PRIVATE inline short UnmarshallShort(char **buffer)
{
  /*
   * modifies: buffer
   * effects:  Returns the short encoded by buffer.
   *           Advances buffer pointer by size of encoded short.
   *
   */
  
  short value = (short) *((short *) *buffer);
  *buffer += sizeof(short);
  return(value);
}

PRIVATE inline float UnmarshallFloat(char **buffer)
{
  /*
   * modifies: buffer
   * effects:  Returns the float encoded by buffer.
   *           Advances buffer pointer by size of encoded float.
   *
   */
  
#ifdef decstation
  float value = (float) *((float *) *buffer);
#else		/* Sparcs don't like unaligned accesses */
  float value;
  memcpy((void *) (&value), (void *) *buffer, (size_t) sizeof(float));
#endif /* decstation */
  *buffer += sizeof(float);
  return(value);
}

PRIVATE inline double UnmarshallDouble(char **buffer)
{
  /*
   * modifies: buffer
   * effects:  Returns the double encoded by buffer.
   *           Advances buffer pointer by size of encoded double.
   *
   */

#ifdef decstation  
  double value = (double) *((double *) *buffer);
#else		/* Sparcs don't like unaligned accesses */
  double value;
  memcpy((void *) (&value), (void *) *buffer, (size_t) sizeof(double));
#endif /* decstation */
  *buffer += sizeof(double);
  return(value);
}

PRIVATE inline char *UnmarshallString(char **buffer)
{
  /*
   * modifies: buffer
   * effects:  Returns the string encoded by buffer.
   *           Advances buffer pointer by size of encoded string.
   *
   */

  char *rv;
  int size = UnmarshallInt(buffer);

  rv = (char *) malloc(size+1);
  memcpy((void *) rv, (void *) *buffer, (size_t) size);
  rv[size] = '\0';
  *buffer += size;
  return(rv);
}

PRIVATE inline char *UnmarshallBin(char **buffer, int *size)
{
  /*
   * modifies: buffer
   * effects:  Returns the binary data encoded by buffer.
   *           Advances buffer pointer by size of encoded binary data.
   *
   */

  char *rv;

  *size = UnmarshallInt(buffer);
  rv = (char *) malloc((*size)+1);
  memcpy((void *) rv, (void *) *buffer, (size_t) *size);
  rv[*size] = '\0';	/* Add a \0 for those string binary images */
  *buffer += *size;
  return(rv);
}
#endif /* NO_MARSHALLING */

#endif /* defined(ROVER_CLIENT) || defined(ROVER_ACCESSMGR) */
#endif /* HTROVER_H */
