 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTSessionClient.c
 *
 * Description: Client-side procedures for session creation, rebinding,
 * 		and consistency control
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/
#define ROVER_CLIENT

#include "sys/types.h"
#include "sys/stat.h"
#include "strings.h"
#include "fcntl.h"

#include "RoverBase.h"
#include "HTUtils.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTSession.h"


PUBLIC RoverSession SessionTable[MAXSES];

PRIVATE void RoverSessionCallback(int appID, int toFD, int msgID,
				  RoverType type, int size, char *buf,
				  char *data);


/* Create a session by binding a service URN to a particular server.
   Returns the session number or ROVER_ERR if no more sessions are available.
 */
PUBLIC RoverErr RoverSessionClientCreate(char *ServiceURN, int consistency,
				      RoverLoc exclusivity,
				      int applicationID, char *username,
				      int authSize, char *authenticator)
{
  char *buf, *buf2;
  int rc = -1;
  int msgID, size;

  /* Send a message to get our session ID */
  size = (6*sizeof(int))+strlen(ServiceURN)+strlen(username)+authSize;
  if ((buf = buf2 = malloc(size)) == NULL) {
    fprintf(stderr,
	    "Unable to allocate buffer [RoverSessionClientCreate:HTSessionClient.c]\n");
    return(ROVER_NO_MEM);
  }
  /* buf should contain Service URN, consistency req., exclusivity req.,
     and appID
   */
  MarshallString(&buf2, ServiceURN);
  MarshallInt(&buf2, consistency);
  MarshallInt(&buf2, (int) exclusivity);
  MarshallInt(&buf2, applicationID);
  MarshallString(&buf2, username);
  MarshallBin(&buf2, authSize, authenticator);
  if (RoverDebug)
    fprintf(stderr, "About to send to server...\n");
  msgID = RoverClientSend(ROVER_NEW_SESSION, -1, RoverSessionCallback,
			  (void *)(&rc), ROVER_NO_DELETE, size, buf, 1);
  if (msgID < 0) {
    fprintf(stderr,
	    "Error talking to server %d [RoverSessionClientCreate:HTSessionClient.c]\n",
	    msgID);
    return(ROVER_SEND_ERR);
  }
  if (RoverDebug)
    fprintf(stderr, "Sent to server, waiting for callback on %d\n", msgID);

  RoverClientWaitMsgID(msgID, 1);	/* Block until we get a response */
  RoverDeleteCallback(msgID);

  SessionTable[rc].inuse = 1;
  SessionTable[rc].ServiceURN = malloc(strlen(ServiceURN)+1);
  strcpy(SessionTable[rc].ServiceURN, ServiceURN);
  SessionTable[rc].CurrentServer = NULL;
  SessionTable[rc].consistency = consistency;
  SessionTable[rc].applicationID = applicationID;
  SessionTable[rc].dsize = 0;
  SessionTable[rc].username = username;
  SessionTable[rc].authSize = authSize;
  SessionTable[rc].authenticator = authenticator;
  SessionTable[rc].dv = (dependV *) NULL;
  if (RoverDebug)
    fprintf(stderr, "CLIENT: Done creating session %d\n", rc);
  return(rc);
}

/* Callback procedure for client-side session processing */
PRIVATE void RoverSessionCallback(int appID, int toFD, int msgID,
				  RoverType type, int size, char *buf,
				  char *data)
{
  RoverCB *cbentry = NULL;;

  if (RoverDebug)
    fprintf(stderr, "Session Callback Type %s/%d for msgID %d\n",
	    RoverTypeString(type), type, msgID);
  switch(type) {
  case ROVER_STATUS:
    if (!(cbentry = RoverGetCB(msgID))) {
      fprintf(stderr, "Msg ID %d not found [RoverSessionCallback:HTSessionClient.c]\n", msgID);
      return;
    }
    cbentry->status = UnmarshallInt(&buf);	/* Update the callback's logID */
    return;
    break;
  case ROVER_NEW_SESSION:
  case ROVER_END_SESSION:
  case ROVER_REPLY:
    if (size != sizeof(int)) {
      fprintf(stderr,
	      "Server response size %d != sizeof(int) [RoverSessionCallback:HTSessionClient.c]\n",
	      size);
      return;
    }
    if (!(cbentry = RoverGetCB(msgID))) {
      fprintf(stderr, "Msg ID %d not found [RoverSessionCallback:HTSessionClient.c]\n", msgID);
      return;
    }
    *((int *)(cbentry->data)) = UnmarshallInt(&buf);
    if (RoverDebug)
      fprintf(stderr, "Session callback got %d\n", *((int *)(cbentry->data)));
    break;

  default:
    fprintf(stderr,
	    "Bad ROVER AM resp %s/%d [RoverSessionCallback:HTSessionClient.c]\n",
	    RoverTypeString(type), type);
    break;
  }
}


/* Release a session */
PUBLIC RoverErr RoverSessionClientRelease(int SessionID, int applicationID)
{
  char *buf, *buf2;
  int msgID;
  int rbuf = -1, size;

  size = (2 * sizeof(int));
  if ((buf = buf2 = malloc(size)) == NULL) {
    fprintf(stderr,
	    "Unable to allocate buffer [RoverSessionClientRelease:HTSessionClient.c]\n");
    return(ROVER_NO_MEM);
  }
  /* buf should contain Session ID and AppID */
  MarshallInt(&buf2, SessionID);
  MarshallInt(&buf2, applicationID);
  msgID = RoverClientSend(ROVER_END_SESSION, SessionID, RoverSessionCallback,
			  (void *) (&rbuf), ROVER_NO_DELETE, size, buf, 1);
  if (msgID < 0) {
    fprintf(stderr,
	    "Server send error %d [RoverSessionClientRelease:HTSessionClient.c]\n",
	    msgID);
    return(ROVER_SEND_ERR);
  }

  RoverClientWaitMsgID(msgID, 1);	/* Block until we get a response */
  RoverDeleteCallback(msgID);
  
  SessionTable[SessionID].inuse = 0;
  SessionTable[SessionID].ServiceURN = NULL;
  SessionTable[SessionID].CurrentServer = NULL;
  SessionTable[SessionID].consistency = 0;
  SessionTable[SessionID].applicationID = 0;
  SessionTable[SessionID].dsize = 0;
  SessionTable[SessionID].username = NULL;
  SessionTable[SessionID].authSize = 0;
  SessionTable[SessionID].authenticator = NULL;
  SessionTable[SessionID].dv = (dependV *) NULL;
  return(rbuf);
}

/* Reset a session */
PUBLIC RoverErr RoverSessionClientReset(int SessionID, char *url)
{
  char *buf, *buf2;
  int msgID, size, rbuf = -1;

  if (RoverDebug)
    fprintf(stderr, "Sending session %d and url %s\n", SessionID, url);
/*  sleep(1); */
  size = strlen(url) + 2*sizeof(int);
  if ((buf = buf2 = malloc(size)) == NULL) {
    fprintf(stderr,
	    "Unable to allocate buffer [RoverSessionClientReset:HTSessionClient.c]\n");
    return(ROVER_NO_MEM);
  }
  /* buf should contain Session ID and url */
  MarshallInt(&buf2, SessionID);
  MarshallString(&buf2, url);
  msgID = RoverClientSend(ROVER_RESET_COMMAND, SessionID,
			  RoverSessionCallback, (void *) (&rbuf),
			  ROVER_NO_DELETE, size, buf, 1);
  if (msgID < 0) {
    fprintf(stderr,
	    "Server send error %d [RoverSessionClientReset:HTSessionClient.c]\n",
	    msgID);
    return(ROVER_SEND_ERR);
  }

/*  sleep(1); */
  RoverClientWaitMsgID(msgID, 1);	/* Block until we get a response */
  RoverDeleteCallback(msgID);
  return(rbuf);
}

