/*									 HTTP.c
**	MULTITHREADED IMPLEMENTATION OF HTTP CLIENT
**
**	(c) COPYRIGHT CERN 1994.
**	Please first read the full copyright statement in the file COPYRIGH.
**
**	This module implments the HTTP protocol as a state machine
**
** History:
**    < May 24 94 ??	Unknown - but obviously written
**	May 24 94 HF	Made reentrent and cleaned up a bit. Implemented
**			Forward, redirection, error handling and referer field
**	 8 Jul 94  FM	Insulate free() from _free structure element.
**	Jul 94 HFN	Written on top of HTTP.c, Henrik Frystyk
**
*/
/* #define ROVER_RTT */
/*
HTTP Request as per

http://www.w3.org/hypertext/WWW/Protocols/HTTP1.0/HTTP1.0-ID_19.html#HEADING29

Request         =       Simple-Request | Full-Request

Simple-Request  =       "GET" SP URI CRLF       ; HTTP/0.9 request
Full-Request    =       Request-Line            ; see Section 5.1
                        *General-Header         ; see Section 4.3
                        *Request-Header         ; see Section 5.5
                        *Object-Header          ; see Section 7
                        CRLF
                        [ Object-Body ]         ; see Section 3.2

Request-Line    =       Method SP URI SP HTTP-Version CRLF
General-Header  = Connection
                | Date
                | Forwarded
                | Mandatory
                | Message-ID
                | MIME-Version
Request-Header  = User-Agent
                | If-Modified-Since
                | Pragma
                | Authorization
                | Proxy-Authorization
                | Referer
                | From
                | Accept
                | Accept-Encoding
                | Accept-Language
Object-Header   = Allow
                | Content-Length
                | Content-Type
                | Content-Encoding
                | Content-Transfer-Encoding
                | Content-Language
                | Expires
                | Last-Modified
                | URI-header
                | Location
                | Version
                | Derived-From
                | Title
                | Link
                | extension-header
extension-header = HTTP-header
Object-Body     =       *OCTET
*/

/* Library include files */
#include "RoverBase.h"
#include "tcp.h"
#include "HTUtils.h"
#include "HTString.h"
#include "HTParse.h"
#include "HTTCP.h"
#include "HTFormat.h"
#include "HTAlert.h"
#include "HTMIME.h"
#include "HTAccess.h"		/* HTRequest */
#include "HTFWrite.h"		/* Write to cache file */
#include "HTError.h"
#include "HTThread.h"
#include "HTTP.h"					       /* Implements */
#include "zlib.h"
#include "RoverHost.h"
#include "RoverEscape.h"
#include "RoverTCP.h"

extern  HTStream* ObjectReceive ARGS5(
	HTRequest *,		request,
	void *,			param,
        HTFormat,               input_format,
        HTFormat,               output_format,
	HTStream *,		output_stream);

/* Macros and other defines */
#define HTTP_VERSION	"HTTP/1.0"
#define PUTC(c)		(*me->target->isa->put_character)(me->target, c)
#define PUTS(s)		(*me->target->isa->put_string)(me->target, s)
#define PUTBLOCK(b, l)	(*me->target->isa->put_block)(me->target, b, l)
#define FREE_TARGET	(*me->target->isa->_free)(me->target)
#define ABORT_TARGET	(*me->target->isa->abort)(me->target, e)

/* Globals */
extern char * HTAppName;		  /* Application name: please supply */
extern char * HTAppVersion;	       /* Application version: please supply */
extern BOOL using_proxy;		    /* are we using a proxy gateway? */

PUBLIC int  HTMaxRedirections = 10;	       /* Max number of redirections */
PUBLIC BOOL HTEnableFrom = NO;			      /* Enable From header? */
PUBLIC BOOL HTEnableAgent = NO;			     /* Enable Agent header? */
PUBLIC BOOL HTEnableAccept = NO;		    /* Enable Accept header? */

#ifdef ROVER
/* Temporary, broken interface to let application pass in data to post */
#ifndef MAXDIGITS
#define MAXDIGITS 9
#endif
#endif /* ROVER */

/* Type definitions and global variables etc. local to this module */
/* This is the local definition of HTRequest->net_info */
typedef enum _HTTPState {
    HTTP_ERROR		= -3,
    HTTP_NO_DATA	= -2,
    HTTP_GOT_DATA	= -1,
    HTTP_BEGIN		= 0,
    HTTP_NEED_CONNECTION,
    HTTP_NEED_REQUEST,
    HTTP_SENT_REQUEST,
    HTTP_NEED_BODY,
    HTTP_REDIRECTION,
    HTTP_AA
} HTTPState;

#define MAX_STATUS_LEN		150	   /* Max number of chars to look at */

struct _HTStream {
    CONST HTStreamClass *	isa;
    HTStream *		  	target;
    HTRequest *			request;
    HTNetInfo *			http;
    int				cnt;
    HTSocketEOL			state;
    BOOL			transparent;
    float			version;
    int				status;
    char 			buffer[MAX_STATUS_LEN+1];
    char *			bufptr;
};

/* ------------------------------------------------------------------------- */
/* 			          Help Functions			     */
/* ------------------------------------------------------------------------- */

/*                                                                  HTTPCleanup
**
**      This function closes the connection and frees memory.
**
**      Returns 0 on OK, else -1
*/
PRIVATE int HTTPCleanup ARGS1(HTRequest *, request)
{
    HTNetInfo *http = NULL;
    int status = 0;
    if (!request || !request->net_info) {
	if (PROT_TRACE) fprintf(TDEST, "HTTPCleanup. Bad argument!\n");
	status = -1;
    } else {
	http = request->net_info;
	if (http->sockfd != INVSOC) {
	    if (PROT_TRACE)
		fprintf(TDEST,"HTTP........ Closing socket %d\n",http->sockfd);
	    if ((status = NETCLOSE(http->sockfd)) < 0) {
		HTErrorSysAdd(http->request, ERR_FATAL, socerrno, NO,
			      "NETCLOSE");
		request->noretry = 1;
	    }
	    HTThreadState(http->sockfd, THD_CLOSE);
	    HTThread_clear(http);
	    http->sockfd = INVSOC;
	}
	if (http->isoc)
	    HTInputSocket_free(http->isoc);
	if (http->transmit)
	    RoverBufferFree(http->transmit);
    }	
    free(http);
    request->net_info = NULL;
    return status;
}


/*                                                              HTTPSendRequest
**
**      This function composes and sends a request to the connected server
**      specified.
**
**	Returns		<0		Error has occured or interrupted
**			HT_WOULD_BLOCK  if operation would have blocked
**			HT_INTERRUPTED  if interrupted
**
**      Note: The function NEVER closes the connection
*/
PRIVATE int HTTPSendRequest ARGS3(HTRequest *, request,
				  HTNetInfo *, http, char *, url)
{
    int status = 0;
    char *p1;
    char line[256];    /*@@@@ */
#ifdef ROVER
    extern int RoverTestPrep;
    extern void TestSaveTime();
    extern int RoverDebug;
    char *content;
    int content_length;			/* Length of post data */
    char content_length_str[MAXDIGITS];	/* String representation of length */
#ifdef ROVER_RTT
    struct timeval tv;
    struct timezone tz;
#endif ROVER_RTT
#endif /* ROVER */
  /*
   * Request-Line
   */

  /* If first time through then generate HTTP request */
    if (!http->transmit) {
	RoverBuffer *command = RoverBufferInit(2048);
	http->transmit = command;
	if (request->method != METHOD_INVALID) {
	    RoverBufferAppend(command, HTMethod_name(request->method));
	    RoverBufferAppendC(command, ' ');
	} else
	  RoverBufferAppend(command, "GET ");
	
	/* if we are using a proxy gateway don't copy in the first slash
	 ** of say: /gopher://a;lkdjfl;ajdf;lkj/;aldk/adflj
	 ** so that just gohper://.... is sent. */
	p1 = HTParse(url, "", PARSE_PATH|PARSE_PUNCTUATION);
	if (using_proxy)
	  RoverBufferAppend(command, p1+1);
	else
	  RoverBufferAppend(command, p1);
	free(p1);
	RoverBufferAppendC(command, ' ');
	RoverBufferAppend(command, HTTP_VERSION);
	RoverBufferAppendC(command, CR);		     /* CR LF, as in rfc 977 */
	RoverBufferAppendC(command, LF);
	
	/*
	 * General-Header
	 * (NULL)
	 */
	/*
	 * Request-Header
	 */
	/* If no conversion list, then put it up, but leave initialization
	   to the client */
	if (!HTConversions)
	  HTConversions = HTList_new();
	
	/* Run through both lists and generate `accept' lines */
	if (HTEnableAccept) {
	  int i;
	  HTList *conversions[2];
	  conversions[0] = HTConversions;
	  conversions[1] = request->conversions;
	  
	  for (i=0; i<2; i++) {
	    HTList *cur = conversions[i];
	    HTPresentation *pres;
	    while ((pres =(HTPresentation *) HTList_nextObject(cur))) {
	      if (pres->rep_out == WWW_PRESENT) {
		if (pres->quality != 1.0) {
		  sprintf(line, "Accept: %s; q=%.3f%c%c",
			  HTAtom_name(pres->rep),
			  pres->quality, CR, LF);
		} else {
		  sprintf(line, "Accept: %s%c%c",
			  HTAtom_name(pres->rep), CR, LF);
		}
		RoverBufferAppend(command, line);
	      }
	    }
	  }
	}
	    
	/* Put out referer field if any parent */
	if (request->parentAnchor) {
	  char *me = HTAnchor_address((HTAnchor *) request->anchor);
	  char *parent = HTAnchor_address((HTAnchor *)
					  request->parentAnchor);
	  char *relative = HTParse(parent, me,
				   PARSE_ACCESS|PARSE_HOST|PARSE_PATH|PARSE_PUNCTUATION);
	  if (relative && *relative) {
	    sprintf(line, "Referer: %s%c%c", parent, CR, LF);
	    RoverBufferAppend(command, line);
	  }
	  free(me);
	  free(parent);
	  free(relative);
	}
	
	/* Put out from field if enabled by client */
	if (HTEnableFrom) {
	  CONST char *mailaddress = RoverGetMailAddress();
	  if (mailaddress != NULL) {
	    sprintf(line, "From: %s%c%c", mailaddress, CR, LF);
	    RoverBufferAppend(command, line);
	  }
	}
	    
	/* Put out user-agent field if enabled by client */
	if (HTEnableAgent) {
	  sprintf(line, "User-Agent: %s/%s  libwww/%s%c%c",
		  HTAppName ? HTAppName : "unknown",
		  HTAppVersion ? HTAppVersion : "0.0",
		  HTLibraryVersion, CR, LF);
	  RoverBufferAppend(command, line);
	}
	
	/* Put out authorization */
	if (request->authorization != NULL) {
	  RoverBufferAppend(command, "Authorization: ");
	  RoverBufferAppend(command, request->authorization);
	  RoverBufferAppendC(command, CR);
	  RoverBufferAppendC(command, LF);
	}

#ifdef ROVER
	if (RoverTestPrep) TestSaveTime();
	/*
	 * Object-Header
	 */
	if (request->method == METHOD_POST) {
	    content = request->post_content;
	    content_length = content ? strlen(content) : 0;
	    if (request->compress) {
	        int outsize = ((content_length < 12) ? 32 : 
			       (content_length << 1));
		/* Double to allow overhead */
		char *buf = malloc(outsize);
		z_stream c_stream;	/* Compression stream */
		int rc;

		if (RoverDebug)
		  fprintf(stderr,
			  "Input size is %d, compressed output size is %d\n",
			  content_length, outsize);
	        RoverBufferAppend(command, "Compress: ");
		sprintf(content_length_str, "%d", request->compress);
		RoverBufferAppend(command, content_length_str);
		RoverBufferAppendC(command, CR);
		RoverBufferAppendC(command, LF);

	        RoverBufferAppend(command, "Length: ");
		sprintf(content_length_str, "%d", content_length);
		RoverBufferAppend(command, content_length_str);
		RoverBufferAppendC(command, CR);
		RoverBufferAppendC(command, LF);
		
		/* Compress the Object-Body before sending it */
		c_stream.zalloc = (alloc_func) 0;
		c_stream.zfree = (free_func) 0;
		c_stream.opaque = (voidpf) 0;
		rc = deflateInit(&c_stream, request->compress);
		if (rc != Z_OK) {
		  fprintf(stderr, "deflateInit error %d \n", rc);
		  free(buf);
		  return -1;
		}
		c_stream.next_in = (Bytef *)request->post_content;
		c_stream.avail_in = content_length;
		c_stream.next_out = (Bytef *)buf;
		c_stream.avail_out = outsize;
		rc = deflate(&c_stream, Z_FINISH);
		if (RoverDebug)
		  fprintf(stderr,"Deflate: Input size: %ld Output size: %ld Ratio: %f\n",
			  c_stream.total_in, c_stream.total_out, 
			  (double)(((double)c_stream.total_in)/((double)c_stream.total_out)));
		if (rc != Z_STREAM_END) {
		  fprintf(stderr, "deflate error %d \n", rc);
		  free(buf);
		  return -1;
		}
		rc = deflateEnd(&c_stream);
		if (rc != Z_OK) {
		  fprintf(stderr, "deflateEnd error %d \n", rc);
		  free(buf);
		  return -1;
		}
		content = buf;
		content_length = c_stream.total_out;
	    }
	    RoverBufferAppend(command, "Content-type: ");
	    if (request->post_content_type) {
		RoverBufferAppend(command, request->post_content_type);
	    } else {
		RoverBufferAppend(command, "none");
	    }		
	    RoverBufferAppendC(command, CR);
	    RoverBufferAppendC(command, LF);
	    sprintf(content_length_str, "%d", content_length);
	    RoverBufferAppend(command, "Content-length: ");
	    RoverBufferAppend(command, content_length_str);
	    RoverBufferAppendC(command, CR);
	    RoverBufferAppendC(command, LF);

	    /*
	     * CRLF
	     */
	    RoverBufferAppendC(command, CR);
	    RoverBufferAppendC(command, LF);

	    /*
	     * Object-Body
	     */
	    if (request->post_content) {
	        if (request->compress) {
		    int i;
		    char *ptr;
		    for (i = 0, ptr = content; i < content_length; i++, ptr++)
		      RoverBufferAppendC(command, *ptr);
		    free(content);
		} else
		    RoverBufferAppend(command, content);
	    }
	}

	if (!request->compress) {
	  RoverBufferAppendC(command, CR);	   /* Blank line means "end" */
	  RoverBufferAppendC(command, LF);
	}
#endif /* ROVER */

	RoverBufferTerminate(command);

	if (PROT_TRACE) fprintf(TDEST,"HTTP Tx... %s",
				(char *)RoverBufferBuf(command));
	
	/* Translate into ASCII if necessary */
#ifdef NOT_ASCII
	{
	    char * p;
	    for(p = (char *)RoverBufferBuf(command); *p; p++) {
		*p = TOASCII(*p);
	    }
	}
#endif
    }

    /* Now, we are ready for sending the request */
#ifdef ROVER
    if (RoverDebug) fprintf(stderr, "Sending command = \"%s\"\n",
			    (char *)RoverBufferBuf(http->transmit));
#ifdef ROVER_RTT
    fprintf(stderr, "RTT Send Size is %d\n",RoverBufferSize(http->transmit)-1);
    gettimeofday(&tv, &tz);
    fprintf(stderr,"%lu.%lu [HTTPSendRequest:HTTP.c]\n", tv.tv_sec, tv.tv_usec);
#endif /* ROVER_RTT */
#endif /* ROVER */
    status = NETWRITE(http->sockfd, (char *)RoverBufferBuf(http->transmit),
		      RoverBufferSize(http->transmit)-1);
    if (status < 0) {
      if (WouldBlock(socerrno))	{
	if (PROT_TRACE)
	  fprintf(TDEST, "HTTP Tx..... Write operation would block\n");
	HTThreadState(http->sockfd, THD_SET_WRITE);
	return HT_WOULD_BLOCK;
      } else {	    			 /* A real error has occured */
	char *unescaped = NULL;
	HTErrorSysAdd(request, ERR_FATAL, socerrno, NO, "NETWRITE");
	StrAllocCopy(unescaped, url);
	RoverUnescape(unescaped);
	HTErrorAdd(request, ERR_FATAL, NO, HTERR_INTERNAL,
		   (void *) unescaped, (int) strlen(unescaped),
		   "HTTPSendRequest");
	free(unescaped);
	return -1;
      }
    }
    HTThreadState(http->sockfd, THD_CLR_WRITE);			/* Write OK */
    return status;
}

/*
**	This is a big switch handling all HTTP return codes. It puts in any
**	appropiate error message and decides whether we should expect data
**	or not.
*/
PRIVATE void HTTPResponse ARGS1(HTStream *, me)
{
    switch (me->status) {

      case 0:						     /* 0.9 response */
      case 200:
      case 201:
      case 202:
      case 203:
      case 205:
      case 206:
	break;

      case 204:						      /* No Response */
	me->http->state = HTTP_NO_DATA;
	break;

      case 301:						   	    /* Moved */
      case 302:							    /* Found */
	me->http->state = HTTP_REDIRECTION;
	break;
	
      case 303:							   /* Method */
	HTAlert("This client doesn't support automatic redirection of type `Method'");
	me->http->state = HTTP_ERROR;
	break;
	
      case 400:						      /* Bad Request */
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_BAD_REQUEST,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;

      case 401:
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_UNAUTHORIZED,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_AA;
	break;
	
      case 402:						 /* Payment required */
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_PAYMENT_REQUIRED,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;
	
      case 403:							/* Forbidden */
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_FORBIDDEN,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;
	
      case 404:							/* Not Found */
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_NOT_FOUND,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;
	
      case 405:						      /* Not Allowed */
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_NOT_ALLOWED,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;

      case 406:						  /* None Acceptable */
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_NONE_ACCEPTABLE,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;

      case 407:			       	    /* Proxy Authentication Required */
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_PROXY,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;

      case 408:						  /* Request Timeout */
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_TIMEOUT,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;

      case 500:
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_INTERNAL,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;
	
      case 501:
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_NOT_IMPLEMENTED,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;

      case 502:
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_BAD_GATE,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;

      case 503:
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_DOWN,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;

      case 504:
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_GATE_TIMEOUT,
		   NULL, 0, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;

      default:						       /* bad number */
	HTErrorAdd(me->request, ERR_FATAL, NO, HTERR_BAD_REPLY,
		   (void *) me->buffer, me->cnt, "HTLoadHTTP");
	me->request->noretry = 1;
	me->http->state = HTTP_ERROR;
	break;
    }
}

/* ------------------------------------------------------------------------- */
/* 			    HTTP Status Line Stream			     */
/* ------------------------------------------------------------------------- */

/*
**	Analyse the string we have read. If it is a HTTP 1.0 or higher
**	then create a MIME-stream. Otherwise, abort. We need to copy the
**	buffer as we don't know if we can modify the contents or not.
**
**	Stream handling is a function of the status code returned from the 
**	server:
**		200:	 Use `output_stream' in HTRequest structure
**		else:	 Use `output_flush' in HTRequest structure
*/
PRIVATE void flush ARGS1(HTStream *, me)
{
    HTRequest *req = me->request;
    me->transparent = YES;				/* Only do this once */
    if (me->state == EOL_FLF) {
	if (strncasecmp(me->buffer, "http/", 5) ||
	    sscanf(me->buffer+5, "%f %d", &me->version, &me->status) < 2) {
	  char *msg = NULL;
	  StrAllocCopy(msg, "Can't convert from unknown format to ");
	  StrAllocCat(msg, HTAtom_name(req->output_format));
	  StrAllocCat(msg, " [Interrupted connection to server?]");
	  req->noretry = 1;
	  HTErrorAdd(req, ERR_FATAL, NO, HTERR_NOT_IMPLEMENTED,
		     (void *) msg, (int) strlen(msg), "flush/EOL_FLF");
	  free(msg);
	  me->target = ObjectReceive(req, NULL, WWW_MIME, req->output_format,
				     req->output_stream);
	} else {
	  if (me->status==200 && ((req->method==METHOD_GET) ||
				  (req->method==METHOD_POST))) {
	    me->target = HTStreamStack(WWW_MIME,req->output_format,
				       req->output_stream, req, NO);
	  } else {
	    me->target = HTMIMEConvert(req, NULL, WWW_MIME,
				       WWW_SOURCE, req->output_flush ?
				       req->output_flush : HTBlackHole());
	  }
	  if (!me->target)
	    me->target = HTBlackHole();			/* What else */
	}
    } else {
      char *msg = NULL;
      StrAllocCopy(msg, "Can't convert from unknown format to ");
      StrAllocCat(msg, HTAtom_name(req->output_format));
      StrAllocCat(msg, " [Interrupted connection to server?]");
      req->noretry = 1;
      HTErrorAdd(req, ERR_FATAL, NO, HTERR_NOT_IMPLEMENTED,
		 (void *) msg, (int) strlen(msg), "flush");
      free(msg);
      me->target = ObjectReceive(req, NULL, WWW_MIME, req->output_format,
				 req->output_stream);
    }
}

PRIVATE void HTTPStatus_put_character ARGS2(HTStream *, me, char, c)
{
    if (me->transparent)
	PUTC(c);
    else {
	if (me->state == EOL_FCR) {
	    if (c == LF) {
		me->state = EOL_FLF;			       /* Line found */
		flush(me);
	    } else {
		me->state = EOL_BEGIN;
		me->cnt += 2;
		*me->bufptr++ = CR;		      /* Need to put it back */
		*me->bufptr++ = c;
	    }
	} else if (c == CR) {
	    me->state = EOL_FCR;
	} else if (c == LF) {
	    me->state = EOL_FLF;			       /* Line found */
	    me->cnt++;
	    *me->bufptr++ = LF;
	    flush(me);
	} else {
	    me->cnt++;
	    *me->bufptr++ = c;
	    if (me->cnt >= MAX_STATUS_LEN)
		flush(me);
	}
    }
}

PRIVATE void HTTPStatus_put_string ARGS2(HTStream *, me, CONST char*, s)
{
    while (!me->transparent && *s)
	HTTPStatus_put_character(me, *s++);
    if (*s) PUTS(s);
}

PRIVATE void HTTPStatus_put_block ARGS3(HTStream *, me, CONST char*, b, int, l)
{
    while (!me->transparent && l-- > 0)
	HTTPStatus_put_character(me, *b++);
    if (l > 0) PUTBLOCK(b, l);
}

PRIVATE int HTTPStatus_free ARGS1(HTStream *, me)
{
    int status = me->status;
    HTTPResponse(me);					   /* Get next state */
    if (!me->transparent)
	flush(me);
    if (me->target)
	FREE_TARGET;
    free(me);
    return status;	     /* Return the HTTP status value - 0 if HTTP 0.9 */
}

PRIVATE int HTTPStatus_abort ARGS2(HTStream *, me, HTError, e)
{
    if (me->target)
	ABORT_TARGET;
    free(me);
    if (PROT_TRACE)
	fprintf(TDEST, "HTTPStatus.. ABORTING LOAD...\n");
    return EOF;
}

/*	HTTPStatus Stream
**	-----------------
*/
PRIVATE CONST HTStreamClass HTTPStatusClass =
{		
    "HTTPStatus",
    HTTPStatus_free,
    HTTPStatus_abort,
    HTTPStatus_put_character,
    HTTPStatus_put_string,
    HTTPStatus_put_block
};

PUBLIC HTStream * HTTPStatus_new ARGS2(HTRequest *, request,
				       HTNetInfo *, http)
{
    HTStream * me = (HTStream *) calloc(1, sizeof(HTStream));
    if (!me) outofmem(__FILE__, "HTTPStatus_new");
    me->isa = &HTTPStatusClass;
    me->request = request;
    me->http = http;
    me->bufptr = me->buffer;
    me->state = EOL_BEGIN;
    me->status = 999;
    return me;
}


PRIVATE char *HTTP_state(int state)
{
  switch (state) {
  case HTTP_BEGIN:
    return("HTTP_BEGIN");
    break;
	    
  case HTTP_NEED_CONNECTION:
    return("HTTP_NEED_CONNECTION");
    break;

  case HTTP_NEED_REQUEST:
    return("HTTP_NEED_REQUEST");
    break;

  case HTTP_SENT_REQUEST:
    return("HTTP_SENT_REQUEST");
    break;

  case HTTP_NEED_BODY:
    return("HTTP_NEED_BODY");
    break;

  case HTTP_REDIRECTION:
    return("HTTP_REDIRECTION");
    break;

  case HTTP_AA:
    return("HTTP_AA");
    break;

  case HTTP_GOT_DATA:
    return("HTTP_GOT_DATA");
    break;
	    
  case HTTP_NO_DATA:
    return(" HTTP_NO_DATA");
    break;

  case HTTP_ERROR:
    return("HTTP_ERROR");
    break;

  default:
    return("<unknown>");
  }
  return("<unknown>");
}    


/* ------------------------------------------------------------------------- */

/*		Load Document from HTTP Server			     HTLoadHTTP
**		==============================
**
**	Given a hypertext address, this routine loads a document.
**
** On entry,
**      request		This is the request structure
** On exit,
**	returns		<0		Error has occured or interrupted
**			HT_WOULD_BLOCK  if operation would have blocked
**			HT_LOADED	if return status 200 OK
**			HT_NO_DATA	if return status 204 No Response
*/
PUBLIC int HTLoadHTTP ARGS1 (HTRequest *, request)
{
    int status = HT_ERROR;
    char *url;				  /* Gets initialized on every entry */
    HTNetInfo *http;			    /* Specific protocol information */

    if (!request || !request->anchor) {
        if (PROT_TRACE) fprintf(TDEST, "HTLoadHTTP.. Bad argument\n");
        return HT_ERROR;
    }
    url = HTAnchor_physical(request->anchor);
    
    /* Only do the setup first time through. This is actually state HTTP_BEGIN
       but it can't be in the state machine as we need the structure first */
    if (!request->net_info) {
	/*
	** Initiate a new http structure and bind to request structure
	** This is actually state HTTP_BEGIN, but it can't be in the state
	** machine as we need the structure first.
	*/
	if (PROT_TRACE) fprintf(TDEST, "HTTP........ Looking for `%s\'\n", url);
	if ((http = (HTNetInfo *) calloc(1, sizeof(HTNetInfo))) == NULL)
	    outofmem(__FILE__, "HTLoadHTTP");
	http->sockfd = INVSOC;			    /* Invalid socket number */
	http->request = request;
	http->state = HTTP_BEGIN;
	request->net_info = http;
	HTThread_new(http);
    } else
	http = request->net_info;		/* Get existing copy */
 
    /* Now jump into the machine. We know the state from the previous run */
    while (1) {
      if (PROT_TRACE)
	fprintf(TDEST, "HTTP........ HTTP state for %d is %s\n", http->sockfd,
		HTTP_state(http->state));
	switch (http->state) {
	  case HTTP_BEGIN:
	    http->state = HTTP_NEED_CONNECTION;
	    break;
	    
	  case HTTP_NEED_CONNECTION: 	    /* Now let's set up a connection */
	    status = HTDoConnect(http, url, TCP_PORT, NULL, NO);
	    if (!status) {
		if (PROT_TRACE)
		    fprintf(TDEST, "HTTP........ Connected, socket %d\n",
			    http->sockfd);
		http->isoc = HTInputSocket_new(http->sockfd);
		http->state = HTTP_NEED_REQUEST;
	    } else if (status == HT_WOULD_BLOCK) {
		if (PROT_TRACE)
		    fprintf(TDEST, "HTTP........ Would block, socket %d\n",
			    http->sockfd);
		return status;
	    } else {
		if (PROT_TRACE)
		    fprintf(TDEST, "HTTP........ Error %d, socket %d\n", status,
			    http->sockfd);
		http->state = HTTP_ERROR;	       /* Error or interrupt */
	    }
	    break;

	  case HTTP_NEED_REQUEST:	  /* Compose the request and send it */
	    if ((status = HTTPSendRequest(request, http, url)) < 0) {
		if (status == HT_WOULD_BLOCK)
		    return status;
		else
		    http->state = HTTP_ERROR;
	    } else {
		http->state = HTTP_SENT_REQUEST;	    
	    }
	    break;

	  case HTTP_SENT_REQUEST:			    /* Put up stream */
	    http->target = HTTPStatus_new(request, http);
	    http->state = HTTP_NEED_BODY;
	    break;

	  case HTTP_NEED_BODY:				/* Read the response */
	    status = HTInputSocket_read(http->isoc, http->target);
	    if (status == HT_WOULD_BLOCK)
		return HT_WOULD_BLOCK;
	    else if (status == HT_INTERRUPTED) {
		(*http->target->isa->abort)(http->target, NULL);
		http->state = HTTP_ERROR;
		HTThreadState(http->isoc->input_file_number, THD_CLR_INTR);
	    } else if (status == HT_LOADED) {
		(*http->target->isa->_free)(http->target);
		if (http->state == HTTP_NEED_BODY)
		    http->state = HTTP_GOT_DATA;
	    } else {
		(*http->target->isa->_free)(http->target);
		http->state = HTTP_ERROR;
	    }
	    break;

          case HTTP_REDIRECTION:
	    if (request->redirect) {
		HTAnchor *anchor;
		if (status == 301) {
		    HTErrorAdd(request, ERR_INFO, NO, HTERR_MOVED,
			       (void *) request->redirect,
			       (int) strlen(request->redirect), "HTLoadHTTP");
		} else if (status == 302) {
		    HTErrorAdd(request, ERR_INFO, NO, HTERR_FOUND,
			       (void *) request->redirect,
			       (int) strlen(request->redirect), "HTLoadHTTP");
		}
		anchor = HTAnchor_findAddress(request->redirect);
		if (++request->redirections < HTMaxRedirections) {
#if 0
		    HTTPCleanup(request);
		    return HTLoadAnchorRecursive((HTAnchor *) anchor, request);
		} else {
#endif
		    HTErrorAdd(request, ERR_FATAL, NO, HTERR_MAX_REDIRECT,
			       NULL, 0, "HTLoadHTTP");
		    http->state = HTTP_ERROR;
		}
	    } else {
		HTErrorAdd(request, ERR_FATAL, NO, HTERR_BAD_REPLY,
			   NULL, 0, "HTLoadHTTP");
		http->state = HTTP_ERROR;
	    }
	    break;

	  case HTTP_AA:
	    {
	        char *unescaped = NULL;
		StrAllocCopy(unescaped, url);
		RoverUnescape(unescaped);
		HTErrorAdd(request, ERR_FATAL, NO, HTERR_UNAUTHORIZED,
			   (void *) unescaped,
			   (int) strlen(unescaped), "HTLoadHTTP");
		free(unescaped);
		http->state = HTTP_ERROR;
	    }
	    break;

	  case HTTP_GOT_DATA:
	    HTTPCleanup(request);
	    return HT_LOADED;
	    break;
	    
	  case HTTP_NO_DATA:
	    HTTPCleanup(request);
	    return HT_NO_DATA;
	    break;

	  case HTTP_ERROR:
	    HTTPCleanup(request);
	    return HT_ERROR;
	    break;
	}
    } /* End of while(1) */
}    

/* Protocol descriptor */

GLOBALDEF PUBLIC HTProtocol HTTP = {
    "http", SOC_NON_BLOCK, HTLoadHTTP, NULL, NULL
};

