 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: RoverClientIE.c
 *
 * Description: Handles client-side object import/export invocations
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/
#define ROVER_CLIENT

#include "RoverClient.h"
#include "RoverHost.h"
#include "RoverEscape.h"

PRIVATE char *ImportHook = NULL;
PRIVATE char *ExportHook = NULL;

PRIVATE void Eval_ExportHook();
PRIVATE void Eval_ImportHook();

PRIVATE int Rover_InvokeCallback(RoverObj *obj, char *callback, HTOper *oper, 
				 int logID, RoverType type, char *name,
				 int import, Tcl_Interp *interp)
{
  /* <Callback fcn> <Value> {<DV server> <DV time>} <Value's type> */
  char *str = "%s {%s} {%u %u} {%s}\n", *rv;
  RoverErr rc;

 /* Execute a callback */
  if (callback == NULL) return(TCL_OK);
  if (!(rv = RoverTclAlloc(strlen(callback)+strlen(name)+strlen(str)+
			   strlen(RoverTypeString(type))+32, interp)))
    return TCL_ERROR;
  sprintf(rv, str, callback, name, obj->dv[0].ServerID, obj->dv[0].ServerClock,
	  RoverTypeString(type));
  
  if (RoverDebug) fprintf(stderr, "Callback is |%s| [Rover_InvokeCallback:RoverClientIE.c]\n", rv);
  free(callback);
  rc = Tcl_GlobalEval(globalTclInterp, rv);
  free(rv);
  if (!TclGoodEval(rc)) {
    showTclError(rc, "[Rover_InvokeCallback:RoverClientIE.c]",globalTclInterp);
    return(rc);
  } else if (RoverDebug)
    fprintf(stderr, "Good callback. [Rover_InvokeCallback:RoverClientIE.c]\n");
  if (import || (logID == -1)) {	/* Imports don't have oper. lists */
    if (oper) {
      free(oper->url);
      free(oper);
    }
    return(TCL_OK);
  }

  if (!oper) {
    /* Search for the operation in the object's operations list */
    oper = obj->oper;
    while (oper) {
      if (logID == oper->logID) break;
      oper = oper->next;
    }
  } else if (logID != oper->logID) {
    /* Verify that the operation's log ID matches the specified one */
    oper = NULL;
  }
  if (oper) {
    obj->operCount--;
    if (!oper->prev) obj->oper = oper->next;
    else oper->prev->next = oper->next;
    if (oper->next) oper->next->prev = oper->prev;
    free(oper->url);
    free(oper);
    return(TCL_OK);
  }

  fprintf(stderr,
	  "Op log ID (%d) not found [Rover_InvokeCallback:RoverClientIE.c].\n",
	  logID);
  return(TCL_ERROR);
}

PRIVATE int RoverMarshallObject(char *name, Tcl_Interp *interp)
{
  char *buf, *err;
  int rc;

  if ((buf = RoverTclAlloc(strlen(name)+16, interp)) == NULL) return TCL_ERROR;
  sprintf(buf, "%s marshall\n", name);
  rc = Tcl_GlobalEval(interp, buf);
  if (!TclGoodEval(rc)) {
    showTclError(rc, "[RoverMarshallObject:RoverClientIE.c]", interp);
    if ((err = RoverAlloc(strlen(name)+64)) == NULL) return TCL_ERROR;
    sprintf(err, "\n\t(\"%s marshall\" body line %d)",name, interp->errorLine);
    Tcl_AddErrorInfo(interp, err);
    free(err);
    return(rc);
  } else
    if (RoverDebug) 
      fprintf(stderr, "Object data successfully marshalled, result is %s\n",
	      interp->result);
  free(buf);
  return(TCL_OK);
}
  
PRIVATE int RoverApplyOperation(char *operation, Tcl_Interp *interp)
{
  char *err;
  int rc;

  struct timeval ap_tv;
  struct timezone tz;
  gettimeofday(&ap_tv, &tz);
/*  fprintf(stderr,"%lu.%lu [RoverApplyOperation:HTTRoverClient.c]\n", ap_tv.tv_sec, ap_tv.tv_usec); */
  /* First, perform the operation */
  if (RoverDebug)
    fprintf(stderr, "Applying op `%s' [RoverApplyOperation:RoverClientIE.c]\n",
	    operation);
  rc = Tcl_GlobalEval(interp, operation);
  if (!TclGoodEval(rc)) {
    showTclError(rc, "[RoverApplyOperation:RoverClientIE.c]", interp);
    if ((err = RoverTclAlloc(strlen(operation)+32, interp)) == NULL)
      return TCL_ERROR;
    sprintf(err, "\n\t(\"%s\" body line %d)", operation, interp->errorLine);
    Tcl_AddErrorInfo(interp, err);
    free(err);
    return(rc);
  }
  if (RoverDebug) 
    fprintf(stderr,
	    "Oper. success ==> `%s' [RoverApplyOperation:RoverClientIE.c]\n",
	    interp->result);
  gettimeofday(&ap_tv, &tz);
/*  fprintf(stderr,"%lu.%lu [RoverApplyOperation:HTTRoverClient.c]\n", ap_tv.tv_sec, ap_tv.tv_usec);*/
  return TCL_OK;
}

PRIVATE int RoverApplyOps(RoverObj *obj, char *name, Tcl_Interp *interp)
{
  HTOper *op, *next;
  int rc;

  if (!obj->operCount) {
    obj->oper = NULL;
    if (obj->tdsize) {
      free(obj->tdata);
      obj->tdsize = 0;
      obj->tdata = NULL;
    }
  } else {
    op = obj->oper;
    while (op) {
      next = op->next;
      rc = RoverApplyOperation(op->oper, globalTclInterp);
      if (!TclGoodEval(rc)) {
	fprintf(stderr,
		"TCL APPLYOP ERR: %s [RoverApplyOps:RoverClientIE.c]\n",
		globalTclInterp->result);
	/* Error: Invoke operation's callback and delete operation */
	if (op->callback) {
	  Rover_InvokeCallback(obj, op->callback, op, op->logID,
			       ROVER_ERROR_REPLY, name, op->import,
			       globalTclInterp);
	  /* ### Check callback status here */
	}
      }
      op = next;
    }
/*    fprintf(stderr,"%lu.%lu (marshall tent)[RoverApplyOps:HTTRoverClient.c]\n", tv.tv_sec, tv.tv_usec);*/
    /* Marshall new tentative version of object */
    rc = RoverMarshallObject(name, globalTclInterp);
    if (!TclGoodEval(rc)) {
      showTclError(rc, "[RoverApplyOps:RoverClientIE.c]", globalTclInterp);
      return(-1);
    }
    obj->tdsize = strlen(globalTclInterp->result)+1;
    if ((obj->tdata = RoverCopy(globalTclInterp->result)) == NULL) return(-1);
  }
  return(0);
}  

PRIVATE char *RoverLoadObject(RoverObj *obj, int msgID, int code, 
			      int tentative, Tcl_Interp *interp)
{
  char *objData, space[64], *Ldata = NULL, *rv;
  RoverErr rc;
  int retVal, hasCode, hasData, hasUnmarshall, hasName;
  RoverCB *cbentry;

  hasCode = (obj->code && (strlen(obj->code) != 0));
  hasData = (obj->dsize || (tentative && obj->tdsize &&
			    (SessionTable[obj->session].consistency &
			     ROVER_ACCEPT_TENT)));
  hasUnmarshall = (obj->unmarshall && strlen(obj->unmarshall) != 0);
  hasName = (obj->name && strlen(obj->name) != 0);
    
  /* First, we load the code */
  if (code) {
    if (!hasCode) {
      if (!hasUnmarshall) {
	if (hasData) fprintf(stderr, "WARNING! Object has data, but no code or unmarshall procedure [RoverLoadObject:RoverClientIE.c]\n");
	obj->unmarshall = RoverAlloc(0);
      }
    } else {
      retVal = Tcl_GlobalEval(globalTclInterp, obj->code);
      if (!TclGoodEval(retVal)) {
      showTclError(retVal, "(Code load) [RoverLoadObject:RoverClientIE.c]",
		   globalTclInterp);
      } else if (RoverDebug)
	fprintf(stderr,"Code loaded [RoverLoadObject:RoverClientIE.c]\n");
      if (strlen(globalTclInterp->result) == 0) {
	/* Object doesn't want to load data? */
	if (!hasUnmarshall) {
	  if (hasData) fprintf(stderr,"WARNING! Obj code load ==> null, unmarshall is 0, data != null [RoverLoadObject:RoverClientIE.c]\n");
	  obj->unmarshall = RoverAlloc(0);
	}
      } else {
	/* Save the unmarshall function */
	if (obj->unmarshall) free(obj->unmarshall);
	obj->unmarshall = RoverCopy(globalTclInterp->result);
	hasUnmarshall = 1;
      }
    }
  }

  /* Now we can insert the object into the interpreter */
  if (!hasUnmarshall) return("");
    
  if ((Ldata = RoverAlloc(strlen("Rover_Unmarshall_Arg_")+
			  strlen(obj->unmarshall)+32)) == NULL) return(NULL);
  sprintf(Ldata,"%s $Rover_Unmarshall_%d\n", obj->unmarshall, msgID);
  sprintf(space, "Rover_Unmarshall_%d", msgID);
  
  /* Check if we should use tentative data (only if the session's consistency
     accepts tentative data and the tentative variable is set) */
  if (obj->data) 
    objData = obj->data;
  else
    objData = "";
  if (!(cbentry = RoverGetCB(msgID))) {
    fprintf(stderr, "Msg ID %d not found [RoverLoadObject:RoverClientIE.c]\n",
	    msgID);
    if (Ldata) free(Ldata);
    return(NULL);
  }
  if (tentative && obj->tdsize && (SessionTable[cbentry->session].consistency &
				   ROVER_ACCEPT_TENT)) {
    objData = obj->tdata;
  }
  rv = Tcl_SetVar(interp, space, objData, TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
  if (rv == NULL) {
    fprintf(stderr,"TCL SETVAR ERR: %s [RoverLoadObject:RoverClientIE.c]\n",
	    interp->result);
  }
  if (RoverDebug) fprintf(stderr, "Unmarshall operation is |%s|\n", Ldata);
  rc = Tcl_GlobalEval(interp, Ldata);
  if (!TclGoodEval(rc)) {
      showTclError(rc, "[RoverLoadObject:RoverClientIE.c]", interp);
      return(NULL);
  } else
    if (RoverDebug)
      fprintf(stderr,"Object data successfully unmarshalled, result is `%s'\n",
	      interp->result);
  free(Ldata);
  /* Save return value */
  if ((rv = RoverCopy(interp->result)) == NULL) return(NULL);
  if (hasName && strcmp(rv, obj->name))
    fprintf(stderr, "Warning! New name: %s doesn't match existing name %s  [RoverLoadObject:RoverClientIE.c]\n",
	    rv, obj->name);
  if (obj->name) free(obj->name);
  obj->name = rv;
  retVal = Tcl_UnsetVar(interp, space, TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
  if (!TclGoodEval(retVal)) {
    fprintf(stderr,"TCL UNSETVAR ERR: %s [RoverLoadObject:RoverClientIE.c]\n",
	    interp->result);
  }
  return(rv);
}

/* 
 * Apply a log suffix, including indicated tentative operations,
 *   invoking callbacks if operations are committed or fail
 */
PRIVATE int ApplyLogSuffix(RoverObj *obj, RoverObj *lbobj, HTOper *oper,
			   char *name, RoverType type, int msgID)
{
  char *lbuf, *idx, host[256], *result = NULL;
  RoverErr rc;
  int i, logID, status, oplen, lcount, local, timestamp, oplogID;
  RoverType cbtype;

  oplogID = oper->logID;
  lbuf = lbobj->log;
  if (!*lbuf || !lbobj->lsize) return(0);	/* No log suffix to apply */

  if (RoverDebug) fprintf(stderr, "Log suffix is %s\n", lbuf);
  if (!(idx = index(lbuf, '\n'))) {
    fprintf(stderr, "No Log suffix count [ApplyLogSuffix:RoverClientIE.c].\n");
    return(-1);
  }
  /* Set the server ID */
  obj->dv[0].ServerID = lbobj->dv[0].ServerID;
  *idx = '\0';
  lcount = atoi(lbuf);
  if (RoverDebug) fprintf(stderr, "Log suffix count is %d\n", lcount);
  lbuf = idx+1;
  while (lcount-- > 0) {
    idx = index(lbuf, '\n');
    if (!idx) {
      fprintf(stderr, "Short log suffix [ApplyLogSuffix:RoverClientIE.c].\n");
      return(-1);
    }
    *idx = '\0';
    i = sscanf(lbuf, "%d %s %d %d %d", &timestamp, host,&logID,&status,&oplen);
    if (RoverDebug)
      fprintf(stderr,
	      "timestamp = %d, host = \"%s\", logID = %d, status = %d, oplen = %d [ApplyLogSuffix:RoverClientIE.c]\n", timestamp, host, logID, status, oplen);
    /* Increment the suffix pointer */
    lbuf = idx+1;
    idx = lbuf+oplen;
    *idx = '\0';
    if (i != 5) {	/* Check the number of scanned elements */
      fprintf(stderr,
	      "Skipping short log suffix (%d != 5) [ApplyLogSuffix:RoverClientIE.c].\n", i);
      lbuf = idx+1;	/* Increment pointer to next record */
      continue;
    }
    /* The operation is local if the host matches, but we don't accept that
       for now */
    if (!strcmp(host, RoverHostName) && (oplogID == logID)) local = 1;
    else local = 0;
    if (obj->dv[0].ServerClock >= timestamp) {
      /* We've already seen this operation? */
      if (!local) {
	if (RoverDebug)
	  fprintf(stderr, "Skipping replayed log suffix record [ApplyLogSuffix:RoverClientIE.c].\n");
	lbuf = idx+1;	/* Increment pointer to next record */
	continue;
      } else 
	fprintf(stderr, "WARNING: Stale log suffix record has time %u vs. %u [ApplyLogSuffix:RoverClientIE.c].\n",
		timestamp, obj->dv[0].ServerClock);
    }
    rc = RoverApplyOperation(lbuf, globalTclInterp);
    obj->dv[0].ServerClock = timestamp;
    if (RoverDebug)
      fprintf(stderr,
	      "local = %d, logID = %d, oper->logID = %d [ApplyLogSuffix:RoverClientIE.c]\n",
	      local, logID, oplogID);
    if (local) {
      if ((result = RoverCopy(globalTclInterp->result)) == NULL) return(-1);
      /* Create an updated promise */
      i = FillPromise(msgID, oplogID, result, globalTclInterp);
      free(result);
      if (!TclGoodEval(i)) return(-1);
    }
    if (!TclGoodEval(rc)) {
      fprintf(stderr,
	      "TCL APPLYOP ERR: %s [ApplyLogSuffix:RoverClientIE.c]\n",
	      result);
      /* Error: For local operations, invoke its callback and delete it */
      cbtype = ROVER_ERROR_REPLY;
    } else {
      /* OK: Invoke operation's callback and delete operation */
      cbtype = type;
    }
    if (local) {
      if (!oper) {
	fprintf(stderr,	"Got local operation twice? [ApplyLogSuffix:RoverClientIE.c]\n");
	RoverExit(-1);
      }
      Rover_InvokeCallback(obj, oper->callback, oper, logID, cbtype, name,
			   oper->import, globalTclInterp);
      oper = NULL;	/* Freed by Rover_InvokeCallback */
      oplogID = -1;
      /* ### Check callback status here */
    }
    if (obj->deleted) {
      /* Check if this was the last operation */
      if (!obj->operCount) {
	obj->oper = NULL;
	if (obj->tdsize) {
	  free(obj->tdata);
	  obj->tdsize = 0;
	  obj->tdata = NULL;
	}
      }
      /* Should warn about (Cache_Flush puts out a warning) and
	 clean up remaining ops! */
      if ((rc = Cache_Flush(obj->url, obj->session)) < 0)
	fprintf(stderr,"Error while flushing %d::%s\n",obj->session, obj->url);
      return(-1);
    }
    lbuf = idx+1;	/* Increment pointer to next record */
  }
  return(1);
}

PRIVATE void Rover_Finish(int appID, int toFD, int msgID, RoverType type,
			  int size, char *buf, char *data)
{
  RoverObj *obj, *lbobj;
  char *ptr = buf, *name = "", *tmp;
  RoverErr rc;
  int i, import = 0;
  HTOper *oper = (HTOper *) data;
/*  struct timeval tv;
  struct timezone tz;
  gettimeofday(&tv, &tz);
  fprintf(stderr,"%lu.%lu [Rover_Finish:HTTRoverClient.c]\n", tv.tv_sec, tv.tv_usec);*/

  while (ClientInit < 0)	/* Wait until we've initialized Tcl/Tk */
    RoverClientRcvWait();

  if (RoverDebug) fprintf(stderr,"%s: type is %s (%d) [Rover_Finish:RoverClientIE.c]\n",
			  myAppID, RoverTypeString(type), type);
  
  switch (type) {

  case ROVER_STATUS:
    oper->status = UnmarshallInt(&buf);	/* Update the operation's status */
    if (RoverDebug)
      fprintf(stderr,
	      "%s: Status for logid %d is %s (%d) [Rover_Finish:RoverClientIE.c]\n",
	      myAppID, oper->logID, RoverStatusType(oper->status), oper->status);
    free(ptr); 
    CreatePromise(msgID, 0, globalTclInterp);
    break;

  case ROVER_LOGID:
    oper->logID = UnmarshallInt(&buf);	/* Update the operation's logID */
    if (RoverDebug) fprintf(stderr,"%s: logID %d [Rover_Finish:RoverClientIE.c]\n",
			    myAppID, oper->logID);
    free(ptr); 
    CreatePromise(msgID, 0, globalTclInterp);
    break;

  case ROVER_ERROR_REPLY:
  case ROVER_REPLY:
    if (type == ROVER_ERROR_REPLY) {
      obj = UnmarshallObj(&buf);
      free(ptr); 
      tmp = obj->data;
    } else {
      if ((obj = Cache_check(oper->url, oper->session)) == NULL) {
	fprintf(stderr,"No obj for %sreply [Rover_Finish:RoverClientIE.c]\n",
		((type == ROVER_ERROR_REPLY) ? "error " : ""));
	RoverExit(-1);
      }
      if (!obj) RoverExit(-1);
      i = UnmarshallInt(&buf);	/* Get error code */
      if ((tmp = RoverAlloc(32)) == NULL) RoverExit(-1);
      sprintf(tmp, "Error: %d", i);
      free(ptr);
    }
    /* Create an updated promise */
    rc = FillPromiseError(msgID, oper->logID, tmp, globalTclInterp);
    if (!TclGoodEval(rc)) {
      showTclError(rc, "[Rover_Finish:RoverClientIE.c]", globalTclInterp);
      if (oper->callback) free(oper->callback);
      free(oper->url);
      free(oper);
      if (type == ROVER_ERROR_REPLY) ObjectFree(obj);
      return;
    }
    /* Now execute the callback */
    Rover_InvokeCallback(obj, oper->callback, oper, oper->logID, type,
			 "", oper->import, globalTclInterp);
    /* ### Check callback status here */
    if (type == ROVER_ERROR_REPLY) ObjectFree(obj);
    break;

  case ROVER_LOGSUFFIX:
    import = oper->import;
    /* ### We need to add object locking here!! ### */
    if ((obj = Cache_check(oper->url, oper->session)) == NULL) {
      fprintf(stderr,"No obj for log suffix [Rover_Finish:RoverClientIE.c]\n");
      RoverExit(-1);
    }

    /* 1. Restore persistent copy of object */
    if (!obj->unmarshall || (strlen(obj->unmarshall) == 0)) {
      fprintf(stderr,"No obj unmarshall fcn [Rover_Finish:RoverClientIE.c]\n");
      RoverExit(-1);
    }
    if ((name = RoverLoadObject(obj, msgID, 0, 0, globalTclInterp)) == NULL) {
      fprintf(stderr,"Object load failed [Rover_Finish:RoverClientIE.c]\n");
      RoverExit(-1);
    }

    /* 2.  Apply log suffix, including indicated tentative operations,
           invoking callbacks if operations are committed or fail */
    lbobj = UnmarshallObj(&buf);
    if ((rc = ApplyLogSuffix(obj, lbobj, oper, name, type, msgID)) == -1) {
      free(ptr);
      ObjectFree(lbobj);
      return;
    }

    /* Only reset the time if it is larger than what we have */
    if (obj->dv[0].ServerClock < lbobj->dv[0].ServerClock) {
      /* This was already set above 
	 obj->dv[0].ServerID = lbobj->dv[0].ServerID; */
      obj->dv[0].ServerClock = lbobj->dv[0].ServerClock;
    }
    ObjectFree(lbobj);

    /* 3.  Marshall and save new persistent copy of object */
    rc = RoverMarshallObject(name, globalTclInterp);
    if (!TclGoodEval(rc)) {
      showTclError(rc, "[Rover_Finish:RoverClientIE.c]", globalTclInterp);
      return;
    }
    if (obj->data) free(obj->data);
    if ((obj->data = RoverCopy(globalTclInterp->result)) == NULL) return;
    obj->dsize = strlen(obj->data)+1;

    /* 4.  Reapply tentative operations, invoking callbacks if any
           errors are encountered */
    if ((rc = RoverApplyOps(obj, name, globalTclInterp)) == -1) {
      free(ptr);
      return;
    }

    /* 5.  Pass the marshalled object to Access Manager. */
    Cache_ClientUpdate(obj);

    /* 6. Create an updated promise for imports (updated promises for
          exports are only created if the log suffix contained an entry
          that matched the logID) */
    if (import) {
      rc = FillPromise(msgID, oper->logID, name, globalTclInterp);
      if (!TclGoodEval(rc)) {
	showTclError(rc, "[Rover_Finish:RoverClientIE.c]", globalTclInterp);
	if (oper->callback) free(oper->callback);
	free(oper->url);
	free(oper);
	return;
      }

      /* 7. Execute the import callback */
      Rover_InvokeCallback(obj, oper->callback, oper, oper->logID, type,
			   name, oper->import, globalTclInterp);
      /* ### Check callback status here */
    }

    /* Clean up */
    free(ptr);
    if (import)
      Eval_ImportHook();
    else
      Eval_ExportHook();
    break;

  case ROVER_CACHE_OBJECT:
  case ROVER_OBJECT:

    import = oper->import;
    if (RoverDebug) fprintf(stderr, "In Rover_Finish (%s), got %sobject:\n",
			    (import ? "import" : "export"),
			    (type == ROVER_CACHE_OBJECT) ? "cache " : "");

    if (type == ROVER_CACHE_OBJECT) {
      if (!import) {
	fprintf(stderr,"Rover_Finish export called w/ ROVER_CACHE_OBJECT!\n");
	RoverExit(-1);
      }
      if ((obj = Cache_check(oper->url, oper->session)) == NULL) {
	fprintf(stderr,"No obj for oper. [Rover_Finish:RoverClientIE.c]\n");
	RoverExit(-1);
      }
      if (obj->name && strlen(obj->name) != 0) {
	name = obj->name;
	goto step4;    /* Skip unnecessary steps (saving and loading object) */
      }
    } else {
      obj = UnmarshallObj(&buf);
      free(ptr);
    }

    /* if (RoverDebug) RoverObj_print(obj);  */

    /* 1. Make the new object the persistent copy */
    if (obj->cache != ROVER_NO_CACHE) {
      if (RoverDebug) fprintf(stderr, "Inserting %s::%s into cache\n",
			      SessionTable[obj->session].ServiceURN, obj->url);
      if ((rc=Cache_ClientInsert(obj))!=ROVER_OK) { /* Insert obj in cache */
	fprintf(stderr,"Cache insert failed [Rover_Finish:RoverClientIE.c]\n");
	RoverExit(-1);
      }
    }
    
    /* 2. Insert the object into the interpreter */
    if ((name = RoverLoadObject(obj, msgID, 1, 0, globalTclInterp)) == NULL) {
      fprintf(stderr,"Object load failed [Rover_Finish:RoverClientIE.c]\n");
      RoverExit(-1);
    }
    
    /* 3. Apply log suffix, including indicated tentative operations,
          invoking callbacks if operations are committed or fail */
    if ((rc = ApplyLogSuffix(obj, obj, oper, name, type, msgID)) == -1) {
      if (oper->callback) free(oper->callback);
      free(oper->url);
      free(oper);
      return;
    } else if ((rc == 1) || (obj->operCount)) { /* Log sfx or pending ops? */ 

      /* 3a.  Marshall and save new persistent copy of object */
      rc = RoverMarshallObject(name, globalTclInterp);
      if (!TclGoodEval(rc)) {
	showTclError(rc, "[Rover_Finish:RoverClientIE.c]", globalTclInterp);
	if (oper->callback) free(oper->callback);
	free(oper->url);
	free(oper);
	return;
      }
      if (obj->data) free(obj->data);
      if ((obj->data = RoverCopy(globalTclInterp->result)) == NULL) return;
      obj->dsize = strlen(obj->data)+1;
      
      /* 3b.  Reapply tentative operations, invoking callbacks if any
	        errors are encountered */
      if ((rc = RoverApplyOps(obj, name, globalTclInterp)) == -1) {
	if (oper->callback) free(oper->callback);
	free(oper->url);
	free(oper);
	return;
      }

      /* 3c.  Pass the marshalled object to Access Manager. */
      Cache_ClientUpdate(obj);
    }

    /* 4. Create an updated promise */
  step4:
    rc = FillPromise(msgID, oper->logID, name, globalTclInterp);
    if (!TclGoodEval(rc)) {
      showTclError(rc, "[Rover_Finish:RoverClientIE.c]", globalTclInterp);
      if (oper->callback) free(oper->callback);
      free(oper->url);
      free(oper);
      return;
    }

    /* 5. Execute the callback */
    Rover_InvokeCallback(obj, oper->callback, oper, oper->logID, type,
			 name, oper->import, globalTclInterp);
    /* ### Check callback status here */
    if (obj->cache == ROVER_NO_CACHE) ObjectFree(obj);

    /* Clean up */
    if (import)
      Eval_ImportHook();
    else
      Eval_ExportHook();
    break;
    
  default:
    fprintf(stderr,
	    "Unexpected AM response %s (%d) [Rover_Finish:RoverClientIE.c]\n",
	    RoverTypeString(type), type);
    return;
  }
}

/* Rover_Export <Session> <Object> <Operation> <Unmarshall Fcn> [<Callback>] */
PUBLIC int Rover_Export(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
  int msgID, session, size, rc, new = 0, opsize;
  char *op, *sbuf, *sbuf2, *url, *cbuf, *err = NULL, *operation;
  RoverObj *obj;
  HTOper *oper, *opr;
/*  struct timeval tv;
  struct timezone tz;
  gettimeofday(&tv, &tz);
  fprintf(stderr,"%lu.%lu [Rover_Export:HTTRoverClient.c]\n", tv.tv_sec, tv.tv_usec);*/

  if ((argc < 5) || (argc > 6)) {
    interp->result =
      "wrong # args: Rover_Export <Session> <Object> <Operation> <Unmarshall Fcn> [<Callback>]";
    return TCL_ERROR;
  }

  session = atoi(argv[1]);
  if (!SessionTable[session].inuse) {
    fprintf(stderr, "Invalid session ID [Rover_Export:RoverClientIE.c]\n");
    interp->result = "Invalid session ID [Rover_Export:RoverClientIE.c]";
    return TCL_ERROR;
  }
  url = argv[2];
  operation = argv[3];
  if (argc == 6) {
    /* Save the callback for later */
    if ((cbuf = RoverTclCopy(argv[5], interp)) == NULL) return TCL_ERROR;
  } else
    cbuf = NULL;

  if ((Location(session) != ROVER_APPL) && (Location(session) != ROVER_ANY) &&
      (Location(session) != ROVER_MH)) {
    rc = Location(session);
    sprintf(err, "Session %d (%s) isn't local, mh, or any!\n", session,
	    LocationStr(rc));
    Tcl_SetResult(interp, err, TCL_VOLATILE);
    return TCL_ERROR;
  }

  if (RoverDebug)
    fprintf(stderr, "Exporting |%s| on session %d (callback: %s)\n", url,
	    session, cbuf);

  /* Check if the object is already here */
  if (RoverDebug)
    fprintf(stderr, "Looking for object locally (URL: %s Session: %d, %s)\n",
	    url, session, SessionTable[session].ServiceURN);
/*  fprintf(stderr,"%lu.%lu (Cache_check)[Rover_Export:HTTRoverClient.c]\n", tv.tv_sec, tv.tv_usec);*/
  if ((obj = Cache_check(url, session)) == NULL) {
    /* Create a new object */
    if (RoverDebug)
      fprintf(stderr,"Obj not cached, new? [Rover_Export:RoverClientIE.c]\n");
    if (!(obj=(RoverObj*)RoverTclAlloc(sizeof(RoverObj), interp))) return(TCL_ERROR);
    obj->lsize = obj->tdsize = obj->csize = obj->dsize = 0;
    obj->log = obj->tdata = obj->code = obj->data = NULL;
    obj->dvsize = sizeof(dependV);
    if (!(obj->dv = (dependV *)RoverTclAlloc(sizeof(dependV),interp)))
      return(TCL_ERROR);
    obj->dv[0].ServerID = 0;
    obj->dv[0].ServerClock = 0;
    obj->interp = NULL;
    obj->cache = ROVER_BEST_EFFORT; 	/* Updated later? */
    obj->session = session;
    if ((obj->url = RoverTclCopy(url,interp)) == NULL) return TCL_ERROR;
    obj->inuse = AllFields;
    obj->deleted = obj->operCount = 0;
    obj->anchor = NULL;
    time(&(obj->lastRef));
    obj->loadDelay = obj->lastRef = 0;
    obj->refCount = 1;
    obj->type = ROVER_OBJECT;
    obj->oper = NULL;
    if (!(obj->name=RoverTclAlloc(sizeof(char),interp))) return TCL_ERROR;
    *(obj->name) = '\0';	/* No name for now */
    new = 1;
  }

  /* Save unmarshall function */
  if ((obj->unmarshall=RoverTclCopy(argv[4],interp))==NULL) return TCL_ERROR;

  if (RoverDebug) {
    fprintf(stderr, "In middle of Rover_Export, got object:\n");
    RoverObj_print(obj); 
  }

/*  fprintf(stderr,"%lu.%lu (apply op)[Rover_Export:HTTRoverClient.c]\n", tv.tv_sec, tv.tv_usec);*/
  /* First, perform the operation */
  rc = RoverApplyOperation(operation, interp);
  if (!TclGoodEval(rc)) return TCL_ERROR;	/* Result is already set */

  /* Next, marshall the object */
/*  fprintf(stderr,"%lu.%lu (marshall)[Rover_Export:HTTRoverClient.c]\n", tv.tv_sec, tv.tv_usec); */
  rc = RoverMarshallObject(url, interp);
  if (!TclGoodEval(rc)) return TCL_ERROR;	/* Result is already set */
  if (obj->tdsize) free(obj->tdata);
  obj->tdsize = strlen(interp->result)+1;
  if ((obj->tdata = RoverCopy(interp->result)) == NULL) return TCL_ERROR;

  if (!(oper=(HTOper *)RoverTclAlloc(sizeof(HTOper),interp))) return TCL_ERROR;
  if ((oper->oper = RoverCopy(operation)) == NULL) return TCL_ERROR;
  oper->status = ROVER_NOTQUEUED;
  oper->import = 0;
  oper->callback = cbuf;
  if ((oper->url = RoverTclCopy(url, interp)) == NULL) return TCL_ERROR;
  oper->session = session;
  oper->logID = -1;
  oper->size = strlen(operation)+1;
  op = RoverEscape(operation, URL_XALPHAS);/* Only escape for transmitted op */
  opsize = strlen(op)+1;
  
  /* Add the operation */
  obj->operCount++;
  oper->next = NULL;
  if (!obj->oper) {
    obj->oper = oper;
    oper->prev = NULL;
  } else {
    opr = obj->oper;
    while (opr->next) {
      opr = opr->next;
    }
    opr->next = oper;
    oper->prev = opr;
  }

  if (RoverDebug) {
    fprintf(stderr, "Tentative object:\n");
    RoverObj_print(obj); 
  }

  /* Propagate the object to the access manager's cache with the operation */
  if (!new) {
/*      fprintf(stderr,"%lu.%lu (Cache_ClientUpdate)[Rover_Export:HTTRoverClient.c]\n", tv.tv_sec, tv.tv_usec);*/
    rc = Cache_ClientUpdate(obj);
    if (rc != ROVER_OK) {
      interp->result = "Cache update error";
      return TCL_ERROR;
    }
  } else {
    rc = Cache_insert(obj);
    if (rc != ROVER_OK) {
      interp->result = "Cache insertion error";
      return TCL_ERROR;
    }
  }

  size = (3*sizeof(int))+strlen(url)+opsize;
  if ((sbuf = sbuf2 = RoverTclAlloc(size, interp)) == NULL) return TCL_ERROR;
  
  /* Marshall up url, session number, and operation */
  MarshallString(&sbuf2, url);
  MarshallInt(&sbuf2, session);
  MarshallString(&sbuf2, op);
  if (op != operation) free(op);	/* Free op if RoverEscape allocated
					     a new string */

/*  fprintf(stderr,"%lu.%lu (ClientSend)[Rover_Export:HTTRoverClient.c]\n", tv.tv_sec, tv.tv_usec);*/
  msgID = RoverClientSend(ROVER_EXPORT_OBJ, session, Rover_Finish,
			  (char *) oper, ROVER_NO_DELETE, size, sbuf, 1);
  if (msgID < 0) {
    interp->result = "Rover send error";
    return TCL_ERROR;
  }

  if (!RoverClientCheckMsgID(msgID, 1))
    rc = CreatePromise(msgID, 1, interp);
  else
    rc = PromiseName(msgID, interp);

/*  fprintf(stderr,"%lu.%lu (done)[Rover_Export:HTTRoverClient.c]\n", tv.tv_sec, tv.tv_usec);*/
  return rc;
}

/* Rover_Import <Session> <Object> [<Callback>] */
PUBLIC int Rover_Import(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
  int msgID, session, rc;
  RoverObj *obj;
  HTOper *oper;
  char *url, buf[1024], *cbuf;

  if ((argc < 3) || (argc > 4)) {
    interp->result =
      "wrong # args: Rover_Import <Session> <Object> [<Callback>]";
    return TCL_ERROR;
  }

  session = atoi(argv[1]);
  if (!SessionTable[session].inuse) {
    fprintf(stderr,
	    "Invalid session ID while importing |%s| on session %s (callback: %s) [Rover_Import:RoverClientIE.c]\n",
	    argv[2], argv[1], ((argc == 4) ? argv[3] : ""));
    interp->result = "Invalid session ID [Rover_Import:RoverClientIE.c]";
    return TCL_ERROR;
  }
  url = argv[2];
  if ((Location(session) != ROVER_APPL) && (Location(session) != ROVER_ANY) &&
      (Location(session) != ROVER_MH)) {
    rc = Location(session);
    sprintf(buf, "Session %d (%s) isn't local, mh, or any!\n", session,
	    LocationStr(rc));
    Tcl_SetResult(interp, buf, TCL_VOLATILE);
    return TCL_ERROR;
  }

  if (argc == 4) {
    /* Save the callback for later */
    if ((cbuf = RoverTclCopy(argv[3], interp)) == NULL) return TCL_ERROR;
  } else
    cbuf = NULL;

  if (!(oper=(HTOper *)RoverTclAlloc(sizeof(HTOper),interp))) return TCL_ERROR;
  oper->status = ROVER_NOTQUEUED;
  oper->oper = NULL;
  oper->import = 1;
  oper->callback = cbuf;

  if ((oper->url = RoverTclCopy(url, interp)) == NULL) return TCL_ERROR;
  oper->session = session;
  oper->logID = -1;
  oper->size = 1;

  if (RoverDebug)
    fprintf(stderr, "Importing |%s| on session %d (callback: %s)\n", url,
	    session, cbuf);

  /* Check if object is already here */
  if (RoverDebug)
    fprintf(stderr, "Checking for object locally (URL: %s Session: %d, %s)\n",
	    url, session, SessionTable[session].ServiceURN);
  msgID = -1;
  if ((obj = Cache_fetch(url, session, ROVER_IMPORT_CMD, Rover_Finish,
			 (void *)oper, &msgID))
      != NULL) {
    /* Create a completed promise */
    rc = FillPromise(msgID, oper->logID, url, interp);
    free(oper->url);
    free(oper);
    if (!TclGoodEval(rc)) return(rc);
    if (RoverDebug)  fprintf(stderr, "New promise successfully invoked\n");
    
    /* Now execute the callback */
    sprintf(buf, "Rover__Promise__%d", msgID);
    rc = Rover_InvokeCallback(obj, cbuf, NULL, -1, ROVER_OBJECT, buf,
			      1, globalTclInterp);
    /* Check callback status */
    if (!TclGoodEval(rc)) {
      sprintf(buf, "\n\t(\"%s Rover__Promise__%d\" body line %d)", cbuf,
	      msgID, interp->errorLine);
      Tcl_AddErrorInfo(interp, buf);
      return(rc);
    }

    /* Finally, return the filled-in promise */
    sprintf(buf, "Rover__Promise__%d", msgID);
    Tcl_SetResult(interp, buf, TCL_VOLATILE);
    return TCL_OK;
  }
  if (msgID < 0) {
    interp->result = "Rover send error";
    return TCL_ERROR;
  }
  
  if (!RoverClientCheckMsgID(msgID, 1))
    rc = CreatePromise(msgID, 1, interp);
  else
    rc = PromiseName(msgID, interp);
  return rc;
}

PRIVATE void Eval_ImportHook()
{
  int retVal;
  if (RoverDebug) fprintf(stderr,"[Eval_ImportHook:RoverClientIE.c]\n");
  if(ImportHook != NULL) {
    retVal = Tcl_GlobalEval(globalTclInterp, ImportHook);
    if (!TclGoodEval(retVal)) {
      showTclError(retVal,"[Eval_ImportHook:RoverClientIE.c]",globalTclInterp);
    } else if (RoverDebug)
      fprintf(stderr,"Success.[Eval_ImportHook:RoverClientIE.c]\n");
  }
}

PUBLIC int Rover_SetImportHook(ClientData clientData, Tcl_Interp *interp,
			       int argc, char *argv[])
{
 if (argc != 2) {
    interp->result = "wrong # args: Rover_SetImportHook <callback>";
    return TCL_ERROR;
 }
 
 if ((ImportHook = RoverTclCopy(argv[1], interp)) == NULL)  return TCL_ERROR;
 return TCL_OK;
}

PUBLIC int Rover_UnsetImportHook(ClientData clientData, Tcl_Interp *interp,
				 int argc, char *argv[])
{
 if (argc != 1) {
    interp->result = "wrong # args: Rover_UnsetImportHook";
    return TCL_ERROR;
 }
 
 if(ImportHook != NULL) {
   free(ImportHook);
   ImportHook = NULL;
 }
 return TCL_OK;
}

PRIVATE void Eval_ExportHook()
{
  int retVal;
/*  struct timeval ev_tv;
  struct timezone tz;
  gettimeofday(&ev_tv, &tz);
  fprintf(stderr,"%lu.%lu [Eval_ExportHook:HTTRoverClient.c]\n", ev_tv.tv_sec, ev_tv.tv_usec); */
  if (RoverDebug) fprintf(stderr,"[Eval_ExportHook:RoverClientIE.c]\n");
  if(ExportHook != NULL) {
    retVal = Tcl_GlobalEval(globalTclInterp, ExportHook);
    if (!TclGoodEval(retVal)) {
      showTclError(retVal,"[Eval_ExportHook:RoverClientIE.c]",globalTclInterp);
    } else if (RoverDebug)
      fprintf(stderr,"Success.[Eval_ExportHook:RoverClientIE.c]\n");
  }
/*  fprintf(stderr,"%lu.%lu [Eval_ExportHook:HTTRoverClient.c]\n", ev_tv.tv_sec, ev_tv.tv_usec); */
}

PUBLIC int Rover_SetExportHook(ClientData clientData, Tcl_Interp *interp,
			       int argc, char *argv[])
{
 if (argc != 2) {
    interp->result = "wrong # args: Rover_SetExportHook <callback>";
    return TCL_ERROR;
 }
 
 if ((ExportHook = RoverTclCopy(argv[1], interp)) == NULL) return TCL_ERROR;
 return TCL_OK;
}

PUBLIC int Rover_UnsetExportHook(ClientData clientData, Tcl_Interp *interp,
				 int argc, char *argv[])
{
 if (argc != 1) {
    interp->result = "wrong # args: RoverUnsetExportHook";
    return TCL_ERROR;
 }
 
 if(ExportHook != NULL) {
   free(ExportHook);
   ExportHook = NULL;
 }
 return TCL_OK;
}

