 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: base.c
 *
 * Description: Application chooser for ROVER
 *
 * Initial Version:  9-Feb-95 (adj)
 *
 * Last Modified:  13-Jul-95 (adj)
 *
 * Global Functions:
 *    main
 *
 * Global Variables: none
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define ROVER_CLIENT

#include "RoverBase.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTSession.h"
#include "stdio.h"

static int FolderUnmarshall(ClientData clientData, Tcl_Interp *interp,
			    int argc, char *argv[]);
void usage(char *arg)
{
  if (arg) fprintf(stderr, "Bad Command Line Argument: %s\n", arg);
  fprintf(stderr, "Usage: base [-d <level>] [-- <args>]\n");
  RoverExit(0);
}

void BaseInit(int applID)
{
  Tcl_CreateCommand(globalTclInterp, "FolderUnmarshall", FolderUnmarshall,
		    NULL, NULL);
}

int main(int argc, char **argv)
{
  int appID, arg, AppArgc = 0, ignore = 0;
  char **AppArgv;

  fprintf(stderr,
	  "Rover base.c client (Rover version %s, WWW Library Version %s)\n",
	  RoverLibVersion, HTLibraryVersion);

  if ((AppArgv = (char **)malloc((argc+1)*sizeof(char*))) == NULL) {
    /* Allocate space for image name and trailing null argv entry */
    fprintf(stderr, "Out of memory [main:base.c]\n");
    RoverExit(-1);
  }

  /* Check for command line options */
  for (arg=0; arg<argc ; arg++) {
    if (!ignore && (*argv[arg] == '-')) {
      
      /* - alone => error */
      if (argv[arg][1] == 0) {
	usage(argv[arg]);
	
	/* --: ignore all else (as args); treat as keywords */
      } else if (!strcmp(argv[arg], "--")) {
	ignore = 1;
	
	/* -? or -help: show the command line help page */
      } else if (!strcmp(argv[arg], "-?") ||
		 !strcmp(argv[arg], "-help")) {
	usage(NULL);
	
	/* Debug flag */
      } else if (!strcmp(argv[arg], "-d")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverDebug = atoi(argv[++arg]);

	/* endif long list of argument options */
      } else {
	usage(argv[arg]);
      }
    } else {	/* If no leading `-' (or ignore is set),
		   then check for main argument */
      if (RoverDebug)
	fprintf(stderr, "Arg %d is |%s|, storing in arg %d\n", arg, argv[arg],
		AppArgc);
      AppArgv[AppArgc++] = argv[arg];
    } /* Not an option '-'*/
  } /* End of argument loop */

  AppArgv[AppArgc] = NULL;

  if (RoverDebug) {
    int i;
    fprintf(stderr, "Got %d application args\n", AppArgc);
    for (i = 0; i < AppArgc; i++)
      fprintf(stderr, "|%s| ", AppArgv[i]);
    fprintf(stderr, "\n");
  }

  RoverAppInit = BaseInit;
  if (RoverDebug) fprintf(stderr, "Initializing client library...\n");
  appID = RoverClientInit(TCL_AND_TK, AppArgc, AppArgv);
  if (RoverDebug) fprintf(stderr, "appID is %d\n", appID);
  if (appID < 0) exit(-1);
  RoverClientMainLoop();		/* Run the chooser */
  return(0);
}

/* FolderUnmarshall <User> <Self> <Buffer> */
PRIVATE int FolderUnmarshall(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[])
{
#define TCL_FLAGS	TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG
  char *self, *user, *ptr, *next, *sp, *msgid, name[1024], number[32], *rv;
  int num, count = 0;

  if (argc < 4) {
    interp->result = "wrong # args: FolderUnmarshall <User> <Self> <Buffer>";
    return TCL_ERROR;
  }

  user = argv[1];
  self = argv[2];
  ptr = argv[3];

  while (*ptr != '\0') {
    if ((next = index(ptr, '\n')) == NULL) break;
    *next++ = '\0';
    if ((sp = index(ptr, ' ')) == NULL) {
      ptr = next;
      continue;
    }
    *sp = '\0';
    msgid = ptr;
    ptr = sp+1;
    num = atoi(ptr);
    sprintf(number, "%d", num);
    ptr += 5;

    sprintf(name, "%s(MsgIDs,%d)", self, num);
    rv = Tcl_SetVar(globalTclInterp, name, msgid, TCL_FLAGS);
    if (!rv) {
      sprintf(name, "TCL SETVAR ERR: %s [FolderUnmarshall:base.c]\n",
	      globalTclInterp->result);
      interp->result = name;
      return(TCL_ERROR);
    }

    sprintf(name, "%s(Currents,%s)", self, msgid);
    rv = Tcl_SetVar(globalTclInterp, name, " ", TCL_FLAGS);
    if (!rv) {
      sprintf(name, "TCL SETVAR ERR: %s [FolderUnmarshall:base.c]\n",
	      globalTclInterp->result);
      interp->result = name;
      return(TCL_ERROR);
    }

    sprintf(name, "%s(Numbers,%s)", self, msgid);
    rv = Tcl_SetVar(globalTclInterp, name, number, TCL_FLAGS);
    if (!rv) {
      sprintf(name, "TCL SETVAR ERR: %s [FolderUnmarshall:base.c]\n",
	      globalTclInterp->result);
      interp->result = name;
      return(TCL_ERROR);
    }

    sprintf(name, "%s(Text,%s)", self, msgid);
    rv = Tcl_SetVar(globalTclInterp, name, ptr, TCL_FLAGS);
    if (!rv) {
      sprintf(name, "TCL SETVAR ERR: %s [FolderUnmarshall:base.c]\n",
	      globalTclInterp->result);
      interp->result = name;
      return(TCL_ERROR);
    }
    
    sprintf(name, "Rover__Email__%s__Message__%s", user, msgid);
    rv = Tcl_GetVar(globalTclInterp, name, TCL_FLAGS);
    sprintf(name, "%s(Status,%s)", self, msgid);
    if (rv) rv = Tcl_SetVar(globalTclInterp, name, "Loaded", TCL_FLAGS);
    else rv = Tcl_SetVar(globalTclInterp, name, "Not Loaded", TCL_FLAGS);
    if (!rv) {
      sprintf(name, "TCL SETVAR ERR: %s [FolderUnmarshall:base.c]\n",
	      globalTclInterp->result);
      interp->result = name;
      return(TCL_ERROR);
    }
    ptr = next;
    count++;
  }

  sprintf(number, "%d", count);
  sprintf(name, "%s(Count)", self);
  rv = Tcl_SetVar(globalTclInterp, name, number, TCL_FLAGS);
  if (!rv) {
    sprintf(name, "TCL SETVAR ERR: %s [FolderUnmarshall:base.c]\n",
	    globalTclInterp->result);
    interp->result = name;
    return(TCL_ERROR);
  }
  return(TCL_OK);
}
